//---------------------------------------------------------------------------
#include <vcl.h>
#include <fstream.h>
#include <stdio.h>
#include <math.h>
#pragma hdrstop

#include "LocateSet.h"
#include "Main.h"
#include "serviceroutines.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TLocation *Location;

extern int *FavLocs;          //Favorite location indices
extern int FavLocNum;         //number of "  "  "
extern Settings PS;
extern Angles AL;
extern String SystemMessage;
extern bool ProceedFlag;
int InfoIndex;

void LoadLocation(int n);
void SaveSettings(String filename);
extern int *i_vector(int nl,int nh);
extern void free_i_vector(int *v, int nl, int nh);
extern void LoadLocation(int n,String *name,float *longitude,float *latitude,\
  float *elevation, float *TimeZone,bool *DaylightSavings);
extern String GetDecString(float dec);

//---------------------------------------------------------------------------
__fastcall TLocation::TLocation(TComponent* Owner)
    : TForm(Owner)
{
  //Load geographic location information
  FILE *infile;
  String filename,cbuf;
  char buf[33];
  int *iptr,idum=0,n,j;
  iptr=&idum;

  filename=PS.ProgPath+"Cities.dat";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }
  infile=fopen(filename.c_str(),"rb");
  fseek(infile,4,SEEK_SET);

  //Read in the city name list, but don't load the coordinates
  FavLocs=i_vector(0,FavLocNum-1);
  for (int i=0;i<FavLocNum;i++){
    fread(iptr,4,1,infile);
    FavLocs[i]=*iptr;
  }

  //Get the number of main records
  fread(iptr,4,1,infile);
  n=*iptr;
  j=0;
  for (int i=0;i<n;i++){
    //Read in the city name (32 bytes + 1 null character byte)
    fgets(buf,33,infile);
    cbuf=buf;
    cbuf=cbuf.Trim();
    //Add the name to the list
    MList->Items->Add(cbuf);
    if (i==FavLocs[j]){
      //Add to the favorite list
      FList->Items->Add(cbuf);
      MainF->LocationBox->Items->Add(cbuf);
      j++;
    }
    //Skip to next record
    fseek(infile,17,SEEK_CUR);
  }

  //Find the selected city name in the favorites list
  j=0;
  while (FavLocs[j]!=PS.location) {j++;}
  FList->ItemIndex=j;
  MainF->LocationBox->ItemIndex=j;
  MainF->StatusBar->Panels->Items[2]->Text=MainF->LocationBox->Text;

  fclose(infile);
}
//---------------------------------------------------------------------------
//What to do when the form opens
void __fastcall TLocation::FormActivate(TObject *Sender)
{
  DrawInfo(PS.location);
  MList->ItemIndex=PS.location;
  InfoIndex=MList->ItemIndex;
  SLoc->Text=MList->Items->Strings[PS.location];
  CurLoc->Caption=MList->Items->Strings[PS.location];
}
//---------------------------------------------------------------------------
//Select the location
void __fastcall TLocation::MListClick(TObject *Sender)
{
  DrawInfo(MList->ItemIndex);
  InfoIndex=MList->ItemIndex;
}
//---------------------------------------------------------------------------
//Draw the object information
void __fastcall TLocation::DrawInfo(int n)
{
  String name;
  float longitude, latitude, TimeZone, elevation;
  float deg_,min_,sec_;
  int deg,min;
  int sign=1;
  bool DaylightSavings;

  LoadLocation(n,&name,&longitude,&latitude,&elevation,&TimeZone,&DaylightSavings);

  //The name
  SLoc->Text=MList->Items->Strings[n];
  CurLoc->Caption=MList->Items->Strings[PS.location];

  //The latitude
  deg_=latitude*360/twopi;
  sign=1;
  if (deg_<0) {sign=-1; deg_=-deg_;}
  deg=deg_;
  if (deg>0) {
    if (deg>deg_) deg-=1;
    min_=(deg_-deg)*60;
  } else {
    if (deg<deg_) deg+=1;
    min_=(deg-deg_)*60;
  }
  min=min_;
  if (min>0) {
    if (min>min_) min-=1;
    sec_=(min_-min)*60;
  } else {
    if (min<min_) min+=1;
    sec_=(min-min_)*60;
  }
  LatD->Text=IntToStr(deg);
  LatM->Text=IntToStr(min);
  LatS->Text=FormatFloat("#0.00",sec_);
  if (sign<0) LatC->Caption="S";
  else LatC->Caption="N";

  //The longitude
  deg_=longitude*360/twopi;
  sign=1;
  if (deg_<0) {sign=-1; deg_=-deg_;}
  deg=deg_;
  if (deg>0) {
    if (deg>deg_) deg-=1;
    min_=(deg_-deg)*60;
  } else {
    if (deg<deg_) deg+=1;
    min_=(deg-deg_)*60;
  }
  min=min_;
  if (min>0) {
    if (min>min_) min-=1;
    sec_=(min_-min)*60;
  } else {
    if (min<min_) min+=1;
    sec_=(min-min_)*60;
  }
  LngD->Text=IntToStr(deg);
  LngM->Text=IntToStr(min);
  LngS->Text=FormatFloat("#0.00",sec_);
  LngC->Caption="W";
  if (sign<0) LngC->Caption="W";
  else LngC->Caption="E";

  //The elevation
  Elev->Text=FormatFloat("######0",elevation);

  //The time zone
  TZ->Text=FormatFloat("#0.0",TimeZone);

  //Daylight savings flag
  DS->Checked=DaylightSavings;
}
//---------------------------------------------------------------------------
//Add selected item to favorites
void __fastcall TLocation::AddToFavClick(TObject *Sender)
{
  AddItem(InfoIndex);
}
//---------------------------------------------------------------------------
//Remove the selected item from the favorites list
void __fastcall TLocation::RemFromFavClick(TObject *Sender)
{
  if (FavLocNum>1){
    //Delete from the list boxes
    int n=FList->ItemIndex;
    FList->Items->Delete(n);
    MainF->LocationBox->Items->Delete(n);

    //Decrement the FavLocNum index
    FavLocNum--;
    //Copy FavLocs[] into the temporary array (removing the old item)
    int *ibuf;
    ibuf=i_vector(0,FavLocNum);
    for (int i=0;i<FavLocNum;i++){
      if (i<n){
        ibuf[i]=FavLocs[i];
      }
      if (i>=n){
        ibuf[i]=FavLocs[i+1];
      }
    }
    //Reallocate FavLocs and copy ibuf
    free_i_vector(FavLocs,0,FavLocNum);
    FavLocs=i_vector(0,FavLocNum-1);
    for (int i=0;i<FavLocNum;i++){FavLocs[i]=ibuf[i];}
    free_i_vector(ibuf,0,FavLocNum-1);

    //Find a new location if the selected location is deleted
    int index;
    if (n<FavLocNum) index=n;
    else index=FavLocNum-1;
    PS.location=FavLocs[index];
    InfoIndex=PS.location;
    CurLoc->Caption=FList->Items->Strings[index];
    FList->ItemIndex=index;
    UseAsCur1Click(Sender);
  }
  //Save changes
  SaveLocs(false,false,0,false);
}
//---------------------------------------------------------------------------
//Change the currently selected location (MList selection method)
void __fastcall TLocation::UseAsCur1Click(TObject *Sender)
{
  //Check to see if the item is in the favorites list and if
  //not found add it to the list
  AddItem(InfoIndex);

  //Find the index in FList
  int j=0;
  while (FavLocs[j]!=InfoIndex) {j++;}

  PS.location=FavLocs[j];
  CurLoc->Caption=FList->Items->Strings[j];
  MainF->StatusBar->Panels->Items[2]->Text=FList->Items->Strings[j];

  //Change settings
  LoadLocation(InfoIndex);
}
//---------------------------------------------------------------------------
//Load location information into system variables
void LoadLocation(int n)
{
  String name;
  float longitude,latitude,elevation,timezone;
  bool daylightsavings;

  LoadLocation(n,&name,&longitude,&latitude,&elevation,&timezone,&daylightsavings);
  PS.location=n;
  PS.TimeZone=timezone;
  PS.DaylightSavings=daylightsavings;
  AL.alt=elevation;
  AL.lng=longitude;
  AL.lat=latitude;
  AL.cos_lat=cos(latitude);
  AL.sin_lat=sin(latitude);
}
//---------------------------------------------------------------------------
//Add an item to the lists
void __fastcall TLocation::AddItem(int n)
{
  //Look to see if the item is already listed
  int j=0;
  bool found=false;
  while (!found&&j<FavLocNum) {if (FavLocs[j]==n) found=true; j++;}

  //Add the item (if not found), but it must be added in
  //the correct order (i.e. not just appended to the list)
  if (!found){
    j=0;
    while ((n>FavLocs[j])&&(j<FavLocNum)) {j++;}
    FList->Items->Insert(j,MList->Items->Strings[n]);

    //Increment the FavLocNum index
    FavLocNum++;
    //Copy FavLocs[] into the temporary array (inserting the new item)
    int *ibuf;
    ibuf=i_vector(0,FavLocNum-1);
    for (int i=0;i<FavLocNum;i++){
      if (i<j){
        ibuf[i]=FavLocs[i];
      }
      if (i==j) ibuf[i]=n;
      if (i>j){
        ibuf[i]=FavLocs[i-1];
      }
    }
    //Reallocate FavLocs and copy ibuf
    free_i_vector(FavLocs,0,FavLocNum-2);
    FavLocs=i_vector(0,FavLocNum-1);
    for (int i=0;i<FavLocNum;i++){FavLocs[i]=ibuf[i];}
    free_i_vector(ibuf,0,FavLocNum-1);

    //Rewrite the header of the database file
    SaveLocs(false,false,0,false);
  }
}
//---------------------------------------------------------------------------
//Edit location (sets the MList index to the selected FList itemindex)
void __fastcall TLocation::EditClick(TObject *Sender)
{
  MList->ItemIndex=FavLocs[FList->ItemIndex];
  PageControl1->SelectNextPage(false);
  DrawInfo(MList->ItemIndex);
}
//---------------------------------------------------------------------------
//Blanks out the edit boxes for new input
void __fastcall TLocation::NewLocClick(TObject *Sender)
{
  SLoc->Text="";
  LngD->Text="";
  LngM->Text="";
  LngS->Text="";
  LatD->Text="";
  LatM->Text="";
  LatS->Text="";
  Elev->Text="";
  TZ->Text="";
  DS->Checked=false;
}
//---------------------------------------------------------------------------
//Save the new location entry
void __fastcall TLocation::SaveClick(TObject *Sender)
{
  String name;
  int item;

  //Check that formats are O.K.
  if (SLoc->Text=="") {
    FocusControl(SLoc);
    Application->MessageBox("Location must have a name!", "SkyAtlas", MB_OK);
    return;
  } else {
    if (LngD->Text=="") LngD->Text="0";
    if (LngM->Text=="") LngM->Text="0";
    if (LngS->Text=="") LngS->Text="0";
    if (LatD->Text=="") LatD->Text="0";
    if (LatM->Text=="") LatM->Text="0";
    if (LatS->Text=="") LatS->Text="0";
    if (Elev->Text=="") Elev->Text="0";
    if (TZ->Text=="") TZ->Text="0";

    //Find where to insert the new item
    String name=SLoc->Text.Trim();
    int comp=1;
    item=-1;
    while (comp>0 && item<MList->Items->Count) {
      item++;
      comp=name.AnsiCompareIC(MList->Items->Strings[item]);
    }

    if (comp==0) {
      //An item with the same name exists
      MList->Items->Delete(item);
      MList->Items->Insert(item,name);
      SaveLocs(false,true,item,false);
    } else {
      //Completely new item
      MList->Items->Insert(item,name);
      //Increment the location index if necessary... also
      //in the FavLocs
      if (PS.location>=item) PS.location=PS.location+1;
      for (int i=0;i<FavLocNum;i++){
        if (FavLocs[i]>=item) {
          FavLocs[i]=FavLocs[i]+1;
        }
      }
      SaveLocs(true,true,item,false);
    }
  }

  //Apply to current settings if appropriate
  if (PS.location==item){
    UseAsCur1Click(Sender);
  }
}
//---------------------------------------------------------------------------
//Saves changes to the database, includes insertions or
//deletions of entries
void __fastcall TLocation::SaveLocs(bool NewName,bool EditItem,int item, bool DeleteFlag)
{
  FILE *infile, *outfile;
  String FilenameIn,FilenameOut,sbuf;
  char nbuf[33],DSbuf[2];
  int *iptr,idum=0,ibuf,n,OldFavLocNum;
  float *fptr,fdum=0,fbuf;
  fptr=&fdum;
  iptr=&idum;

  FilenameIn=PS.ProgPath+"Cities.dat";
  if (!FileExists(FilenameIn)) {
    String osbuf = "Could not open file " + FilenameIn;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }
  infile=fopen(FilenameIn.c_str(),"rb");
  fseek(infile,0,SEEK_SET); //Read the old FavLocNum
  fread(iptr,4,1,infile);
  OldFavLocNum=*iptr;
  fseek(infile,OldFavLocNum*4,SEEK_CUR);

  FilenameOut=PS.ProgPath+"Cities.tmp";
  outfile=fopen(FilenameOut.c_str(),"wb");

  if (!EditItem){
    //Only the header needs to be rewritten
    fwrite(&FavLocNum,4,1,outfile);
    for (int i=0;i<FavLocNum;i++){
      ibuf=FavLocs[i];
      fwrite(&ibuf,4,1,outfile);
    }
    //Read and rewrite the remainder of the file
    fread(iptr,4,1,infile);
    n=*iptr;
    fwrite(&n,4,1,outfile);
    for (int i=0;i<n;i++){
      fgets(nbuf,sizeof(nbuf),infile);
      sbuf=nbuf;
      fprintf(outfile,"%s",sbuf.c_str());
      fread(fptr,4,1,infile);
      fwrite(fptr,4,1,outfile);
      fread(fptr,4,1,infile);
      fwrite(fptr,4,1,outfile);
      fread(fptr,4,1,infile);
      fwrite(fptr,4,1,outfile);
      fread(fptr,4,1,infile);
      fwrite(fptr,4,1,outfile);
      fgets(DSbuf,sizeof(DSbuf),infile);
      sbuf=DSbuf;
      fprintf(outfile,"%s",sbuf.c_str());
    }

  } else {

    //Rewrite the header
    fwrite(&FavLocNum,4,1,outfile);
    for (int i=0;i<FavLocNum;i++){
      ibuf=FavLocs[i];
      fwrite(&ibuf,4,1,outfile);
    }
    //Read and rewrite the remainder of the file
    fread(iptr,4,1,infile);
    n=*iptr;
    if (NewName) n++;    //An item will be added, increment the counter
    if (DeleteFlag) n--; //An item will be deleted, decrement the counter
    fwrite(&n,4,1,outfile);
    for (int i=0;i<n;i++){
      if (i!=item){
        fgets(nbuf,sizeof(nbuf),infile);
        sbuf=nbuf;
        fprintf(outfile,"%s",sbuf.c_str());
        fread(fptr,4,1,infile);
        fwrite(fptr,4,1,outfile);
        fread(fptr,4,1,infile);
        fwrite(fptr,4,1,outfile);
        fread(fptr,4,1,infile);
        fwrite(fptr,4,1,outfile);
        fread(fptr,4,1,infile);
        fwrite(fptr,4,1,outfile);
        fgets(DSbuf,sizeof(DSbuf),infile);
        sbuf=DSbuf;
        fprintf(outfile,"%s",sbuf.c_str());
      } else {
        if (!DeleteFlag){
          sbuf=SLoc->Text+"                                  ";
          sbuf=sbuf.SubString(1,32);
          fprintf(outfile,"%s",sbuf.c_str());
          fbuf=LatD->Text.ToDouble()+LatM->Text.ToDouble()/60+LatS->Text.ToDouble()/3600;
          if (LatC->Caption=="S") fbuf=-fbuf;
          fwrite(&fbuf,4,1,outfile);
          fbuf=LngD->Text.ToDouble()+LngM->Text.ToDouble()/60+LngS->Text.ToDouble()/3600;
          if (LngC->Caption=="W") fbuf=-fbuf;
          fwrite(&fbuf,4,1,outfile);
          fbuf=Elev->Text.ToDouble();
          fwrite(&fbuf,4,1,outfile);
          fbuf=TZ->Text.ToDouble();
          fwrite(&fbuf,4,1,outfile);
          if (DS->Checked) sbuf="1";
          else sbuf="0";
          fprintf(outfile,"%s",sbuf.c_str());
          //The following sets overwrite of prexisting
          //location entries
          if (!NewName) fseek(infile,49,SEEK_CUR);
        } else {
          //Skip the entry (delete)
          fseek(infile,49,SEEK_CUR);
        }
      }
    }
  }
  fclose(infile);
  fclose(outfile);

  //Rename the rewritten file and delete the temp file
  DeleteFile(FilenameIn);
  RenameFile(FilenameOut,FilenameIn);
}
//---------------------------------------------------------------------------
void __fastcall TLocation::OKButClick(TObject *Sender)
{
  //Save the lists (in particular the favorites list)
  SaveLocs(false,false,0,false);

  //Copy FList into the list on the main form
  MainF->LocationBox->Clear();
  for (int i=0;i<FavLocNum;i++){
    MainF->LocationBox->Items->Add(FList->Items->Strings[i]);
    if (FavLocs[i]==PS.location) {
      MainF->LocationBox->ItemIndex=i;
      //Make location current
      FList->ItemIndex=i;
      UseAsCur1Click(Sender);
    }
  }

  MainF->StatusBar->Panels->Items[2]->Text=CurLoc->Caption;

  //Save the changes (if any) made to the location index
  String filename=PS.ProgPath+"SkyAtlas.ini";
  SaveSettings(filename);
}
//---------------------------------------------------------------------------
//Change the E-W reference
void __fastcall TLocation::LngCClick(TObject *Sender)
{
  if (LngC->Caption=="W") LngC->Caption="E";
  else LngC->Caption="W";
}
//---------------------------------------------------------------------------
//Change the N-S reference
void __fastcall TLocation::LatCClick(TObject *Sender)
{
  if (LatC->Caption=="N") LatC->Caption="S";
  else LatC->Caption="N";
}
//---------------------------------------------------------------------------
//Delete an entry from the file
void __fastcall TLocation::DeleteButClick(TObject *Sender)
{
  ProceedFlag=false;
  if (Application->MessageBox("Proceed with delete operation?", "SkyAtlas", MB_OKCANCEL) == IDOK) {
    ProceedFlag=true;
  }

  if (ProceedFlag){
    MList->Items->Delete(InfoIndex);
    //Delete from the favorite list if it is there
    int i=0;
    while (InfoIndex!=FavLocs[i] && i<FavLocNum){i++;}
    if (i<FavLocNum){
      FList->Items->Delete(i);
      for (int j=i;j<FavLocNum-1;j++){
        FavLocs[j]=FavLocs[j+1]-1;
      }
      FavLocNum--;
    }

    //Change current settings if appropriate
    if (PS.location==InfoIndex){
      //Look to see if another favorite is available
      if (FavLocNum>0){
        PS.location=i;
        if (i>=FavLocNum) i=FavLocNum-1;
      } else {
        PS.location=InfoIndex;
      }
      UseAsCur1Click(Sender);
    } else {
      if (PS.location>InfoIndex){
        PS.location=PS.location-1;
      }
    }

    //Save the changes
    SaveLocs(false,true,InfoIndex,true);
  }
}
//---------------------------------------------------------------------------
//Set the info index
void __fastcall TLocation::FListClick(TObject *Sender)
{
  InfoIndex=FavLocs[FList->ItemIndex];
}
//---------------------------------------------------------------------------

