//---------------------------------------------------------------------------
#include <vcl.h>
#include <math.h>
#include <fstream.h>
#include <stdio.h>
#pragma hdrstop

#include "MapForm.h"
#include "ServiceRoutines.h"
#include "Main.h"
#include "Printers.hpp"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "cspin"
#pragma link "CSPIN"
#pragma resource "*.dfm"
TMapF *MapF;

//Pointers to object database matrices
//Stars
extern float *ra_star;
extern float *dec_star;
extern int *num_star;
extern short *symb_star;
extern int **StarBlkIndex;
extern short **StarBlkNum;
extern int NStars;
//RNGC objects
extern float **ra_ngc;
extern float **dec_ngc;
extern short **mag_ngc;
extern short **num_ngc;
extern short **ext_ngc;
extern short **type_ngc;
extern short **cat_ngc;
extern bool **show_ngc;
extern short *NumNGC;
//IC objects
extern float **ra_ic;
extern float **dec_ic;
extern short **mag_ic;
extern short **num_ic;
extern short **type_ic;
extern short **cat_ic;
extern bool **show_ic;
extern short *NumIC;
//Messier objects
extern float ra_mes[110];
extern float dec_mes[110];
extern short mag_mes[110];
extern short type_mes[110];
extern short mes_ngc[110];
extern bool show_mes[110];
//Lynd clouds
extern float **ra_lyn;
extern float **dec_lyn;
extern short **num_lyn;
extern short *NumLyn;
//Barnard clouds
extern float *ra_bar;
extern float *dec_bar;
extern short *num_bar;
extern short *ext_bar;
//Arp galaxies
extern float *ra_arp;
extern float *dec_arp;

//Constellations
extern int ConstTot[103];
extern float ConstRA[10000];
extern float ConstDec[10000];
extern int ConstFigNSeg[79];
extern int ConstFigNPts[180];
extern float ConstFigRA[1000];
extern float ConstFigDec[1000];
extern float ConstLabRA[88];
extern float ConstLabDec[88];
extern String ConstLabel[88];
//Solar system objects
extern double OElem[213][15]; //Instantaneous orbital elements
                         //0-99: satelites of the sun
                         //100-199: satelites of earth
                         //200-201: satelites of mars
                         //202-208: Cassini and Herschel's satelites of saturn
                         //209-212: jupiter's galilean satelites
extern OrbitalElements SSData[213]; //Information for all solar system objects

//Functions defined in ServiceRoutines.cpp
extern void FirstCircle(float *i);
extern void Planar(float *X1,float *x,float *y);
extern void Lambert(float *X1,float *x,float *y);
extern void Lagrange(float *X1,float *x,float *y);
extern void Mercator(float *X1,float *x,float *y);
extern void InversePlanar(float *X);
extern void InverseLambert(float *X);
extern void InverseMercator(float *X);
extern void SphToCosines(float *X);
extern void CosinesToSph(float *X);
extern void XRotate(float *X,float *sinangle,float *cosangle);
extern void ZRotate(float *X,float *sinangle,float *cosangle);
extern void YRotate(float *X,float *sinangle,float *cosangle);

extern float *vector(int nl, int nh);
extern void free_vector(float *v, int nl, int nh);
extern int *i_vector(int nl,int nh);
extern void free_i_vector(int *v, int nl, int nh);
extern short *s_vector(int nl,int nh);
extern void free_s_vector(short *v, int nl, int nh);
extern float **matrix(int nrl, int nrh, int ncl, int nch);
extern void free_matrix(float **m,int nrl,int nrh,int ncl,int nch);
extern short **s_matrix(int nrl, int nrh, int ncl, int nch);
extern void free_s_matrix(short **m,int nrl,int nrh,int ncl,int nch);
extern int **i_matrix(int nrl, int nrh, int ncl, int nch);
extern void free_i_matrix(int **m,int nrl,int nrh,int ncl,int nch);
extern void free_matrices(int database,int vector_size);

extern void ScreenToSky(int *x,int *y,float *q,float *lambda,float *R0,int *w2,int *h2);
extern void ReturnToScreenEdge(int *w,int *h,int *i);
extern void LoadStarMap(int MaxVmag);
extern void NextDay(struct TIME_ *t);
extern struct TIME_ GetTime(void);
extern float GetLST(void);
extern double GregToJul(TIME_ Greg_t);

extern void EclSatCoord(int n,double *T);
extern void GeoSatCoord(int n);
extern void GetOE(double *T);
extern void GetObliquity(double *T);
extern void LoadOElem(String filename);
extern void PrecessClass(int Class,double jday,double startepoch);

extern void LoadCatalog(int catalog,bool startflag);
extern void LoadConstBound(void);
extern void LoadSettings(String filename);
extern void LoadLocation(int n);
extern void SaveSettings(String filename);

extern String GetTimeDateS(TIME_ time);
extern String GetDecString(float dec, bool HiPrecision);
extern String GetRAString(float ra);
extern void HorToEqu(float *h,float *Az,float *q,float *lambda);
extern void EquToHor(float *lambda,float *q,float *Az,float *h,float SidTime);
extern void EqToGal(float *lambda, float *q);
extern void GalToEq(float *lambda, float *q);
extern void EqToEcl(float *lambda, float *q);
extern void EclToEq(float *lambda, float *q);
extern void AdvanceTime(void);
extern struct TIME_ AddSeconds(struct TIME_ t,double d);
extern void Precess_sp(float *ra,float *dec,double *T, double *T0);

//Global variables
int h_,w_,h_2,w_2,PRes;
float q_0,lambda_0,sinq0,cosq0;  //Coordinates of the map center
float R0;
float StarScale;  //Sets the scaling of the size of the star symbols
//int lastx=0,lasty=0;
//int lastXSB=150,lastYSB=100;
//int x0=0,y0=0;
//bool drawing=false;
bool StarLabels=true;
extern bool SouthHemisphere=false;
extern bool initialize;
extern bool DrawTag[432];
//extern bool TimerOn;
//bool ResetTimer=true;
//bool DrawingMap=false; //Flag to indicate that the drawing thread is currently active
//int sign=1;
//struct TIME_ StartTime;
extern bool Redraw;
extern TPoint points[10000];
Graphics::TBitmap *Image = new Graphics::TBitmap();
Graphics::TBitmap *Map = new Graphics::TBitmap();
Graphics::TBitmap *Legend = new Graphics::TBitmap();
Graphics::TBitmap *text = new Graphics::TBitmap();

//Variables that are shared across units
extern int HighlightNum=0;
extern SelectedObjects SO[100]; //Holds information for up to 100 selected objects
extern HighlightObjects HO[100];//Holds information for up to 100 highlighted objects
extern Settings PS;
extern Colors CS;
extern Angles AL;
extern String SystemMessage;

//Local functions
bool SkyToMap(float *q,float *lambda,int *x1,int *y1);
void MapToSky(int *x,int *y,float *q,float *lambda);
void FindPlaneIntersect_(int *x1,int *y1,int d,int *PlaneIndex,bool XFixed);
void DrawPlaneLine_(int *x1,int *y1,int sign, int *PlaneIndex);
bool PlaneToScreen_(int *x1, int *y1, float *lambda, int *PlaneIndex);
void ScreenToEq_(int *x, int *y, float *q, float *lambda, int *PlaneIndex);
void LongLatRange_(float *min_lambda, float *max_lambda,\
  float *min_q, float *max_q, float lambda_incr, float q_incr);

//---------------------------------------------------------------------------
__fastcall TMapF::TMapF(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
//What to do when the form opens
void __fastcall TMapF::FormActivate(TObject *Sender)
{
  String filename = ProgPath + "imblank.bmp";
  if (!FileExists(filename) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
  }
  Map->LoadFromFile(filename);
  Map->Height=MapImage->Height;
  Map->Width=MapImage->Width;
  h_=Map->Height;
  h_2=h_/2;
  w_=Map->Width;
  w_2=w_/2;
  R0=(PS.R0*h_)/PS.h;
  StarScale=0.6;
  int FOV=180*w_*2/(R0*3);
  if (FOV>30){RAGrid->Value=60; DecGrid->Value=10;   } else {
    if (FOV>20){RAGrid->Value=30; DecGrid->Value=5;  } else {
      if (FOV>10){RAGrid->Value=15; DecGrid->Value=4;} else {
        if (FOV>5){RAGrid->Value=5; DecGrid->Value=2;} else {
          RAGrid->Value=4; DecGrid->Value=1;
        }
      }
    }
  }

  //Get the equatorial coordinates of the screen center (center of the map)
  ScreenToSky(&(PS.w2),&(PS.h2),&q_0,&lambda_0,&(PS.R0),&(PS.w2),&(PS.h2));
  FirstCircle(&lambda_0);
  sinq0=sin(pihalved-q_0);
  cosq0=cos(pihalved-q_0);

  //Draw the map
  DrawMap(Sender);

  //Set up form controls
  SetForm(Sender);
}
//---------------------------------------------------------------------------
//Set up form controls
void __fastcall TMapF::SetForm(TObject *Sender)
{
  //Format the R.A. and declination masks
  float val, s;
  int h,m;
  String h_,m_,s_,sign;

  //RA mask
  val=lambda_0*12.0/pi;
  h=floor(val);
  val=(val-h)*60;
  m=floor(val);
  s=(val-m)*60;
  if (h<10) h_="0"+IntToStr(h);
  else h_=IntToStr(h);
  if (m<10) m_="0"+IntToStr(m);
  else m_=IntToStr(m);
  if (s<10) s_=" "+FormatFloat("0.0",s);
  else s_=FormatFloat("00.0",s);
  RA->Text=h_+"h "+m_+"m "+s_+"s";

  //Declination mask
  val=q_0*180.0/pi;
  if (val<0) sign="-";
  else sign="+";
  val=fabs(val);
  h=floor(val);
  val=(val-h)*60;
  m=floor(val);
  s=(val-m)*60;
  if (h<10) h_=sign+"0"+IntToStr(h);
  else h_=sign+IntToStr(h);
  if (m<10) m_="0"+IntToStr(m);
  else m_=IntToStr(m);
  if (s<10) s_=" "+FormatFloat("0.0",s);
  else s_=FormatFloat("00.0",s);
  Dec->Text=h_+" "+m_+"m "+s_+"s";

  StarSizes->Value=1+(StarScale-0.6)/0.2;

  FOV->Value=180*w_*2/(R0*3);
  FOV->Repaint();

  ShowStarLabels2->Checked=StarLabels;
}
//---------------------------------------------------------------------------
//Load settings from form controls
void __fastcall TMapF::LoadSettings(TObject *Sender)
{
  String h_,m_,s_,sbuf;
  float sign;

  sbuf=RA->Text.Trim();
  lambda_0=(sbuf.SubString(1,2).ToInt()+(sbuf.SubString(5,2).ToInt())/60.0+(sbuf.SubString(9,4).ToDouble())/3600.0)*pi/12.0;
  sbuf=Dec->Text.Trim();
  if (sbuf.SubString(1,1)=="-") sign=-1;
  else sign=1;
  q_0=sign*(sbuf.SubString(2,2).ToInt()+(sbuf.SubString(6,2).ToInt())/60.0+(sbuf.SubString(10,4).ToDouble())/3600.0)*pi/180.0;
  sinq0=sin(pihalved-q_0);
  cosq0=cos(pihalved-q_0);

  StarScale=(StarSizes->Value-1)*0.2 + 0.6;

  R0=180*w_*2/(3*FOV->Value);

  StarLabels=ShowStarLabels2->Checked;
}
//---------------------------------------------------------------------------
//Controls drawing of the map
void __fastcall TMapF::DrawMap(TObject *Sender)
{
  //Draw the sky
  Map->Canvas->Pen->Color=clWhite;
  Map->Canvas->Brush->Color=clWhite;
  if (R0<4*w_){
    if (PS.Projection==0){
      Map->Canvas->Ellipse(w_2-R0/2,h_2-R0/2,w_2+R0/2,h_2+R0/2);
    } else {if (PS.Projection==1){
      Map->Canvas->Ellipse(w_2-R0,h_2-R0,w_2+R0,h_2+R0);
    } else {if (PS.Projection==2){
      Map->Canvas->Rectangle(0,0,w_,h_);
    } else {
      Map->Canvas->Rectangle(w_2-R0*pihalved,0,w_2+R0*pihalved,h_);
    }}}
  } else {
    //In this case the edge of the field of view is not visible,
    //so just paint the entire screen
    Map->Canvas->Rectangle(0,0,w_,h_);
  }

  //Get the latitude and longitude range
  float min_lambda,max_lambda,min_q,max_q,lambda_incr,q_incr;
  lambda_incr=pi*RAGrid->Value/(12*60);
  q_incr=pi*DecGrid->Value/180;
  LongLatRange_(&min_lambda,&max_lambda,&min_q,&max_q,lambda_incr,q_incr);

  if (StarLabels) DrawStarLabels();
//  if (PS.DrawConstLab) DrawConstLab();
//  if (PS.DrawConstBound&&!OnlyStars) DrawConstBound();
  if (PS.DrawConstFig) DrawConstFig();
  if (PS.ShowHorLine) DrawIntersects(0);
  if (PS.ShowEclLine) DrawIntersects(1);
  if (PS.ShowGalLine) DrawIntersects(2);
  if (Stars1->Checked) ShowStars(StarScale);
//  if (PS.ShowPlanets&&!OnlyStars) DrawPlanets(PS.time.jul);
//  if (PS.ShowDeepSky&&!OnlyStars) DrawRNGC();
//  DrawHighlights();
  if (PS.ShowGrid) DrawGrid(&min_lambda,&max_lambda,&min_q,&max_q);

  Map->Canvas->Pen->Color=clBlack;
  Map->Canvas->Brush->Style=bsClear;
  Map->Canvas->Pen->Width=w_/200;
  Map->Canvas->Rectangle(0,0,w_,h_);
  Map->Canvas->Brush->Style=bsSolid;

  DrawLegend(Sender,StarScale);

  MapImage->Picture->Bitmap=Map;
  LegendImg->Picture->Bitmap=Legend;
}
//---------------------------------------------------------------------------
//Set the drawing tag array and determine the range of
//latitudes and longitudes to use when drawing the grid
void LongLatRange_(float *min_lambda, float *max_lambda,\
  float *min_q, float *max_q, float lambda_incr, float q_incr)
{
  int i,j,x1=0,y1=0,n,m;
  float q, lambda;
  bool proceed;
  bool NPoleVisible=false, SPoleVisible=false;
  *max_q=-pihalved;
  *min_q=pihalved;
  *max_lambda=0;
  *min_lambda=twopi;

  for (lambda=0; lambda<=twopi; lambda+=lambda_incr){
    for (q=-pihalved; q<=pihalved; q+=q_incr){
      i=18*(lambda*12/pi)+18*(pihalved-q)/pi;
      DrawTag[i]=SkyToMap(&q,&lambda,&x1,&y1);
      if (DrawTag[i]){
        if (lambda>*max_lambda) *max_lambda=lambda;
        if (lambda<*min_lambda) *min_lambda=lambda;
        if (q>*max_q) *max_q=q;
        if (q<*min_q) *min_q=q;
      }
    }
  }
  *max_lambda=*max_lambda+lambda_incr;
  *min_lambda=*min_lambda-lambda_incr;
  *max_q=*max_q+q_incr;
  *min_q=*min_q-q_incr;
  if (*max_lambda>twopi) *max_lambda=twopi;
  if (*min_lambda<0) *min_lambda=0;
  if (*max_q>pihalved) *max_lambda=pihalved;
  if (*min_q<-pihalved) *min_q=-pihalved;

  //Make sure the block in the center of the screen is
  //checked (and its nearest neighbours)
  x1=w_2;
  y1=h_2;
  MapToSky(&x1,&y1,&q,&lambda);
  i=lambda_incr;
  j=(pihalved-q)*q_incr;
  for (int k=i-1;k<i+2;k++){
    for (int l=j-1;l<j+2;l++){
      n=k;
      m=l;
      if (k>23) n-=24;
      if (k<0) n+=24;
      if (l>17) m-=18;
      if (l<0) m+=18;
      DrawTag[n*18+m]=true;
    }
  }
  for (i=0;i<432;i++){
    DrawTag[i]=true;
  }

  //See if one of the poles are visible
  q=pihalved; lambda=0;
  proceed=SkyToMap(&q,&lambda,&i,&j);
  if (proceed) {
    NPoleVisible=true;
  } else {
    q=-pihalved; lambda=0;
    proceed=SkyToMap(&q,&lambda,&i,&j);
    if (proceed){
      SPoleVisible=true;
    }
  }

  //Return the minimum and maximum declinations/RA's to
  //their closest integral grid value
  *min_lambda-=lambda_incr;
  *max_lambda+=lambda_incr;
  if (*max_lambda>twopi) *max_lambda=twopi;

  *min_q-=q_incr;
  if (*min_q<-pihalved) *min_q=-pihalved;

  *max_q+=q_incr;
  if (*max_q>pihalved) *max_q=pihalved;

  if (*max_lambda==0) *max_lambda=twopi;
  if (*min_lambda==twopi) *min_lambda=0;
  if (*max_lambda==twopi && *min_lambda==0){
    if (SouthHemisphere) *min_q=-pihalved;
    else *max_q=pihalved;
  } else {
    if (*max_q==-pihalved) *max_q=pihalved;
  }
  if (*max_lambda==twopi && *min_lambda==0){
    if (SouthHemisphere) *min_q=-pihalved;
    else *max_q=pihalved;
  } else {
    if (*min_q==pihalved) *min_q=-pihalved;
  }

  //If the south and/or north pole is visible, then
  //use the following settings
  if (NPoleVisible){ //North pole visible
    *max_q=pihalved; *min_lambda=0; *max_lambda=twopi;
  }
  if (SPoleVisible){ //South pole visible
    *min_q=-pihalved; *min_lambda=0; *max_lambda=twopi;
  }
}
//---------------------------------------------------------------------------
//Draw the grid
void __fastcall TMapF::DrawGrid(float *min_lambda,float *max_lambda,\
  float *min_q,float *max_q)
{
  float incr=0,fine_incr,MajLngInc,MajLatInc,val;
  float lambda=0,lambda1=0,q=0,q1=0,h_,m_,m__;
  float startlat=0,endlat=0,startlong=0,endlong=0;
  int i,x1=0,y1=0;
  bool testexpression, SpecialIncr, drawingline, proceed, proceed2, first_point;
  String label;
  TRect ImArea=Rect(0,0,w_,h_);
  Map->Canvas->Brush->Color=clWhite;

  if (AL.lat>pihalved) SouthHemisphere=true;

  Map->Canvas->Pen->Width=1;

  //Grid colors
  Map->Canvas->Pen->Color=clBlack;
  Map->Canvas->Font->Size=8;
  Map->Canvas->Font->Color=clBlack;
  Map->Canvas->Font->Height=Map->Width/40;
  Map->Canvas->Font->Name="Arial";

  //Draws the lines of longitude
  MajLngInc=RAGrid->Value*pi/(60*12);

  for (lambda=*min_lambda;lambda<=*max_lambda;lambda+=MajLngInc){
    if (((lambda<=1e-5)&&(lambda>=0))\
      ||((lambda<=pihalved+1e-5)&&(lambda>=pihalved-1e-5))\
      ||((lambda<=pi+1e-5)&&(lambda>=pi-1e-5))\
      ||((lambda<=threepihalved+1e-5)&&(lambda>=threepihalved-1e-5))){
      SpecialIncr=true; //These lines are drawn all the way to the pole
    } else {
      SpecialIncr=false;
    }

    if (SouthHemisphere){
      if (!PS.HorizonMask){
        if (SpecialIncr) startlat=pihalved;
        else startlat=pihalved-0.174532925;
      } else {
        //Find the start latitude
        float arg,longitude;
        if (AL.sin_lat==0) arg=1e+15;
        else arg=AL.sin_lat/AL.cos_lat;
        longitude=lambda-AL.l_ori+pihalved;
        FirstCircle(&longitude);
        startlat=atan2(-sin(longitude),arg);
      }
      if (startlat>*min_q) startlat=*min_q;

      if (SpecialIncr) endlat=-pihalved;
      else endlat=-pihalved+0.174532925;
      if (endlat<*max_q) endlat=*max_q;
    } else {
      if (!PS.HorizonMask){
        if (SpecialIncr) startlat=-pihalved;
        else startlat=-pihalved+0.174532925;
      } else {
        //Find the start latitude
        float arg,longitude;
        if (AL.sin_lat==0) arg=1e+15;
        else arg=AL.sin_lat/AL.cos_lat;
        longitude=lambda-AL.l_ori+pihalved;
        FirstCircle(&longitude);
        startlat=atan2(-sin(longitude),arg);
      }
      if (startlat<*min_q) startlat=*min_q;

      if (SpecialIncr) endlat=pihalved;
      else endlat=pihalved-0.174532925;
      if (endlat>*max_q) endlat=*max_q;
    }
    i=R0/w_2+1;
    if (PS.Projection==3) incr=(endlat-startlat)/(60*i);
    else incr=(endlat-startlat)/(30*i);
    fine_incr=incr/20;

    //Draw a longitude line
    i=0;
    first_point=true;
    q=startlat;
    if (startlat<endlat) testexpression=(q<=endlat);
    else testexpression=(q>=endlat);
    while (testexpression){
      q1=q;

      //Get the screen coordinates
      proceed=SkyToMap(&q1,&lambda,&x1,&y1);
      if (first_point) {
        drawingline=proceed;
        first_point=false;
      }

      //Add the point to the point array
      if (proceed){
        if (drawingline){
          points[i].x=x1;
          points[i].y=y1;
          i++;

          //Increment the latitude
          q+=incr;
          if (!SouthHemisphere) testexpression=(q<=endlat+incr/2);
          else testexpression=(q>=endlat+incr/2);
        } else {
          //Find the beginning of the line
          q1-=incr;
          do {
            proceed2=SkyToMap(&q1,&lambda,&x1,&y1);
            q1+=fine_incr;
          } while (!proceed2);
          points[i].x=x1;
          points[i].y=y1;
          i++;
          drawingline=true;
          val=lambda*3.819718634;
          h_=floor(val);
          m__=(val-h_)*60.0;
          if (m__>59.5) {m__=0; h_++;}
          if (h_>23) h_-=24;
          if (m__>.5) label=IntToStr((int)h_)+"h "+FormatFloat("#0",m__)+"m";
          else label=IntToStr((int)h_)+"h";
          DrawGridLabel(x1,y1,label);
        }
      } else {
        if (drawingline){
          //Find the end of the line
          q1-=incr;
          do {
            proceed2=SkyToMap(&q1,&lambda,&x1,&y1);
            q1+=fine_incr;
          } while (proceed2);
          points[i].x=x1;
          points[i].y=y1;
          DrawLine(i,false);
          i=0;
          drawingline=false;
          val=lambda*3.819718634;
          h_=floor(val);
          m__=(val-h_)*60.0;
          if (m__>59.5) {m__=0; h_++;}
          if (h_>23) h_-=24;
          if (m__>.5) label=IntToStr((int)h_)+"h "+FormatFloat("#0",m__)+"m";
          else label=IntToStr((int)h_)+"h";
          DrawGridLabel(x1,y1,label);
        } else {
          //Increment the latitude
          q+=incr;
          if (!SouthHemisphere) testexpression=(q<=endlat+incr/2);
          else testexpression=(q>=endlat+incr/2);
        }
      }
    }
    if (drawingline) {
      //All of the last points were good
      DrawLine(i-1,false);
    }
  }

  //Draws the lines of latitude
  startlat=*min_q;
  endlat=*max_q;
  MajLatInc=DecGrid->Value*pi/180;

  for (q=startlat;q<=endlat;q+=MajLatInc){
    q1=q;
    startlong=*min_lambda;
    endlong=*max_lambda;
    i=R0/w_2+1;
    if (PS.Projection==3){
      incr=(endlong-startlong)/(120*i);
    } else {
      incr=(endlong-startlong)/(80*i);
    }
    fine_incr=incr/20;
    i=0;

    //Draw a latitude line
    first_point=true;
    lambda=startlong;
    while (lambda<=endlong+incr/2){
      lambda1=lambda;

      //Get the screen coordinates
      proceed=SkyToMap(&q1,&lambda1,&x1,&y1);
      if (first_point) {
        drawingline=proceed;
        first_point=false;
      }

      //Add the point to the point array
      if (proceed){
        if (drawingline){
          points[i].x=x1;
          points[i].y=y1;
          i++;

          //Increment the longitude
          lambda+=incr;
        } else {
          //Find the beginning of the line
          lambda1-=incr;
          do {
            proceed2=SkyToMap(&q1,&lambda1,&x1,&y1);
            lambda1+=fine_incr;
          } while (!proceed2);
          points[i].x=x1;
          points[i].y=y1;
          i++;
          drawingline=true;
          val=q*57.29577951;
          label=FormatFloat("#0",val)+"";
          DrawGridLabel(x1,y1,label);
        }
      } else {
        if (drawingline){
          //Find the end of the line
          lambda1-=incr;
          do {
            proceed2=SkyToMap(&q1,&lambda1,&x1,&y1);
            lambda1+=fine_incr;
          } while (proceed2);
          points[i].x=x1;
          points[i].y=y1;
          DrawLine(i,false);
          i=0;
          drawingline=false;
          val=q*57.29577951;
          label=FormatFloat("#0",val)+"";
          DrawGridLabel(x1,y1,label);
        } else {
          //Increment the longitude
          lambda+=incr;
        }
      }
    }
    if (drawingline) {
      //All of the last points were good
      DrawLine(i-1,false);
    }
  }
}
//---------------------------------------------------------------------------
//Draws a line, filtering out discontinuities across
//the screen found when using a Mercator projection
void __fastcall TMapF::DrawLine(int n,bool ForceVerify)
{
  if (PS.Projection==3||ForceVerify){
    TPoint n_points[1000];
    int last_x1,last_y1,i,j;
    i=0;
    while (i<n){
      j=0;
      do {
        n_points[j].x=points[i].x;
        n_points[j].y=points[i].y;
        last_x1=points[i].x;
        last_y1=points[i].y;
        i++;
        j++;
      } while ((abs(points[i].x-last_x1)<h_2/4)&&(abs(points[i].y-last_y1)<h_2/4)&&(i<=n));

      Map->Canvas->Polyline(n_points,j-1);
    }
  } else {
    Map->Canvas->Polyline(points,n);
  }
}
//---------------------------------------------------------------------------
//Draws a grid line label
void __fastcall TMapF::DrawGridLabel(int x, int y,String label)
{
  int tw,tw_,th;
  th=Map->Canvas->TextHeight(label);
  tw=Map->Canvas->TextWidth(label+" ");
  tw_=Map->Canvas->TextWidth(" ");
  if (x<tw_) x=tw_;
  if (x>w_-tw) x=w_-tw;
  if (y<th*.2) y=th*.2;
  if (y>h_-th*1.2) x=h_-th*1.2;
  Map->Canvas->TextOut(x,y,label);
}
//---------------------------------------------------------------------------
//Draw special plane intersections (horizon, ecliptic and galactic plane)
//PlaneIndex = 0 for horizon, 1 for ecliptic, 2 for galactic
void __fastcall TMapF::DrawIntersects(int PlaneIndex)
{
  int x1, dx1, y1, dy1, xint, yint, bestx1, ncross,n;
  float q=0, lambda=0, clambda=0, cq=0, lastq, dq, lastdq, bestq, Minlambda, Maxlambda, dlambda, r02, minq;
  bool proceed;

  //Find the coordinates of the center of the screen (in the coordinate system
  //of interest)
  x1=w_2;
  y1=h_2;
  ScreenToEq_(&x1,&y1,&cq,&clambda,&PlaneIndex);

  //Draw the lines
  Map->Canvas->Pen->Style=psDash;
  if (PlaneIndex==0) {
    Map->Canvas->Pen->Color=clBlack;
    Map->Canvas->Pen->Style=psSolid;
  } else {
    Map->Canvas->Pen->Style=psDash;
  }
  if (PlaneIndex==1) Map->Canvas->Pen->Color=clBlack;
  if (PlaneIndex==2) Map->Canvas->Pen->Color=clBlack;

  //In the case where the scale is sufficiently large, simply draw the full
  //circle
  if (R0<w_){
    n=0;
    for (lambda=0; lambda<=twopi; lambda+=pi/160){
      if (PlaneToScreen_(&x1,&y1,&lambda,&PlaneIndex)){
        points[n].x=x1;
        points[n].y=y1;
        n++;
      }
    }
    DrawLine(n-1,true);
  } else {
    bool Drawing, IsGood;
    float maxlambda;

    lambda=0;
    dlambda=(pi*w_2)/(120*R0);
    x1=w_2; y1=h_2;
    ScreenToEq_(&x1,&y1,&q,&lambda,&PlaneIndex);
    maxlambda=lambda+4*pi*w_2/R0;
    lambda=lambda-4*pi*w_2/R0;

    while (lambda<=maxlambda){
      IsGood=PlaneToScreen_(&x1,&y1,&lambda,&PlaneIndex);
      if (IsGood){
        Drawing=true;
        lambda-=dlambda;
        PlaneToScreen_(&x1,&y1,&lambda,&PlaneIndex);
        n=0;
        points[n].x=x1;
        points[n].y=y1;
        lambda+=dlambda;
        PlaneToScreen_(&x1,&y1,&lambda,&PlaneIndex);
        n++;
        points[n].x=x1;
        points[n].y=y1;

        while (lambda<=maxlambda && Drawing){
          lambda+=dlambda;
          Drawing=PlaneToScreen_(&x1,&y1,&lambda,&PlaneIndex);
          points[n].x=x1;
          points[n].y=y1;
          n++;
        }
        DrawLine(n-1,true);
      }
      lambda+=dlambda;
    }
  }

  if (PlaneIndex==0){
    //Draw the North, South, East, West labels
    float q_=0, lambda_=0;
    TRect source,dest;
    Map->Canvas->CopyMode=cmSrcCopy;
    for (int i=0;i<8;i++){
      q=0;
      lambda=i*piquartered;
      //Convert horizon coordinate to sky coordinates
      HorToEqu(&q,&lambda,&q_,&lambda_);
      //Get the screen coordinates
      proceed=SkyToMap(&q_,&lambda_,&x1,&y1);
      if (proceed&&(x1<=w_)&&(x1>=0)&&(y1<=h_)&&(y1>=0)){
        dest=Rect(x1-10,y1-8,x1+11,y1+9);
        source=Rect(i*21,0,(i+1)*21,17);
        Map->Canvas->CopyRect(dest,text->Canvas,source);
      }
    }
  }
  Map->Canvas->Pen->Style=psSolid;
}
//---------------------------------------------------------------------------
//Find intersection of plane with screen edge
void FindPlaneIntersect_(int *x1,int *y1,int d,int *PlaneIndex,bool XFixed)
{
  float q=0, lambda=0, qinit;

  ScreenToEq_(x1,y1,&q,&lambda,PlaneIndex);
  qinit=q;

  if (!XFixed){
    while (qinit*q>0){
      *x1=*x1+d;
      ScreenToEq_(x1,y1,&q,&lambda,PlaneIndex);
    }
  } else {
    while (qinit*q>0){
      *y1=*y1+d;
      ScreenToEq_(x1,y1,&q,&lambda,PlaneIndex);
    }
  }
}
//---------------------------------------------------------------------------
//Draw the plane line starting at screen coordinate (x1, y1)
void DrawPlaneLine_(int *x1,int *y1,int sign, int *PlaneIndex)
{
  float lambda=0, q=0, dlambda, lambda_;
  int n=0, xinit;

  ScreenToEq_(x1,y1,&q,&lambda,PlaneIndex);
  if (fabs(q)>0.05){
    *y1=*y1-1;
    ScreenToEq_(x1,y1,&q,&lambda,PlaneIndex);
  }
  dlambda=-sign*R0*pihalved/(w_2*60);
  points[0].x=*x1;
  points[0].y=*y1;
  n++;
  lambda+=dlambda;
  if (!PlaneToScreen_(x1,y1,&lambda,PlaneIndex)) {
    lambda=lambda-2*dlambda;
    dlambda=-dlambda;
  }

  while (PlaneToScreen_(x1,y1,&lambda,PlaneIndex)) {
    points[n].x=*x1;
    points[n].y=*y1;
    n++;
    lambda+=dlambda;
  }
  points[n].x=*x1;
  points[n].y=*y1;
  n++;
  MapF->DrawLine(n-1,false);
}
//---------------------------------------------------------------------------
//Get screen coordinates for a particular coordinate system, but only along
//its 0 latitude circle (used for drawing equator planes)
bool PlaneToScreen_(int *x1, int *y1, float *lambda, int *PlaneIndex)
{
  float q_=0, lambda_=0, q=0;
  bool Proceed;

  if (*PlaneIndex==0) {
    //Convert horizon coordinates to equatorial coordinates
    HorToEqu(&q,lambda,&q_,&lambda_);
  }
  if (*PlaneIndex==1) {
    //Convert ecliptic coordinates to equatorial coordinates
    lambda_=*lambda;
    EclToEq(&lambda_,&q_);
  }
  if (*PlaneIndex==2){
    //Convert galactic coordinates to equatorial coordinates
    lambda_=*lambda;
    GalToEq(&lambda_,&q_);
  }
  Proceed=SkyToMap(&q_,&lambda_,x1,y1);
  return Proceed;
}
//---------------------------------------------------------------------------
//Get coordinates for a particular coordinate system from the screen (used for
//drawing special planes)
void ScreenToEq_(int *x, int *y, float *q, float *lambda, int *PlaneIndex)
{
  bool Proceed;

  MapToSky(x,y,q,lambda);
  if (*PlaneIndex==0) {
    //Convert equatorial coordinates to horizon coordinates
    float q_, lambda_;
    q_=*q;
    lambda_=*lambda;
    EquToHor(&lambda_,&q_,lambda,q,AL.l_ori);
  }
  if (*PlaneIndex==1) {
    //Convert equatorial coordinates to ecliptic coordinates
    EqToEcl(lambda,q);
  }
  if (*PlaneIndex==2){
    //Convert equatorial coordinates to galactic coordinates
    EqToGal(lambda,q);
  }
}
//---------------------------------------------------------------------------
//Draw the stars on the map
void __fastcall TMapF::ShowStars(float sc)
{
  float Ra,Declination;
  int symb_index,mag_index,base_index,index,size_index,CatNum;

  Map->Canvas->CopyMode=cmSrcPaint;
  mag_index=(PS.MaxStarMag-150)/50;

  for (int l=0;l<432;l++){
    //Only draw stars for blocks that are visible
    if (DrawTag[l]){
      if (l>0) base_index=StarBlkIndex[mag_index][l-1];
      else base_index=0;
      for (int k=0;k<StarBlkNum[mag_index][l];k++){
        index=base_index+k;
        Ra=ra_star[index];
        Declination=dec_star[index];
        size_index=symb_star[index]/7;
        CatNum=num_star[index];
        symb_index=symb_star[index]-size_index*7+PS.StarSizes[size_index];
        DrawStar(&Ra,&Declination,&symb_index,sc,CatNum);
      }
    }
  }
}
//---------------------------------------------------------------------------
//Draws a single star on the map
void __fastcall TMapF::DrawStar(float *RA,float *dec,int *symb_index,float sc,int CatNum)
{
  int x1,y1;
  bool proceed;

  if (AL.lat>pihalved) SouthHemisphere=true;

  //Get the screen coordinates
  proceed=SkyToMap(dec,RA,&x1,&y1);

  Map->Canvas->Brush->Color=clBlack;

  //Draw the star
  if (proceed){
    if (*symb_index<7){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-11*sc,y1-11*sc,x1+11*sc,y1+11*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-10*sc,y1-10*sc,x1+10*sc,y1+10*sc);
    } else { if (*symb_index<14){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-11*sc,y1-11*sc,x1+11*sc,y1+11*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-10*sc,y1-10*sc,x1+10*sc,y1+10*sc);
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Brush->Color=clWhite;
      Map->Canvas->Ellipse(x1-5*sc,y1-5*sc,x1+5*sc,y1+5*sc);
    } else { if (*symb_index<21){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-9*sc,y1-9*sc,x1+9*sc,y1+9*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-8*sc,y1-5,x1+8*sc,y1+8*sc);
    } else { if (*symb_index<28){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-9*sc,y1-9*sc,x1+9*sc,y1+9*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-8*sc,y1-8*sc,x1+8*sc,y1+8*sc);
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Brush->Color=clWhite;
      Map->Canvas->Ellipse(x1-4*sc,y1-4*sc,x1+4*sc,y1+4*sc);
    } else { if (*symb_index<35){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-8*sc,y1-8*sc,x1+8*sc,y1+8*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-7*sc,y1-7*sc,x1+7*sc,y1+7*sc);
    } else { if (*symb_index<42){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-8*sc,y1-8*sc,x1+8*sc,y1+8*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-7*sc,y1-7*sc,x1+7*sc,y1+7*sc);
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Brush->Color=clWhite;
      Map->Canvas->Ellipse(x1-3*sc,y1-3*sc,x1+3*sc,y1+3*sc);
    } else { if (*symb_index<49){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-7*sc,y1-7*sc,x1+7*sc,y1+7*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-6*sc,y1-6*sc,x1+6*sc,y1+6*sc);
    } else { if (*symb_index<56){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-7*sc,y1-7*sc,x1+7*sc,y1+7*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-6*sc,y1-6*sc,x1+6*sc,y1+6*sc);
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Brush->Color=clWhite;
      Map->Canvas->Ellipse(x1-3*sc,y1-3*sc,x1+3*sc,y1+3*sc);
    } else { if (*symb_index<63){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-6*sc,y1-6*sc,x1+6*sc,y1+6*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-5*sc,y1-5*sc,x1+5*sc,y1+5*sc);
    } else { if (*symb_index<70){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-5*sc,y1-5*sc,x1+5*sc,y1+5*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-4*sc,y1-4*sc,x1+4*sc,y1+4*sc);
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Brush->Color=clWhite;
      Map->Canvas->Ellipse(x1-2*sc,y1-2*sc,x1+2*sc,y1+2*sc);
    } else { if (*symb_index<77){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-3*sc,y1-3*sc,x1+3*sc,y1+3*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-2*sc,y1-2*sc,x1+2*sc,y1+2*sc);
    } else { if (*symb_index<84){
      Map->Canvas->Pen->Color=clWhite;
      Map->Canvas->Ellipse(x1-2*sc,y1-2*sc,x1+2*sc,y1+2*sc);
      Map->Canvas->Pen->Color=clBlack;
      Map->Canvas->Ellipse(x1-1*sc,y1-1*sc,x1+1*sc,y1+1*sc);
    }}}}}}}}}}}}
  }
}
//---------------------------------------------------------------------------
//Draw the star labels
void __fastcall TMapF::DrawStarLabels(void)
{
  float Ra,Declination,*fptr,fdum;
  int n,SAO,idum,*iptr,x1,y1;
  iptr=&idum;
  fptr=&fdum;
  char buf[3],bufFlam[10],bufName[24];
  FILE *namefile, *mycat;
  bool proceed;
  String ComName="", Flamsteed="";
  String filename;

  filename=PS.ProgPath+"Catalogs\\SAO\\NAME_CROSSREF.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }
  namefile=fopen(filename.c_str(),"rb");
  filename=PS.ProgPath+"Catalogs\\SAO\\SAO_FULL.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }
  mycat=fopen(filename.c_str(),"rb");

  n=0;
  while (n<3126){
    //Get the SAO number
    fseek(namefile,38*n,SEEK_SET);
    fread(iptr,4,1,namefile);
    SAO=*iptr;

    //Get the star dec and RA
    fseek(mycat,(SAO-1)*31+4,SEEK_SET);
    fread(fptr,4,1,mycat);
    Ra=*fptr;
    fseek(mycat,4,SEEK_CUR);
    fread(fptr,4,1,mycat);
    Declination=*fptr;

    //Get the screen coordinates
    proceed=SkyToMap(&Declination,&Ra,&x1,&y1);

    if (proceed){
      //Get the Flamsteed label
      fgets(bufFlam,11,namefile);
      Flamsteed=bufFlam;
      //Get the common name
      fgets(bufName,25,namefile);
      ComName=bufName;
      ComName=ComName.Trim();
      //Call the formating and display routine
      FormatStarName(Flamsteed,ComName,x1,y1);
    }
    n++;
  }
  fclose(namefile);
  fclose(mycat);
}
//---------------------------------------------------------------------------
//Format and display a star name
void __fastcall TMapF::FormatStarName(String Flamsteed,String ComName,int x1,int y1)
{
  String GLetter[24]={"Alp","Bet","Gam","Del","Eps","Zet",\
    "Eta","The","Iot","Kap","Lam","Mu ","Nu ","Xi ","Omi",\
    "Pi ","Rho","Sig","Tau","Ups","Phi","Chi","Psi","Ome"};
  String GFLetter[24]={"Alpha","Beta","Gamma","Delta",\
    "Epsilon","Zeta","Eta","Theta","Iota","Kappa","Lambda",\
    "Mu","Nu","Xi","Omicron","Pi","Rho","Sigma","Tau",\
    "Upsilon","Phi","Chi","Psi","Omega"};
  String Equival[24]={"\"","$","(","*",",",".","0","2","4",\
    "6","8",":","<",">","@","B","D","F","J","L","N","P","R",\
    "T"};

  int j, left_x, char_w;
  String SrcStr,prefix,grk,multinum,abrev,CommonName;
  bool found;

  //Get the greek letter (if any)
  grk=Flamsteed.SubString(4,3);
  found=false;
  j=-1;
  do {
    j++;
    if (GLetter[j]==grk) found=true;
  } while ((j<24)&&(!found));

  //Assign the appropriate character code
  if (found) grk=Equival[j];
  else grk="";

  //The name prefix (the Flamsteed number)
  prefix=(Flamsteed.SubString(1,3)).Trim();

  Map->Canvas->Font->Color=clBlack;
  Map->Canvas->Font->Height=Map->Width/50;
  char_w=Map->Canvas->Font->Height/2;
  left_x=char_w;

  Map->Canvas->Font->Name="Arial";
  Map->Canvas->TextOut(x1+left_x,y1,prefix);
  left_x+=(Map->Canvas->TextWidth(prefix)+char_w);

  Map->Canvas->Font->Name="WP Greek Century";
  Map->Canvas->TextOut(x1+left_x,y1,grk);
  left_x+=(Map->Canvas->TextWidth(grk)+char_w);

  Map->Canvas->Font->Name="Arial";
  Map->Canvas->TextOut(x1+left_x,y1,ComName);
}
//---------------------------------------------------------------------------
//Returns the map direction cosines (x1,y1) of a sky coordinate (lambda,q)
//If the point is not hidden by other parts of the map then the function
//returns "true".
bool SkyToMap(float *q,float *lambda,int *x1,int *y1)
{
  bool proceed=true;
  float *X1,x,y;
  X1=vector(0,2);

  X1[0]=1;
  X1[1]=*q;
  X1[2]=*lambda-lambda_0-pihalved;  //Rotate about z-axis by the negative central longitude

  //Convert equatorial sky angles to rectangular direction
  //cosines
  SphToCosines(X1);

  //Rotate about the x-axis by the difference in latitude
  XRotate(X1,&sinq0,&cosq0);

  if (PS.Projection==3) Mercator(X1,&x,&y);
  if (PS.Projection==2) Planar(X1,&x,&y);
  if (PS.Projection==1) Lambert(X1,&x,&y);
  if (PS.Projection==0) { //Stereographic
    x=-X1[0]/2;
    y=X1[1]/2;
  }
  *x1=w_2+x*R0;
  *y1=h_2-y*R0;

  //Make sure "hidden" areas of the hemisphere are not drawn
  if ((X1[2]>0)||(PS.Projection==3)) proceed=true;
  else proceed=false;

  //Check that the point is within screen bounds
  if ((*x1<0)||(*x1>w_2*2)||(*y1<0)||(*y1>h_2*2)) proceed=false;

  free_vector(X1,0,2);
  return proceed;
}
//---------------------------------------------------------------------------
//Returns the sky coordinate (lambda,q) of a map coordinate (x,y)
void MapToSky(int *x1,int *y1,float *q,float *lambda)
{
  float arg,*X_;
  X_=vector(0,2);
  X_[0]=(*x1-w_2)/R0;
  X_[1]=(h_2-*y1)/R0;
  if (PS.Projection==0){
    X_[0]=-X_[0]*2;
    X_[1]=X_[1]*2;
  } else {
    if (PS.Projection!=1){
      X_[0]=X_[0]*2;
      X_[1]=X_[1]*2;
    }
  }
  arg=1-X_[0]*X_[0]-X_[1]*X_[1];
  if (arg>0) X_[2]=sqrt(arg);
  else X_[2]=1;

  //Get the rectangular coordinates prior to the projection transform
  if (PS.Projection==3) InverseMercator(X_);
  if (PS.Projection==2) InversePlanar(X_);
  if (PS.Projection==1) InverseLambert(X_);

  //Rotate about the x-axis by the negative central latitude
  arg=-sinq0;
  XRotate(X_,&arg,&cosq0);

  //Convert to spherical coordinates
  CosinesToSph(X_);
  *q=X_[1];
  *lambda=X_[2]+lambda_0+pihalved;  //Rotate about z-axis by the central longitude
  FirstCircle(lambda);

  free_vector(X_,0,2);
}
//---------------------------------------------------------------------------
//Draws the constellation figures
void __fastcall TMapF::DrawConstFig(void)
{
  Map->Canvas->Pen->Width=1;
  Map->Canvas->Pen->Color=clGray;

  float ra,dec;
  bool proceed;
  bool started;
  int index,x1,y1,lastx1=0,lasty1=0;

  int N=0, N2=0;
  for (int i=0;i<79;i++){
    for (int j=0;j<ConstFigNSeg[i];j++){
      index=0;
      started=false;
      for (int k=0;k<ConstFigNPts[N2];k++){
        ra=ConstFigRA[N];
        dec=ConstFigDec[N];
        proceed=SkyToMap(&dec,&ra,&x1,&y1);
        N++;

        if (proceed&&(abs(lastx1-x1)<PS.h)&&(abs(lasty1-y1)<PS.h)){
          started=true;
          //Add the point to the line
          points[index].x=x1;
          points[index].y=y1;
          index++;
        } else {
          if (started&&index>1) {
            //The line started but now has stopped, draw the
            //segment
            Map->Canvas->Polyline(points,index-1);
            //Reset the drawing index and tags so that a
            //new segment may be drawn (if a valid segment
            //comes up)
            started=false;
            index=0;
          }
        }
        lastx1=x1;
        lasty1=y1;
      }
      if (started&&index>1){
        if (proceed) Map->Canvas->Polyline(points,index-1);
        else Map->Canvas->Polyline(points,index);
      }
      N2++;
    }
  }
}
//---------------------------------------------------------------------------
//Redraw the map
void __fastcall TMapF::RedrawBtnClick(TObject *Sender)
{
  LoadSettings(Sender);
  DrawMap(Sender);
}
//---------------------------------------------------------------------------
//Draw the map legend
void __fastcall TMapF::DrawLegend(TObject *Sender, float sc)
{
  int h,w,x1,y1;
  AnsiString text;

  Legend->Assign(Map);
  Legend->Width=Map->Width;
  Legend->Height=Map->Width/7;
  h=Legend->Height;
  w=Legend->Width;

  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Pen->Width=w/200;
  Legend->Canvas->Rectangle(0,0,w,h);

  //Star symbols
  Map->Canvas->Font->Name="Arial";
  Legend->Canvas->Font->Height=h/6;
  text="Stellar magnitudes";
  Legend->Canvas->TextOut(w/30,h/30,text);
  Legend->Canvas->Font->Height=h/7;

  x1=w/20; y1=h/3;
  Legend->Canvas->Pen->Width=1;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-10*sc,y1-10*sc,x1+10*sc,y1+10*sc);
  text=" < 1.5";
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->TextOut(x1+12*sc,y1-h/11,text);

  y1=h/3+h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-10*sc,y1-10*sc,x1+10*sc,y1+10*sc);
  Legend->Canvas->Pen->Color=clWhite;
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->Ellipse(x1-5*sc,y1-5*sc,x1+5*sc,y1+5*sc);
  text=" 1.5 - 2.0";
  Legend->Canvas->TextOut(x1+12*sc,y1-h/11,text);

  y1=h/3+2*h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-8*sc,y1-8*sc,x1+8*sc,y1+8*sc);
  text=" 2.0 - 2.5";
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->TextOut(x1+12*sc,y1-h/11,text);

  y1=h/3+3*h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-8*sc,y1-8*sc,x1+8*sc,y1+8*sc);
  Legend->Canvas->Pen->Color=clWhite;
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->Ellipse(x1-4*sc,y1-4*sc,x1+4*sc,y1+4*sc);
  text=" 2.5 - 3.0";
  Legend->Canvas->TextOut(x1+12*sc,y1-h/11,text);

  x1=w/20+w/7; y1=h/3;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-7*sc,y1-7*sc,x1+7*sc,y1+7*sc);
  text=" 3.0 - 3.5";
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  y1=h/3+h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-7*sc,y1-7*sc,x1+7*sc,y1+7*sc);
  Legend->Canvas->Pen->Color=clWhite;
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->Ellipse(x1-3*sc,y1-3*sc,x1+3*sc,y1+3*sc);
  text=" 3.5 - 4.0";
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  y1=h/3+2*h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-6*sc,y1-6*sc,x1+6*sc,y1+6*sc);
  text=" 4.0 - 5.0";
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  y1=h/3+3*h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-6*sc,y1-6*sc,x1+6*sc,y1+6*sc);
  Legend->Canvas->Pen->Color=clWhite;
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->Ellipse(x1-3*sc,y1-3*sc,x1+3*sc,y1+3*sc);
  text=" 5.0 - 6.0";
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  x1=w/20+2*w/7; y1=h/3;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-5*sc,y1-5*sc,x1+5*sc,y1+5*sc);
  text=" 6.0 - 7.0";
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  y1=h/3+h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-4*sc,y1-4*sc,x1+4*sc,y1+4*sc);
  Legend->Canvas->Pen->Color=clWhite;
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->Ellipse(x1-2*sc,y1-2*sc,x1+2*sc,y1+2*sc);
  text=" 7.0 - 8.0";
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  y1=h/3+2*h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-2*sc,y1-2*sc,x1+2*sc,y1+2*sc);
  text=" 8.0 - 9.5";
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  y1=h/3+3*h/5.5;
  Legend->Canvas->Brush->Color=clBlack;
  Legend->Canvas->Pen->Color=clBlack;
  Legend->Canvas->Ellipse(x1-1*sc,y1-1*sc,x1+1*sc,y1+1*sc);
  text=" > 9.5";
  Legend->Canvas->Brush->Color=clWhite;
  Legend->Canvas->TextOut(x1+10*sc,y1-h/11,text);

  //Other symbols


}
//---------------------------------------------------------------------------
//Print the map
void __fastcall TMapF::Print1Click(TObject *Sender)
{
  int old_w, old_h, old_w2, old_h2, old_R0, sc;
  old_w=w_;
  old_h=h_;
  old_w2=w_2;
  old_h2=h_2;
  old_R0=R0;

  Map->Width=2000;
  Map->Height=Map->Width*2700.0/2250.0;
  w_=Map->Width;
  h_=Map->Height;
  w_2=Map->Width/2;
  h_2=Map->Height/2;
  R0=Map->Width*R0/old_w;
  sc=StarScale*Map->Width/old_w;

  //Clear the canvas
  Map->Canvas->Pen->Width=w_/150;
  Map->Canvas->Pen->Color=clWhite;
  Map->Canvas->Brush->Color=clWhite;
  Map->Canvas->Rectangle(0,0,w_,h_);
  if (R0<4*w_){
    if (PS.Projection==0){
      Map->Canvas->Ellipse(w_2-R0/2,h_2-R0/2,w_2+R0/2,h_2+R0/2);
    } else {if (PS.Projection==1){
      Map->Canvas->Ellipse(w_2-R0,h_2-R0,w_2+R0,h_2+R0);
    } else {if (PS.Projection==2){
      Map->Canvas->Rectangle(0,0,w_,h_);
    } else {
      Map->Canvas->Rectangle(w_2-R0*pihalved,0,w_2+R0*pihalved,h_);
    }}}
  }

  //Get the latitude and longitude range
  float min_lambda,max_lambda,min_q,max_q,lambda_incr,q_incr;
  lambda_incr=pi*RAGrid->Value/(12*60);
  q_incr=pi*DecGrid->Value/180;
  LongLatRange_(&min_lambda,&max_lambda,&min_q,&max_q,lambda_incr,q_incr);

  if (StarLabels) DrawStarLabels();
//  if (PS.DrawConstLab) DrawConstLab();
//  if (PS.DrawConstBound&&!OnlyStars) DrawConstBound();
  if (PS.DrawConstFig) DrawConstFig();
  if (PS.ShowHorLine) DrawIntersects(0);
  if (PS.ShowEclLine) DrawIntersects(1);
  if (PS.ShowGalLine) DrawIntersects(2);
  if (Stars1->Checked) ShowStars(sc);
//  if (PS.ShowPlanets&&!OnlyStars) DrawPlanets(PS.time.jul);
//  if (PS.ShowDeepSky&&!OnlyStars) DrawRNGC();
//  DrawHighlights();
  if (PS.ShowGrid) DrawGrid(&min_lambda,&max_lambda,&min_q,&max_q);

  Map->Canvas->Pen->Color=clBlack;
  Map->Canvas->Brush->Style=bsClear;
  Map->Canvas->Pen->Width=w_/200;
  Map->Canvas->Rectangle(0,0,w_,h_);
  Map->Canvas->Brush->Style=bsSolid;

  DrawLegend(Sender,sc);

  //Paint to the printer canvas
  Printer()->BeginDoc();
    TRect MapRect,LegendRect;
    MapRect = Rect(75,75,2325,2775);
    LegendRect = Rect(75,2780,2325,2778+2325*Legend->Height/Legend->Width);
    Printer()->Canvas->Brush->Color=clWhite;
    Printer()->Canvas->Rectangle(75,75,2325,3100);
    Printer()->Canvas->StretchDraw(MapRect,Map);
    Printer()->Canvas->StretchDraw(LegendRect,Legend);
  Printer()->EndDoc();

  w_=old_w;
  h_=old_h;
  w_2=old_w2;
  h_2=old_h2;
  R0=old_R0;
}
//---------------------------------------------------------------------------


