//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "NonStellSet.h"
#include "ServiceRoutines.h"
#include "Main.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "cspin"
#pragma link "CSPIN"
#pragma resource "*.dfm"
TNonStellarSet *NonStellarSet;

TRect Source,Dest;

extern Settings PS;
Settings BackupPS;  //Holds a backup copy of the original settings
Settings PS_;       //Holds modified settings prior to being copied to PS

extern bool drawing;
extern short **NumNGC;
extern short **type_ngc;
extern float **mag_ngc;
extern bool **show_ngc;
extern short **NumIC;
extern short **type_ic;
extern float **mag_ic;
extern bool **show_ic;
extern short *NumRC3;
extern float *mag_rc3;
extern bool *show_rc3;

extern Graphics::TBitmap *DeepSkySymb;
extern bool DisplayRNGCItem(int type,float mag);
extern bool DisplayICItem(int type,float mag);
extern int *i_vector(int nl,int nh);
extern void free_i_vector(int *v, int nl, int nh);
extern bool *b_vector(int nl,int nh);
extern void free_b_vector(bool *v, int nl, int nh);
extern void free_matrices(int database,int vector_size);
extern bool DrawingMap;
extern void LoadCatalog(int catalog,bool startflag);

String CatalogDesc[10]={"Revised New General Catalogue of nebulae and star clusters (8163 records).",
  "Index Catalogue of nebulae and star clusters.","Messier Catalogue of notable objects visible from northern mid latitudes (110 records).",
  "Lynd catalogue of dark nebulae.","Barnard's catalogue of dark nebulae.","Arp Catalogue of peculiar galaxies and associations.",
  "RC3 catalogue of galaxies; contains size information.","Zwicky catalogue of galaxy clusters",
  "Princeton pulsar catalogue.","Globular clusters compiled by Monela et al."};

  //---------------------------------------------------------------------------
__fastcall TNonStellarSet::TNonStellarSet(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
//Load form components on startup
void __fastcall TNonStellarSet::FormActivate(TObject* Sender)
{
  static bool startup=true;
  if (startup) {
    startup=false;
    CatList->ItemIndex=0;
    ObjList->ItemIndex=0;
  }

  //Keep a copy of the startup settings
  BackupPS=PS;
  PS_=PS;

  DrawForm(Sender);
}
//---------------------------------------------------------------------------
//Draw settings on the form
void __fastcall TNonStellarSet::DrawForm(TObject* Sender)
{
  if (PS_.UseCat[CatList->ItemIndex]) UseCatBut->Caption="YES";
  else UseCatBut->Caption="NO";

  CatDesc->Clear();
  CatDesc->Lines->Add(CatalogDesc[CatList->ItemIndex]);

  if (PS_.ShowObjects[ObjList->ItemIndex]) ObjVisBut->Caption="YES";
  else ObjVisBut->Caption="NO";

  if (PS_.DrawObjLab[ObjList->ItemIndex]) LabelBut->Caption="YES";
  else LabelBut->Caption="NO";

  MaxMag->Value=PS_.MaxMag[ObjList->ItemIndex];

  Dest=Rect(0,0,15,15);
  Source=Rect(PS.Symbol[ObjList->ItemIndex]*15,ObjList->ItemIndex*15,(PS.Symbol[ObjList->ItemIndex]+1)*15,(ObjList->ItemIndex+1)*15);
  Symb->Canvas->CopyRect(Dest,DeepSkySymb->Canvas,Source);
  Dest=Rect(0,0,150,15);
  Source=Rect(0,ObjList->ItemIndex*15,150,(ObjList->ItemIndex+1)*15);
  Image->Canvas->CopyRect(Dest,DeepSkySymb->Canvas,Source);
}
//---------------------------------------------------------------------------
//When the mouse is clicked the selected bitmap is displayed
//and the symbol index is changed
void __fastcall TNonStellarSet::ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
  int index=X/15;
  Dest=Rect(0,0,15,15);
  Source=Rect(index*15,ObjList->ItemIndex*15,(index+1)*15,(ObjList->ItemIndex+1)*15);
  Symb->Canvas->CopyRect(Dest,DeepSkySymb->Canvas,Source);
  PS_.Symbol[ObjList->ItemIndex]=index;
}
//---------------------------------------------------------------------------
void __fastcall TNonStellarSet::ObjVisButClick(TObject *Sender)
{
  PS_.ShowObjects[ObjList->ItemIndex]=!PS_.ShowObjects[ObjList->ItemIndex];
  if (PS_.ShowObjects[ObjList->ItemIndex]) ObjVisBut->Caption="YES";
  else ObjVisBut->Caption="NO";
}
//---------------------------------------------------------------------------
void __fastcall TNonStellarSet::LabelButClick(TObject *Sender)
{
  PS_.DrawObjLab[ObjList->ItemIndex]=!PS_.DrawObjLab[ObjList->ItemIndex];
  if (PS_.DrawObjLab[ObjList->ItemIndex]) LabelBut->Caption="YES";
  else LabelBut->Caption="NO";
}
//---------------------------------------------------------------------------
void __fastcall TNonStellarSet::MaxMagChange(TObject *Sender)
{
  PS_.MaxMag[ObjList->ItemIndex]=MaxMag->Value;
}
//---------------------------------------------------------------------------
void __fastcall TNonStellarSet::UseCatButClick(TObject *Sender)
{
  PS_.UseCat[CatList->ItemIndex]=!PS_.UseCat[CatList->ItemIndex];
  if (PS_.UseCat[CatList->ItemIndex]) UseCatBut->Caption="YES";
  else UseCatBut->Caption="NO";
}
//---------------------------------------------------------------------------
//Restore the original settings before entering the form
void __fastcall TNonStellarSet::Button4Click(TObject *Sender)
{
  PS=BackupPS;
  PS_=PS;

  //Redraw the form
  DrawForm(Sender);

  //Redraw the map
  if (!DrawingMap){
    DrawingMap=true;
    MainF->DrawMap(PS_.MaxStarMag,false);
    DrawingMap=false;
  }
}
//---------------------------------------------------------------------------
//Show how the changes will affect the map
void __fastcall TNonStellarSet::ApplyBtnClick(TObject *Sender)
{
  //Wait for the program to end drawing the sky map (if this is happening)
  //then pause the timer that causes the map to be redrawn
  while (drawing){};
  MainF->RedrawTimer->Enabled=false;

  //Free memory taken up by unused databases and load
  //any new databases
  for (int i=0;i<n_cat;i++){
    if (PS.UseCat[i]!=PS_.UseCat[i]){
      if (PS_.UseCat[i]){
        LoadCatalog(i,true);
      } else {
        free_matrices(i,0);
      }
    }
  }
  PS=PS_;

  //Repaint the map
  if (!DrawingMap){
    DrawingMap=true;
    MainF->DrawMap(PS.MaxStarMag,false);
    DrawingMap=false;
  }
  MainF->RedrawTimer->Enabled=true;
}
//---------------------------------------------------------------------------
void __fastcall TNonStellarSet::ObjListClick(TObject *Sender)
{
  static int LastListIndex;

  //Change the form settings if the list index has changed
  if (LastListIndex!=ObjList->ItemIndex){
    if (PS_.ShowObjects[ObjList->ItemIndex]) ObjVisBut->Caption="YES";
    else ObjVisBut->Caption="NO";

    if (PS_.DrawObjLab[ObjList->ItemIndex]) LabelBut->Caption="YES";
    else LabelBut->Caption="NO";

    MaxMag->Value=PS_.MaxMag[ObjList->ItemIndex];

    Dest=Rect(0,0,15,15);
    Source=Rect(PS_.Symbol[ObjList->ItemIndex]*15,ObjList->ItemIndex*15,(PS_.Symbol[ObjList->ItemIndex]+1)*15,(ObjList->ItemIndex+1)*15);
    Symb->Canvas->CopyRect(Dest,DeepSkySymb->Canvas,Source);
    Dest=Rect(0,0,150,15);
    Source=Rect(0,ObjList->ItemIndex*15,150,(ObjList->ItemIndex+1)*15);
    Image->Canvas->CopyRect(Dest,DeepSkySymb->Canvas,Source);
  }
  LastListIndex=ObjList->ItemIndex;
}
//---------------------------------------------------------------------------
void __fastcall TNonStellarSet::CatListClick(TObject *Sender)
{
  static int LastIndex;

  if (CatList->ItemIndex!=LastIndex){
    if (PS_.UseCat[CatList->ItemIndex]) UseCatBut->Caption="YES";
    else UseCatBut->Caption="NO";
    CatDesc->Clear();
    CatDesc->Lines->Add(CatalogDesc[CatList->ItemIndex]);
  }
  LastIndex=CatList->ItemIndex;
}
//---------------------------------------------------------------------------
void __fastcall TNonStellarSet::BitBtn1Click(TObject *Sender)
{
  ApplyBtnClick(Sender);
  Close();
}
//---------------------------------------------------------------------------

