//---------------------------------------------------------------------------
#include <vcl.h>
#include <math.h>
#include <fstream.h>
#include <stdio.h>
#pragma hdrstop

#include "ObjInfo.h"
#include "ServiceRoutines.h"
#include "Main.h"
#include "ImageForm.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "cspin"
#pragma link "CSPIN"
#pragma resource "*.dfm"
TObjectInfo *ObjectInfo;

extern double junk;
//local globals
float ra_=0,dec_=0,ra_pm=0,dec_pm=0,vismag_;
double dT__=0;
double ViewingTime; //The viewing time for the current object (defaults to PS.time.jul)
bool highlighted;
int CurrentObject;  //Note that index 0 is reserved for the case where none of the listed objects are selected.
int SourceCatalog;  //The catalog from which the search initially sourced the object information
String ObjType="";
String CommonName="";
bool AddToButEn;
bool ViewTimeButEn;
bool CenterViewOnButEn;
bool XMirrorOn=false;
bool YMirrorOn=false;
bool ImageExists;  //An image for the deep sky object exists
String ImageFile;  //The file name of the image for the deep sky object
int ImageIndex;    //The index of the image file in the "file list" file
float PScale;   //Scale factor for the planet image
float BasePScale;
int BaseIndex,NSat;  //Used for planet display
Graphics::TBitmap *TextureMap = new Graphics::TBitmap();
Graphics::TBitmap *PlBitmap = new Graphics::TBitmap();
Graphics::TBitmap *MoonBitmap = new Graphics::TBitmap();

//Variables used in searches
int BestMatches[51][4]; //BestMatch[i][0]=Object catalogue
                        //BestMatch[i][1]=Object number
                        //BestMatch[i][2]=Object extension number
                        //BestMatch[i][3]=Number of matching characters
String BestMatchDesc[51];

//External variables
extern Angles AL;
extern Settings PS;
extern double OElem[213][15];
extern OrbitalElements SSData[213];
extern int CatNum[20];
extern int ExtNum;  //The catalogue extension number of the selected object
extern int SelectedCatalogue;
extern int HighlightNum;
extern bool DrawingMap;
//Used for RC3 catalogue
extern float *ra_rc3;
extern float *dec_rc3;
extern float *a_rc3;  //major axis
extern float *b_rc3;  //minor axis
extern float *pa_rc3; //position angle
extern float *mag_rc3;
extern short *num_rc3;

//Local functions
String DreyerTranslate(String expr);
String RNGCTranslate(String expr);
int CompareString(String Str1, String Str2);
void AddToFoundList(int CatN,int N,int N_Ext,int MatchNChar,String Desc,int MaxNMatch);

//External functions
extern String GetRAString(float ra);
extern String GetDecString(float dec, bool HiPrecision);
extern void EquToHor(float *lambda,float *q,float *Az,float *h,float SidTime);
extern bool SkyToScreen(float *q,float *lambda,int *x1,int *y1);
extern SelectedObjects SO[100];
extern HighlightObjects HO[100];
extern void EclSatCoord(int n,double *T);
extern void GeoSatCoord(int n);
extern bool PlanetToGeoCoord(int *N,double *r,double *T,double *lambda,double *phi,double *dalpha,double *ddelta,bool Initialize);
extern void GetOE(double *T);
extern void GetOE_(int i, double *T);
extern void XRotate2(double *x,double *y,double *z,double *sinangle,double *cosangle);
extern void YRotate2(double *x,double *y,double *z,double *sinangle,double *cosangle);
extern void ZRotate2(double *x,double *y,double *z,double *sinangle,double *cosangle);
extern void FirstCircle(float *angle);
extern void FirstCircle2(double *angle);
extern void GetObliquity(double *T);
extern void EqToGal(float *lambda, float *q);
extern String GetRiseTime(float RA, float Dec, float SidTime, struct TIME_ CurTime);
extern String GetSetTime(float RA, float Dec, float SidTime, struct TIME_ CurTime);
extern void Illumination(int n, double *k, double *mag);
extern void SphToCosines2(double *X);
extern void Precess_sp(float *ra,float *dec,double *T, double *T0);
extern double *d_vector(int nl, int nh);
extern void free_d_vector(double *v, int nl, int nh);

//---------------------------------------------------------------------------
__fastcall TObjectInfo::TObjectInfo(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
//What to do when the form opens
void __fastcall TObjectInfo::FormActivate(TObject *Sender)
{
  PlBitmap->Assign(PImage->Picture->Bitmap);
  String filename = PS.ProgPath+"Planets\\Bitmaps\\Moon.bmp";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }
  MoonBitmap->LoadFromFile(filename);
  SearchIndicator->Caption="";
  PlanetRes->Position = PS.PlanetResolution;
  ImageExists=false;

  static initialize=true;
  if (initialize){
    FindExt_->ItemIndex=0;
    SearchCat_->ItemIndex=2;
    initialize=false;
  }

  SourceCatalog=SelectedCatalogue;

  if (SelectedCatalogue==-4) PageControl->ActivePage=FindPage;

  ShowObjectInfo(Sender);
}
//---------------------------------------------------------------------------
//Show an object's information
void __fastcall TObjectInfo::ShowObjectInfo(TObject *Sender)
{
  ImageExists=false;

  //Case for a deep sky object
  if (SelectedCatalogue>=0){
    PageControl->ActivePage=DeepSky;
    XCatBtn->Enabled=false;
    XCatBtn->Caption="";
    ShowAsIC->Enabled=false;
    ShowAsMes->Enabled=false;
    ShowAsRNGC->Enabled=false;
    Desc->Clear();

    //Case for RNGC catalogue
    if (SelectedCatalogue==0){
      ShowAsRNGC->Enabled=true;
      ShowRNGCInfo(Sender);
    }

    //Case for IC catalogue
    if (SelectedCatalogue==1){
      ShowAsIC->Enabled=true;
      ShowICInfo(Sender);
    }

    //Case for Messier catalogue
    if (SelectedCatalogue==2){
      ShowAsMes->Enabled=true;
      ShowMesInfo(Sender);
    }

    //Case for Lynd catalogue
    if (SelectedCatalogue==3){
      ShowLyndInfo(Sender);
      XCatBtn->Enabled=true;
      XCatBtn->Caption="Lynd Info";
    }

    //Case for Barnard catalogue
    if (SelectedCatalogue==4){
      ShowBarnInfo(Sender);
      XCatBtn->Enabled=true;
      XCatBtn->Caption="Barnard Info";
    }

    //Case for Arp catalogue
    if (SelectedCatalogue==5){
      ShowArpInfo(Sender);
      XCatBtn->Enabled=true;
      XCatBtn->Caption="Arp Info";
    }

    //Case for RC3 catalogue
    if (SelectedCatalogue==6){
      ShowRC3Info(Sender);
      XCatBtn->Enabled=true;
      XCatBtn->Caption="RC3 Info";
    }

    //See if an image file is available
    int NRec=0, j=0, N, N_Ext, CatIndex;
    short *sptr, sdum=0;
    char buf[129];
    sptr=&sdum;
    FILE *infile;
    bool NotFound=true;

    String filename=PS.ProgPath+"Images\\ImageFiles.dat";
    if (!FileExists(filename)) {
      String osbuf = "Could not open file " + filename;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
      return;
    }
    infile=fopen(filename.c_str(),"rb");
    fseek(infile,0,SEEK_SET);
    fread(sptr,2,1,infile);
    NRec=*sptr;

    while (j<NRec && NotFound){
      //Skip to the required catalogue number
      if (SelectedCatalogue<3){
        if (SelectedCatalogue==0){
          fread(sptr,2,1,infile);
          N=*sptr;
          fread(sptr,2,1,infile);
          N_Ext=*sptr;
          if (N==CatNum[SelectedCatalogue+2] && N_Ext==ExtNum){
            NotFound=false;
            fseek(infile,10,SEEK_CUR);
            fgets(buf,129,infile);
            filename=buf;
            ImageFile=PS.ProgPath+"Images\\"+filename.Trim();
          } else {
            fseek(infile,138,SEEK_CUR);
          }
        } else {
          fseek(infile,2+2*SelectedCatalogue,SEEK_CUR);
          fread(sptr,2,1,infile);
          N=*sptr;
          if (N==CatNum[SelectedCatalogue+2]){
            NotFound=false;
            fseek(infile,6+(2-SelectedCatalogue)*2,SEEK_CUR);
            fgets(buf,129,infile);
            filename=buf;
            ImageFile=PS.ProgPath+"Images\\"+filename.Trim();
          } else {
            fseek(infile,134+(2-SelectedCatalogue)*2,SEEK_CUR);
          }
        }
      } else {
        fseek(infile,8,SEEK_CUR);
        fread(sptr,2,1,infile);
        CatIndex=*sptr;
        fread(sptr,2,1,infile);
        N=*sptr;
        fread(sptr,2,1,infile);
        N_Ext=*sptr;
        if (N==CatNum[SelectedCatalogue+2] && N_Ext==ExtNum && CatIndex==SelectedCatalogue){
          NotFound=false;
          fgets(buf,129,infile);
          filename=buf;
          ImageFile=PS.ProgPath+"Images\\"+filename.Trim();
        } else {
          fseek(infile,128,SEEK_CUR);
        }
      }
      j++;
    }
    fclose(infile);
    ImageIndex=j;

    //Display the image
    DisplayImage(!NotFound,ImageFile);
  }

  //Case for a star
  if (SelectedCatalogue==-2){
    PageControl->ActivePage=StarPage;

    //Show the star information
    ShowStarInfo(Sender,CatNum[0]);
  }

  //Case for a planet
  PScale=1.0;
  //Set the base scale of planet image (in degrees)
  //BasePScale is horizontal field of view of the image
  //in degrees.  BaseIndex is the index at which the satelite
  //data for the planet found in OElem and SSData database begins.
  //NSat is the number of satelites defined for the planet in the
  //database.
  if (CatNum[1]==0) {BasePScale=3; BaseIndex=0; NSat=0;}
  if (CatNum[1]==1) {BasePScale=.015; BaseIndex=0; NSat=0;}
  if (CatNum[1]==2) {BasePScale=.05; BaseIndex=0; NSat=0;}
  if (CatNum[1]==3) {BasePScale=.025; BaseIndex=200; NSat=PS.NMarSat;}
  if (CatNum[1]==4) {BasePScale=.2; BaseIndex=209; NSat=PS.NJupSat;}
  if (CatNum[1]==5) {BasePScale=.05; BaseIndex=202; NSat=PS.NSatSat;}
  if (CatNum[1]==6) {BasePScale=.005; BaseIndex=0; NSat=0;}
  if (CatNum[1]==7) {BasePScale=.005; BaseIndex=0; NSat=0;}
  if (CatNum[1]==8) {BasePScale=.001; BaseIndex=0; NSat=0;}
  if (CatNum[1]==100) {BasePScale=1.2; BaseIndex=100; NSat=0;}
  if (CatNum[1]>=101) {BasePScale=.001; BaseIndex=0; NSat=0;}

  if (SelectedCatalogue==-1){
    PageControl->ActivePage=PlanetPage;
    ShowSatPath->Down=!PS.DrawPPath[CatNum[1]];

    //Show the planet information
    PlanetInfo(Sender);
  }

  //Check to see if the object is in the selected object list
  int i=0;
  while ((CatNum[SelectedCatalogue+2]!=SO[i].CatNum||ExtNum!=SO[i].CatExt||\
    SelectedCatalogue!=SO[i].CatIndex)&&(i<ObjList->Items->Count)){
    i++;
  }
  if (i<ObjList->Items->Count){
    //Object found, make the changes to the form
    CurrentObject=i+1;
    AddToListBut->Enabled=false;
    if (SO[i].Highlight) {
      HighlightBut->Caption="Remove Highlight";
      highlighted=true;
    } else {
      HighlightBut->Caption="Highlight";
      highlighted=false;
    }
    SelectBut->Enabled=false;
  } else {
    //Object not found, load form defaults
    AddToListBut->Enabled=true;
    HighlightBut->Caption="Highlight";
    highlighted=false;
  }
}
//---------------------------------------------------------------------------
//Dreyer code translator
String DreyerTranslate(String expr)
{
  int i,j,len,lbuf;
  bool found,proceed,firstone,sameflag,fixsame;
  String abrv,sbuf,ostring="",s,filename,tempostring;
  String lead1,lead2,lead3,follow1,follow2,follow3,follow4;
  char abuf[11],tbuf[81];

  FILE *transl;
  filename=PS.ProgPath+"Catalogs\\RNGC\\NGCSTR.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return "";
  }
  transl=fopen(filename.c_str(),"rb");

  expr=expr.Trim();
  len=expr.Length();
  firstone=true;
  sameflag=false;
  fixsame=false;
  i=1;
  while(i<=len){
    proceed=true;

    //Read in characters until a space or colon are found
    s=""; sbuf="";
    do {sbuf+=s; s=expr.SubString(i,1); i++;}
    while ((((s!=":")&&(s!=" "))||(sbuf==" "))&&(i<=len+1));
    sbuf=sbuf.Trim();
    lbuf=sbuf.Length();

    if (s==":") follow4=",";
    else follow4=" ";

    //Strip off terminating ";" and or "?" and or ")"
    follow3="";
    if ((sbuf.SubString(lbuf,1)==";")||(sbuf.SubString(lbuf,1)=="?")||(sbuf.SubString(lbuf,1)==")")){
      follow3=sbuf.SubString(lbuf,1);
      lbuf--;
      sbuf=sbuf.SubString(1,lbuf);
    }
    follow2="";
    if ((sbuf.SubString(lbuf,1)==";")||(sbuf.SubString(lbuf,1)=="?")||(sbuf.SubString(lbuf,1)==")")){
      follow2=sbuf.SubString(lbuf,1);
      lbuf--;
      sbuf=sbuf.SubString(1,lbuf);
    }
    follow1="";
    if ((sbuf.SubString(lbuf,1)==";")||(sbuf.SubString(lbuf,1)=="?")||(sbuf.SubString(lbuf,1)==")")){
      follow1=sbuf.SubString(lbuf,1);
      lbuf--;
      sbuf=sbuf.SubString(1,lbuf);
    }

    //Strip off leading ";" and or "?" and or "("
    lead1="";
    if ((sbuf.SubString(1,1)==";")||(sbuf.SubString(lbuf,1)=="?")||(sbuf.SubString(lbuf,1)=="(")){
      lead1=sbuf.SubString(1,1);
      lbuf--;
      sbuf=sbuf.SubString(2,lbuf);
    }
    lead2="";
    if ((sbuf.SubString(1,1)==";")||(sbuf.SubString(lbuf,1)=="?")||(sbuf.SubString(lbuf,1)=="(")){
      lead2=sbuf.SubString(1,1);
      lbuf--;
      sbuf=sbuf.SubString(2,lbuf);
    }
    lead3="";
    if ((sbuf.SubString(1,1)==";")||(sbuf.SubString(lbuf,1)=="?")||(sbuf.SubString(lbuf,1)=="(")){
      lead3=sbuf.SubString(1,1);
      lbuf--;
      sbuf=sbuf.SubString(2,lbuf);
    }

    //Now looks for the string "deg" at the end of the
    //sbuf string and replaces it with ""
    if (lbuf>=3){
      if (sbuf.SubString(lbuf-2,3)=="deg"){
        sbuf=sbuf.SubString(1,lbuf-3)+"";
        proceed=false;
      }
    }

    if ((sbuf!="")&&(proceed)){
      //Does a linear search until the desired record is found
      fseek(transl,0,SEEK_SET);
      j=0;
      found=false;
      while ((j<533)&&(!found)){
        fgets(abuf,11,transl);
        abrv=abuf;
        if (abrv.Trim()==sbuf) found=true;
        else fseek(transl,80,SEEK_CUR);
        j++;
      }
      if (found){
        fgets(tbuf,81,transl);
        sbuf=tbuf;
        sbuf=sbuf.Trim();
        if (sbuf=="same as (similar to)") sameflag=true;
      }
    }
    //Capitalize the first letter on the first entry
    if (firstone) {
      lbuf=sbuf.Length();
      sbuf=sbuf.SubString(1,1).UpperCase()+sbuf.SubString(2,lbuf-1);
      firstone=false;
    }
    if (!sameflag){
      if (fixsame){
        if (sbuf.SubString(1,1)=="M"||sbuf.SubString(1,1)=="IC"){
          sbuf="same object as "+sbuf;
        } else {
          sbuf="similar to "+sbuf;
        }
      }
      ostring+=lead1+lead2+lead3+sbuf+follow1+follow2+follow3+follow4;
      fixsame=false;
    } else {
      //Tries to deal with the ambiguity of the use the
      //"equal sign".  If the last entry is an IC # or
      //M #, then "=" means "same as" otherwise use
      //"similar to"
      fixsame=true;
      sameflag=false;
    }
  }
  return ostring;
}
//---------------------------------------------------------------------------
//RNGC code translator
String RNGCTranslate(String expr)
{
  int i,j,k,len,lbuf,range;
  bool found,proceed,firstone,foundcolon;
  String abrv,sbuf,ostring="",s,filename,tempostring;
  String lead1,lead2,lead3,follow1,follow2,follow3,follow4;
  char abuf[21],tbuf[65];

  String Fchar[42]={"=","-","&","*",".","+","0","1","2","3",\
              "4","5","6","7","8","9","A","B","C","D","E",\
              "F","G","H","I","J","K","L","M","N","O","P",\
              "Q","R","S","T","U","V","W","X","Y","Z"};

  int Fpos[42]={1,6,7,13,117,118,119,120,150,241,267,278,\
              286,289,293,293,295,351,527,646,815,864,940,\
              943,955,1029,1032,1071,1111,1124,1210,1245,\
              1330,1330,1347,1534,1555,1561,1655,1715,1715,\
              1715};

  FILE *transl;
  filename=PS.ProgPath+"Catalogs\\RNGC\\RNGCSTR.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return "";
  }  
  transl=fopen(filename.c_str(),"rb");

  expr=expr.Trim();
  len=expr.Length();
  firstone=true;
  foundcolon=false;
  i=1;
  while(i<=len){
    proceed=true;

    //Read in characters until a space or colon are found
    s=""; sbuf="";
    do {sbuf+=s; s=expr.SubString(i,1); i++;}
    while ((((s!=";")&&(s!=" "))||(sbuf==" "))&&(i<=len+1));
    if (s==";") foundcolon=true; //Flag used to distinguish "Sulentic" abbreviation from "spiral"
    sbuf=sbuf.Trim();
    lbuf=sbuf.Length();

    if (s==";") follow4=",";
    else follow4=" ";

    //Strip off terminating or ")" or "))"
    follow3="";
    if (sbuf.SubString(lbuf,1)==")"){
      follow3="?";
      lbuf--;
      sbuf=sbuf.SubString(1,lbuf);
    }
    follow2="";
    if (sbuf.SubString(lbuf,1)==")"){
      follow2="?";
      lbuf--;
      sbuf=sbuf.SubString(1,lbuf);
    }

    //Do a linear search until the desired record is found.
    if ((sbuf!="")&&(proceed)){
      //Use the first character of the record to seek to
      //the approximate location in the file
      k=0;
      found=false;
      while ((k<42)&&!found){
        if (sbuf.SubString(1,1)==Fchar[k]) found=true;
        k++;
      }
      fseek(transl,84*(Fpos[k-1]-1),SEEK_SET);

      found=false;
      if (k<1715) range=Fpos[k]-Fpos[k-1];
      else range=1;
      j=0;
      while ((j<range)&&!found){
        fgets(abuf,21,transl);
        abrv=abuf;
        if (abrv.Trim()==sbuf) found=true;
        else fseek(transl,64,SEEK_CUR);
        j++;
      }
      if (found){
        fgets(tbuf,65,transl);
        sbuf=tbuf;
        sbuf=sbuf.Trim();
      }
    }
    if (!found) sbuf=sbuf.LowerCase();

    //Tries to discriminate between "spiral" and "south"
    if ((sbuf=="spiral")&&(i>=4)){
      sbuf="south";
    }

    //Discriminates between "Sulentic" abreviation "S" and
    //"spiral" abreviation "S" as well as NF flag
    if ((i==len)&&((sbuf=="spiral")||(sbuf=="south"))&&!foundcolon){
      //If there is only one field in the descriptor
      //and one of the strings contained is S then
      //S indicates "J.W. Sulentic (visual interpretation from POSS)
      sbuf="J.W. Sulentic (visual interpretation from POSS)";
    }
    if ((i==len)&&!foundcolon){
      if (ostring.Trim()=="NF") ostring="Not found: ";
    }

    //Capitalize the first letter on the first entry
    if (firstone) {
      lbuf=sbuf.Length();
      sbuf=sbuf.SubString(1,1).UpperCase()+sbuf.SubString(2,lbuf-1);
      firstone=false;
    }
    ostring+=sbuf+follow2+follow3+follow4;
  }
  fclose(transl);
  return ostring;
}
//---------------------------------------------------------------------------
//Displays the IC object information
void __fastcall TObjectInfo::ShowICInfo(TObject *Sender)
{
  String sbuf,sbuf_,filename,typ,OldDesc,OldDescT,NewDesc,NewDescT,XReff;
  int initpos,pos,IC;
  short *sptr,sdum=0;
  short type;
  float *fptr,fdum=0.0;
  sptr=&sdum;
  fptr=&fdum;

  char obuf[51];
  char magbuf[2];
  char pnamebuf[121];
  String TypeS[17]={"Uncertain type or may not exist",\
   "Open cluster","Globular cluster","Diffuse nebula",\
   "Planetary nebula","Galaxy","Cluster associated with nebulosity",\
   "Non existent","Object in Large Magellanic Cloud",\
   "Object in Small Magellanic Cloud",\
   "Uncertain type or may not exist","Double star",\
   "Uncertain type or may not exist","Single star",\
   "Triple star","Asterism","Plate defect"};

  //See if there is a name associated with the catalogue number
  FILE *NAME;
  filename=PS.ProgPath+"Catalogs\\XRef\\Names.dat";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  NAME=fopen(filename.c_str(),"rb");
  fseek(NAME,6,SEEK_SET);
  CommonName="";
  bool found=false;
  int i=0;
  while ((i<193)&&!found) {
    fread(sptr,2,1,NAME);
    IC=*sptr;
    if (IC==CatNum[3]){
      found=true;
      fgets(pnamebuf,121,NAME);
      sbuf=pnamebuf;
      dsName->Caption="Common name(s): "+sbuf;
      CommonName=sbuf;
    } else {
      fseek(NAME,126,SEEK_CUR);
    }
    i++;
  }
  if (!found) dsName->Caption="";
  fclose(NAME);

  //See if there is an RNGC or Messier number associated with
  //the IC number
  FILE *XREF;
  filename=PS.ProgPath+"Catalogs\\IC\\ICXRef.dat";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  XREF=fopen(filename.c_str(),"rb");
  fseek(XREF,0,SEEK_SET);
  found=false;
  i=0;
  while ((i<132)&&!found) {
    fread(sptr,2,1,XREF);
    IC=*sptr;
    if (IC==CatNum[3]){
      found=true;
      fread(sptr,2,1,XREF);
      CatNum[2]=*sptr;
      fread(sptr,2,1,XREF);
      ExtNum=*sptr;
      fread(sptr,2,1,XREF);
      CatNum[4]=*sptr;
      dsName->Caption="Common name(s): "+sbuf;
    } else {
      fseek(XREF,6,SEEK_CUR);
    }
    i++;
  }
  if (CatNum[4]!=0) ShowAsMes->Enabled=true;
  if (CatNum[2]!=0) ShowAsRNGC->Enabled=true;

  FILE *IC_;
  filename=PS.ProgPath+"Catalogs\\IC\\IC_FULL.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  IC_=fopen(filename.c_str(),"rb");

  //Note that there are 65 bytes per record. You must provide
  //an n+1 byte buffer to read in an n byte char because of the
  //addition of an end character during the read. Records are
  //listed by catalogue number, so we can seek directly to
  //record location in the file.
  fseek(IC_,(CatNum[3]-1)*65,SEEK_SET);

  dsNum->Caption="IC Catalog Number: "+IntToStr(CatNum[3]);

  //Object type
  fread(sptr,2,1,IC_);
  type=*sptr;
  dsType->Caption="Object type: "+TypeS[type];
  ObjType=type;

  //RA
  fread(fptr,4,1,IC_);
  ra_=*fptr;

  //Declination
  fread(fptr,4,1,IC_);
  dec_=*fptr;

  //Proper motions
  ra_pm=0;
  dec_pm=0;

  //Magnitude
  fread(fptr,4,1,IC_);
  sbuf=FormatFloat("#0.0",*fptr);
  //...magnitude code
  fgets(magbuf,2,IC_);
  vismag_=*magbuf;

  //Show object astrometric information
  ShowInfo(Sender,PS.time);

  //Old descriptor
  fgets(obuf,51,IC_);
  OldDesc=obuf;
  fclose(IC_);
  Desc->Clear();
  ObjectInfo->Repaint();

  //Format the old descriptor (Dreyer's description)
  if (OldDesc.Trim()!=""){
    OldDescT=DreyerTranslate(OldDesc);
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelAttributes->Color=clBlue;
    Desc->Lines->Add("Dreyer's NGC Catalogue Description");
    initpos=Desc->SelStart;
    Desc->SelAttributes->Color=Desc->DefAttributes->Color;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    Desc->Lines->Add("Code: "+OldDesc);
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=5;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;

    initpos=Desc->SelStart;
    Desc->Lines->Add("Translation: "+OldDescT);
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=12;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;
    Desc->Lines->Add("");
    Desc->SelStart=1;
  }
}
//---------------------------------------------------------------------------
//Displays the Lynd object information
void __fastcall TObjectInfo::ShowLyndInfo(TObject *Sender)
{
  String sbuf,filename,XRef;
  short *sptr,sdum=0;
  short lyndnum,idnum,lyndnum2,opcty;
  float *fptr,fdum=0.0,area;
  sptr=&sdum;
  fptr=&fdum;

  char buf[41];
  String Opacity[6]={"very light","light",\
    "moderately light","moderately dark","dark","very dark"};

  FILE *infile;
  filename=PS.ProgPath+"Catalogs\\Lynd\\Lynd_FULL.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  infile=fopen(filename.c_str(),"rb");

  //Note that there are 60 bytes per record. You must provide
  //an n+1 byte buffer to read in an n byte char because of the
  //addition of an end character during the read.
  int nskipbytes=60*(CatNum[5]-15);
  if (nskipbytes<0) nskipbytes=0;
  fseek(infile,nskipbytes,SEEK_SET);

  //Find the catalogue record (at most 15 files need to be checked)
  lyndnum=0;
  while (lyndnum!=CatNum[5]){
    fread(sptr,2,1,infile);
    lyndnum=*sptr;
    if (lyndnum!=CatNum[5]) fseek(infile,58,SEEK_CUR);
  }
  dsNum->Caption="Lynd Catalog Number: "+IntToStr(lyndnum);

  //RA
  fread(fptr,4,1,infile);
  ra_=*fptr;

  //Declination
  fread(fptr,4,1,infile);
  dec_=*fptr;

  //Proper motions
  ra_pm=0;
  dec_pm=0;

  //Area
  fread(fptr,4,1,infile);
  area=*fptr;

  //Opacity
  fread(sptr,2,1,infile);
  opcty=*sptr;

  //ID#
  fread(sptr,2,1,infile);
  idnum=*sptr;

  //Revised Lynd#
  fread(sptr,2,1,infile);
  lyndnum2=*sptr;

  //X-reference
  fgets(buf,41,infile);
  XRef=buf;
  fclose(infile);
  Desc->Clear();

  //Show object astrometric information
  ShowInfo(Sender,PS.time);

  dsType->Caption="Object type: Dark Nebula";

  //Format the description
  int nchar=0;
  if (XRef.Trim()!=""){
    sbuf="Barnard catalogue cross references: "+XRef;
    Desc->Lines->Add(sbuf);
    Desc->SelStart=0;
    Desc->SelLength=35;
    Desc->SelAttributes->Color=clBlue;
    Desc->SelStart=36;
    nchar=sbuf.Length()+1;
    Desc->SelLength=nchar;
    Desc->SelAttributes->Color=clBlack;
  }

  sbuf="Opacity: "+Opacity[opcty-1];
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+9;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+10;
  if (nchar>0) nchar+=sbuf.Length()+2;
  else nchar+=sbuf.Length()+1;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Area (square degrees): "+FormatFloat("###0.000",area);
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+23;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+24;
  nchar+=sbuf.Length()+2;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Revised Lynd catalog number: "+IntToStr(lyndnum2);
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+29;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+30;
  nchar+=sbuf.Length()+2;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Lynd I.D. number: "+IntToStr(idnum);
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+18;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+19;
  nchar+=sbuf.Length()+2;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;
  Desc->SelStart=nchar;
  Desc->SelLength=0;
  Desc->SelStart=1;
}
//---------------------------------------------------------------------------
//Displays the Barnard cloud information
void __fastcall TObjectInfo::ShowBarnInfo(TObject *Sender)
{
  String sbuf,filename;
  short *sptr,sdum=0;
  short barnnum,barnext,xref;
  float *fptr,fdum=0.0;
  sptr=&sdum;
  fptr=&fdum;

  char buf[21];

  FILE *infile;
  filename=PS.ProgPath+"Catalogs\\Barnard\\Barnard.dat";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  infile=fopen(filename.c_str(),"rb");

  //Note that there are 34 bytes per record. You must provide
  //an n+1 byte buffer to read in an n byte char because of the
  //addition of an end character during the read.
  int nskipbytes=34*(CatNum[6]-22);
  if (nskipbytes<0) nskipbytes=0;
  fseek(infile,nskipbytes,SEEK_SET);

  //Find the catalogue record (at most 15 files need to be checked)
  barnnum=0; barnext=0;
  while (barnnum!=CatNum[6]||barnext!=ExtNum){
    fread(sptr,2,1,infile);
    barnnum=*sptr;
    fread(sptr,2,1,infile);
    barnext=*sptr;
    if (barnnum!=CatNum[6]) fseek(infile,30,SEEK_CUR);
  }
  if (barnext==0) dsNum->Caption="Barnard Catalog Number: "+IntToStr(barnnum);
  else  dsNum->Caption="Barnard Catalog Number: "+IntToStr(barnnum)+"A";

  //RA
  fread(fptr,4,1,infile);
  ra_=*fptr;

  //Declination
  fread(fptr,4,1,infile);
  dec_=*fptr;

  //Proper motions
  ra_pm=0;
  dec_pm=0;

  //Common name
  fgets(buf,21,infile);
  CommonName=buf;
  CommonName=CommonName.Trim();
  dsName->Caption="Common name(s): "+CommonName;

  //X-Reference
  fread(sptr,2,1,infile);
  xref=*sptr;

  fclose(infile);
  Desc->Clear();

  //Show object astrometric information
  vismag_=-100;
  ShowInfo(Sender,PS.time);

  dsType->Caption="Object type: Dark Nebula";

  //Format the description
  if (xref!=0){
    sbuf="Lynd catalogue cross reference: "+IntToStr(xref);
    Desc->Lines->Add(sbuf);
    Desc->SelStart=0;
    Desc->SelLength=32;
    Desc->SelAttributes->Color=clBlue;
    Desc->SelStart=33;
    Desc->SelLength=sbuf.Length()+1;
    Desc->SelAttributes->Color=clBlack;
    Desc->SelStart=128;
    Desc->SelLength=0;
    Desc->SelStart=1;
  }
}
//---------------------------------------------------------------------------
//Displays the Arp galaxy information
void __fastcall TObjectInfo::ShowArpInfo(TObject *Sender)
{
  String sbuf,filename,Descriptor[44];
  int desc1,desc2;
  short *sptr,sdum=0;
  float *fptr,fdum=0.0,size;
  sptr=&sdum;
  fptr=&fdum;

  char descbuf[65];
  char namebuf[31];

  FILE *infile;
  filename=PS.ProgPath+"Catalogs\\Arp\\Arp.dat";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  infile=fopen(filename.c_str(),"rb");

  //Note that there are 46 bytes per record. You must provide
  //an n+1 byte buffer to read in an n byte char because of the
  //addition of an end character during the read.
  fseek(infile,0,SEEK_SET);

  //Read in the descriptor strings
  for (int i=0;i<44;i++){
    fgets(descbuf,65,infile);
    sbuf=descbuf;
    Descriptor[i]=sbuf.Trim();
  }

  //Skip to the catalogue record
  fseek(infile,46*(CatNum[7]-1),SEEK_CUR);

  //Common name
  fgets(namebuf,31,infile);
  CommonName=namebuf;
  CommonName=CommonName.Trim();

  //See if a cross reference to NGC, Messier or IC exists
  FILE *NAME;
  filename=PS.ProgPath+"Catalogs\\Arp\\arp_xref.dat";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  NAME=fopen(filename.c_str(),"rb");
  fseek(NAME,0,SEEK_CUR);
  String com_name;
  char namebuf_[56];
  int i=0, num=0, ngc, ic, mes;
  while ((num!=CatNum[7])&&(i<216)){
    fread(sptr,2,1,NAME);
    num=*sptr;
    if (num==CatNum[7]){
      fread(sptr,2,1,NAME);
      CatNum[2]=*sptr;
      fread(sptr,2,1,NAME);
      ExtNum=*sptr;
      fread(sptr,2,1,NAME);
      CatNum[3]=*sptr;
      fread(sptr,2,1,NAME);
      CatNum[4]=*sptr;
      fgets(namebuf_,56,NAME);
      com_name=namebuf_;
      com_name=com_name.Trim();
    } else {
      fseek(NAME,63,SEEK_CUR);
    }
    i++;
  }
  fclose(NAME);
  if (CatNum[2]!=0) ShowAsRNGC->Enabled=true;
  if (CatNum[3]!=0) ShowAsIC->Enabled=true;
  if (CatNum[4]!=0) ShowAsMes->Enabled=true;

  if (com_name!="") CommonName=com_name;

  dsNum->Caption="Catalogue Number: Arp "+IntToStr(CatNum[7]);
  dsName->Caption="Common name(s): "+CommonName;

  //Size
  fread(fptr,4,1,infile);
  size=*fptr;

  //RA
  fread(fptr,4,1,infile);
  ra_=*fptr;

  //Declination
  fread(fptr,4,1,infile);
  dec_=*fptr;

  //Proper motions
  ra_pm=0;
  dec_pm=0;

  //First desc. index
  fread(sptr,2,1,infile);
  desc1=*sptr;

  //Second desc. index
  fread(sptr,2,1,infile);
  desc2=*sptr;

  fclose(infile);
  Desc->Clear();

  //Show object astrometric information
  vismag_=-100;
  ShowInfo(Sender,PS.time);

  dsType->Caption="Object type: Peculiar galaxy";

  //Format the description
  int nchar=0;
  sbuf="Arp's primary taxonomy: "+Descriptor[desc1];
  Desc->Lines->Add(sbuf);
  Desc->SelStart=0;
  Desc->SelLength=22;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=23;
  nchar=sbuf.Length()+1;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Arp's secondary taxonomy: "+Descriptor[desc2];
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+24;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+25;
  if (nchar>0) nchar+=sbuf.Length()+2;
  else nchar+=sbuf.Length()+1;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Size: "+FormatFloat("###0.0",size)+"\'";
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+5;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+7;
  nchar+=sbuf.Length()+2;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;
  Desc->SelStart=nchar;
  Desc->SelLength=0;
  Desc->SelStart=1;
}
//---------------------------------------------------------------------------
//Displays RC3 galaxy information
void __fastcall TObjectInfo::ShowRC3Info(TObject *Sender)
{
  String sbuf,filename,Descriptor[44],type,lum,hubble,majaxis,minaxis,pa;
  int i, ic_xref, ngc_xref, ic_xref_ext, ngc_xref_ext;
  short *sptr, sdum=0;
  sptr=&sdum;

  char namebuf[57];
  char typebuf[11];
  char hubblebuf[11];
  char lumbuf[11];

  FILE *infile,*xref_file;
  filename=PS.ProgPath+"Catalogs\\RC3\\RC3_NAMES.dat";
  infile=fopen(filename.c_str(),"rb");
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  filename=PS.ProgPath+"Catalogs\\RC3\\RC3_XREF.dat";
  xref_file=fopen(filename.c_str(),"rb");

  //Determine any cross-references to IC or RNGC objects
  fseek(xref_file,8*(CatNum[8]-1),SEEK_SET);
  fread(sptr,2,1,xref_file);
  ngc_xref=*sptr;
  fread(sptr,2,1,xref_file);
  ngc_xref_ext=*sptr;
  fread(sptr,2,1,xref_file);
  ic_xref=*sptr;
  fread(sptr,2,1,xref_file);
  ic_xref_ext=*sptr;
  fclose(xref_file);

  if (ic_xref>0){
    //IC catalogue entry exists
    ShowAsIC->Enabled=true;
    CatNum[3]=ic_xref;
    ExtNum=ic_xref_ext;
  }
  if (ngc_xref>0){
    //RNGC catalogue entry exists
    ShowAsRNGC->Enabled=true;
    CatNum[2]=ngc_xref;
    ExtNum=ngc_xref_ext;
  }

  //Note that there are 86 bytes per record. You must provide
  //an n+1 byte buffer to read in an n byte char because of the
  //addition of an end character during the read.
  fseek(infile,86*(CatNum[8]-1)-1,SEEK_SET);

  //Object names
  fgets(namebuf,57,infile);
  CommonName=namebuf;
  CommonName=CommonName.Trim();
  dsName->Caption="Other designations(s): "+CommonName;

  //Object type
  fgets(typebuf,11,infile);
  type=typebuf;

  //Hubble stage
  fgets(hubblebuf,11,infile);
  hubble=hubblebuf;

  //Luminosity class
  fgets(lumbuf,11,infile);
  lum=lumbuf;
  lum=lum.Trim();
  fclose(infile);

  //Recover the catalogue index (for the sorted database)
  i=0;
  do {i++;} while (num_rc3[i]!=CatNum[8]);
  dsNum->Caption="Catalogue number: RC3 "+IntToStr(CatNum[8]);

  ra_=ra_rc3[i];
  dec_=dec_rc3[i];
  vismag_=mag_rc3[i];
  majaxis=GetDecString(a_rc3[i],false);
  minaxis=GetDecString(b_rc3[i],false);
  pa=FormatFloat("##0.0",pa_rc3[i]*180/pi)+"";

  //Proper motions
  ra_pm=0;
  dec_pm=0;

  Desc->Clear();

  //Show object astrometric information
  ShowInfo(Sender,PS.time);

  dsType->Caption="Object type: galaxy";

  //Format the description
  int nchar=0;
  sbuf="Galaxy type: "+type;
  Desc->Lines->Add(sbuf);
  Desc->SelStart=0;
  Desc->SelLength=12;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=13;
  nchar=sbuf.Length()+1;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Hubble stage: "+hubble;
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+13;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+14;
  if (nchar>0) nchar+=sbuf.Length()+2;
  else nchar+=sbuf.Length()+1;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Luminosity class: "+lum;
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+17;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+18;
  if (nchar>0) nchar+=sbuf.Length()+2;
  else nchar+=sbuf.Length()+1;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Major axis diameter: "+majaxis;
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+20;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+21;
  nchar+=sbuf.Length()+2;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;
  Desc->SelStart=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Minor axis diameter: "+minaxis;
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+20;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+21;
  nchar+=sbuf.Length()+2;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;
  Desc->SelStart=nchar;
  Desc->SelAttributes->Color=clBlack;

  sbuf="Position angle: "+pa;
  Desc->Lines->Add(sbuf);
  Desc->SelStart=nchar;
  Desc->SelLength=nchar+15;
  Desc->SelAttributes->Color=clBlue;
  Desc->SelStart=nchar+16;
  nchar+=sbuf.Length()+2;
  Desc->SelLength=nchar;
  Desc->SelAttributes->Color=clBlack;
  Desc->SelStart=nchar;
  Desc->SelLength=0;
  Desc->SelStart=1;
}
//---------------------------------------------------------------------------
//Displays the RNGC object information
void __fastcall TObjectInfo::ShowRNGCInfo(TObject *Sender)
{
  String sbuf,sbuf_,filename,typ,OldDesc,OldDescT,NewDesc,NewDescT,XReff;
  int type1,type2,initpos,pos,num;
  short *sptr,sdum=0;
  short type,ext;
  float *fptr,fdum=0.0;
  sptr=&sdum;
  fptr=&fdum;

  char obuf[51];
  char nbuf[57];
  char rbuf[33];
  char pnamebuf[121];
  String ExtS[8]={"","A","B","C","D","E","F","G"};
  String TypeS[11]={"Unverified southern object","Open cluster","Globular cluster",\
   "Diffuse nebula","Planetary nebula","Galaxy","Cluster associated with nebulosity",\
   "Non existent","Object in Large Magellanic Cloud",\
   "Object in Small Magellanic Cloud",""};

  //See if there is a name associated with the catalogue number
  FILE *NAME;
  filename=PS.ProgPath+"Catalogs\\XRef\\Names.dat";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  NAME=fopen(filename.c_str(),"rb");
  fseek(NAME,0,SEEK_SET);
  bool found=false;
  CommonName="";
  int i=0,NGC;
  while ((i<193)&&!found) {
    fread(sptr,2,1,NAME);
    NGC=*sptr;
    fread(sptr,2,1,NAME);
    ext=*sptr;
    if ((NGC==CatNum[2])&&(ext==ExtNum)){
      found=true;
      fseek(NAME,4,SEEK_CUR);
      fgets(pnamebuf,121,NAME);
      sbuf=pnamebuf;
      dsName->Caption="Common name(s): "+sbuf;
      CommonName=sbuf;
    } else {
      fseek(NAME,124,SEEK_CUR);
    }
    i++;
  }
  if (!found) dsName->Caption="";
  fclose(NAME);

  FILE *RNGC_;
  filename=PS.ProgPath+"Catalogs\\RNGC\\RNGC_FULL.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  RNGC_=fopen(filename.c_str(),"rb");

  //Note that there are 164 bytes per record. You must provide
  //an n+1 byte buffer to read in an n byte char because of the
  //addition of an end character during the read.  We'll skip
  //a few records (to be safe we'll always undershoot the
  //record number by 10%, this is because there are 4% more
  //RNGC numbers than NGC numbers (new RNGC have a letter
  //extensions in the number eg. 85A)
  num=CatNum[2]*0.9;
  fseek(RNGC_,num*164,SEEK_SET);

  //Read in the catalogue numbers and extensions until you find
  //one that matches
  ext=0;
  int nrec=num;
  while (((ext!=ExtNum)||(num!=CatNum[2]))&&(nrec<8163)){
    if (nrec>=num) fseek(RNGC_,160,SEEK_CUR);
    //Cat. number
    fread(sptr,2,1,RNGC_);
    num=*sptr;
    //Cat. extension
    fread(sptr,2,1,RNGC_);
    ext=*sptr;
    nrec++;
  }
  dsNum->Caption="RNGC Catalog Number: "+IntToStr(num)+ExtS[ext];

  //Object type
  fread(sptr,2,1,RNGC_);
  type=*sptr;
  typ=IntToStr(type);
  type1=(typ.SubString(1,1)).ToInt();
  if (typ.Length()==2) type2=(typ.SubString(2,1)).ToInt();
  else type2=10;
  if (type2==5) {
    sbuf=TypeS[type2];
  } else {
    if (TypeS[type2]!=""){
      sbuf=TypeS[type1]+", "+TypeS[type2];
    } else {
      sbuf=TypeS[type1];
    }
  }
  dsType->Caption="Object type: "+sbuf;
  ObjType=sbuf;

  //RA
  fread(fptr,4,1,RNGC_);
  ra_=*fptr;
  sbuf=GetRAString(ra_);
  dsRA->Caption="Right ascension: "+sbuf;

  //Declination
  fread(fptr,4,1,RNGC_);
  dec_=*fptr;
  sbuf=GetDecString(dec_,false);
  dsDec->Caption="Declination: "+sbuf;

  //Proper motions
  ra_pm=0;
  dec_pm=0;

  //Galactic longitude
  fread(fptr,4,1,RNGC_);
  fdum=*fptr;

  //Galactic Declination
  fread(fptr,4,1,RNGC_);
  fdum=*fptr;

  //Magnitude
  fread(fptr,4,1,RNGC_);
  vismag_=*fptr;

  //Show object astrometric information
  ShowInfo(Sender,PS.time);

  //Old descriptor
  fgets(obuf,51,RNGC_);
  OldDesc=obuf;

  //New descriptor
  fgets(nbuf,57,RNGC_);
  NewDesc=nbuf;
  NewDesc=NewDesc.Trim();

  //Cross reference
  fgets(rbuf,33,RNGC_);
  XReff=rbuf;
  XReff=XReff.Trim();
  fclose(RNGC_);
  Desc->Clear();
  ObjectInfo->Repaint();

  //Format the old descriptor (Dreyer's description)
  if (OldDesc.Trim()!=""){
    OldDescT=DreyerTranslate(OldDesc);
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelAttributes->Color=clBlue;
    Desc->Lines->Add("Dreyer's NGC Catalogue Description");
    initpos=Desc->SelStart;
    Desc->SelAttributes->Color=Desc->DefAttributes->Color;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    Desc->Lines->Add("Code: "+OldDesc);
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=5;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;

    initpos=Desc->SelStart;
    Desc->Lines->Add("Translation: "+OldDescT);
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=12;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;
    Desc->Lines->Add("");
  }

  //Format the new descriptor (Palomar Sky Survey description)
  if (NewDesc.Trim()!=""){
    NewDescT=RNGCTranslate(NewDesc);
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelAttributes->Color=clBlue;
    Desc->Lines->Add("Palomar Sky Survey Description");

    initpos=Desc->SelStart;
    Desc->SelAttributes->Color=Desc->DefAttributes->Color;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    Desc->Lines->Add("Code: "+NewDesc);
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=5;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;

    initpos=Desc->SelStart;
    Desc->Lines->Add("Translation: "+NewDescT);
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=12;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;
  }

  //Format the cross reference string
  if (XReff!=""){
    String OutStrings[10];
    String XRef[33]={"Arp, 1966, Atlas of Peculiar Galaxies",\
     "Arp, 1965, Catalogue of Globular Clusters",\
     "Hodge 1961, Blue LMC Clusters",\
     "de Vaucouleurs and de Vaucouleurs 1964, Reference Catalogue of Bright Galaxies",\
     "de Vaucouleurs and de Vaucouleurs 1964, Reference Catalogue of Bright Galaxies (with supplemental data)",\
     "Gum 1955, Catalogue of Southern HII Regions",\
     "Henize 1956, Catalogue of Nebulae in the Magellanic Clouds",\
     "Hogg 1959, Catalogue of Open and Globular Clusters",\
     "Holmberg 1931, Catalogues of Double and Multiple Galaxies",\
     "Dreyer 1895, 1908, Index Catalogues",\
     "Kron 1956, Catalogue of SMC Clusters",\
     "Linsay 1956, Catalogue of SMC Clusters",\
     "Hodge and Lucke 1970, Catalogue of LMC Associations",\
     "Lynds 1965, Catalogue of Bright Nebulae",\
     "Genkin and Genkina 1969, Catalogue of Galaxies with Determined Masses",\
     "Becvar 1959, Messier Catalogue",\
     "Perek and Kohoutek 1967, Catalogue of Galactic Planetary Nebulae",\
     "Bolton et al. 1964, Price and Milne 1965, Day et al. 1966, Shimins et al. 1966, Shimmins and Day 1968, Parkes Catalogues of Southern Radio Sources",\
     "Alter et al. 1970, Catalogue of Star Clusters and Associations",\
     "Minkowski 1968, Catalogue of Seyfert Galaxies",\
     "Karpowicz and Rudnicki 1968, Catalogue of Supernovae",\
     "Sharpless 1956, Catalogue of HII Regions",\
     "Shapley and Lindsay 1963, Catalogue of Clusters in the LMC",\
     "Vorontsov-Velyaminov, Krasnogorskaya and Arhipova 1962-1968, Morphological Catalogue of Galaxies",\
     "Vorontsov-Velyaminov, Krasnogorskaya and Arhipova 1962-1968, Morphological Catalogue of Galaxies (ID questioned by authors)",\
     "Vorontsov-Velyaminov, Krasnogorskaya and Arhipova 1962-1968, Morphological Catalogue of Galaxies (ID questioned by authors)",\
     "Zwicky et al. 1961-1968, Catalogue of Galaxies and of Clusters of Galaxies",\
     "Zwicky et al. 1961-1968, Catalogue of Galaxies and of Clusters of Galaxies (with additional comments)",\
     "Edge et al. 1959, Third Cambridge Catalogue of Radio Sources",\
     "Pilkington and Scott 1965; Gower et al. 1967, Fourth Cambridge Catalogue of Radio Sources",\
     "Object has supplementary data for RNGC (new) description",\
     "Object has supplementary data for Dreyer (old) description",\
     "Object has supplementary data for cross references"};

    String XRefCode[33]={"A###","AG","BMC","D","D*","G","H","HO",\
     "HO###A","IC####","K","L","LH","LY","M","M###","P","PKS","R",\
     "S","S#","SH","SL","V","V*","(V)","Z","Z*","3C###","4C##.##",\
     "8","9","X"};

    String firstc,secondc,first2c,c,sbuf;
    int len,j;

    //Separate out the different entries
    int NOutStrings=0;
    len=XReff.Length();
    j=1;
    while (j<=len){
      c=""; sbuf="";
      do {sbuf+=c; c=XReff.SubString(j,1); j++;} while((c!=";")&&(c!=" ")&&(j<=len+1));
      OutStrings[NOutStrings]=sbuf.Trim();
      NOutStrings++;
    }

    //Translate the codes
    for (int i=0;i<NOutStrings;i++){
      firstc=OutStrings[i].SubString(1,1);
      //Special cases
      //Arp catalogue
      if (firstc=="A"){
        secondc=XReff.SubString(2,1);
        if (secondc!="G"){
          OutStrings[i]=XRef[0]+": "+OutStrings[i];
        } else {
          OutStrings[i]=XRef[1];
        }
      } else {
        //Holmberg catalogue
        first2c=OutStrings[i].SubString(1,2);
        len=OutStrings[i].Length();
        //Remove ")" character left on the end of some strings
        if (OutStrings[i].SubString(len,1)==")") {OutStrings[i]=OutStrings[i].SubString(1,len-1); len--;}
        if ((first2c=="HO")&&(len>2)){
          OutStrings[i]=XRef[8]+": "+OutStrings[i];
        } else {
          //IC catalogue
          if (first2c=="IC"){
            CatNum[3]=OutStrings[i].SubString(3,len-2).ToInt();
            OutStrings[i]=XRef[9]+": IC"+IntToStr(CatNum[3]);
          } else {
            //Messier catalogue
            if (firstc=="M"){
              if (len>1){
                CatNum[4]=OutStrings[i].SubString(2,len-1).ToInt();
                OutStrings[i]=XRef[15]+": M"+IntToStr(CatNum[4]);
              } else {
                OutStrings[i]=XRef[14];
              }
            } else {
              //Supernova catalogue
              if ((firstc=="S")&&(len>1)){
                secondc=XReff.SubString(2,1);
                if ((secondc!="H")&&(secondc!="L")){
                  OutStrings[i]=XRef[20]+": "+OutStrings[i];
                } else {
                  if (secondc=="H") OutStrings[i]=XRef[21];
                  else OutStrings[i]=XRef[22];
                }
              } else {
                //Cambridge catalogue of radio sources
                if (first2c=="3C"){
                  OutStrings[i]=XRef[28]+": "+OutStrings[i];
                } else {
                  //4th Cambridge catalogue of radio sources
                  if (first2c=="4C"){
                    OutStrings[i]=XRef[28]+": "+OutStrings[i];
                  } else {
                    //All other cases
                    j=0;
                    while ((OutStrings[i]!=XRefCode[j])&&(j<33)) {j++;}
                    OutStrings[i]=XRef[j];
                  }
                }
              }
            }
          }
        }
      }
    }
    if (CatNum[3]!=0) ShowAsIC->Enabled=true;
    if (CatNum[4]!=0) ShowAsMes->Enabled=true;

    //Display results
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelAttributes->Color=clBlue;
    Desc->Lines->Add("");
    Desc->Lines->Add("Cross References");

    initpos=Desc->SelStart;
    Desc->SelAttributes->Color=Desc->DefAttributes->Color;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    Desc->Lines->Add("Code: "+XReff);
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=5;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;

    initpos=Desc->SelStart;
    Desc->Lines->Add("Translation: "+OutStrings[0]);
    for (int i=1;i<NOutStrings;i++){
      Desc->Lines->Add(OutStrings[i]);
    }
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
    pos=Desc->SelStart;
    Desc->SelStart=initpos;
    Desc->SelLength=12;
    Desc->SelAttributes->Color=clGreen;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelStart=pos;

    //Enable cross-reference buttons
    if (CatNum[3]!=0) ShowAsIC->Enabled=true;
    if (CatNum[4]!=0) ShowAsMes->Enabled=true;
  }
  Desc->SelStart=1;
}
//---------------------------------------------------------------------------
//Displays the Messier object information
void __fastcall TObjectInfo::ShowMesInfo(TObject *Sender)
{
  String sbuf,sbuf_,filename,typ,OldDesc,OldDescT,NewDesc,NewDescT,XReff;
  int Messier;
  short *sptr,sdum=0;
  short type;
  float *fptr,fdum=0.0;
  sptr=&sdum;
  fptr=&fdum;

  char obuf[56];
  char pnamebuf[121];

  String TypeS[6]={"Uncertain type or may not exist","Open cluster",\
    "Globular cluster","Diffuse nebula","Planetary nebula","Galaxy"};

  //See if there is a name associated with the catalogue number
  FILE *NAME;
  filename=PS.ProgPath+"Catalogs\\XRef\\Names.dat";
  NAME=fopen(filename.c_str(),"rb");
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  fseek(NAME,4,SEEK_SET);
  bool found=false;
  CommonName="";
  int i=0;
  while ((i<193)&&!found) {
    fread(sptr,2,1,NAME);
    Messier=*sptr;
    if (Messier==CatNum[4]){
      found=true;
      fseek(NAME,2,SEEK_CUR);
      fgets(pnamebuf,121,NAME);
      sbuf=pnamebuf;
      dsName->Caption="Common name(s): "+sbuf;
      CommonName=sbuf;
    } else {
      fseek(NAME,126,SEEK_CUR);
    }
    i++;
  }
  if (!found) dsName->Caption="";
  fclose(NAME);

  FILE *MESSIER;
  filename=PS.ProgPath+"Catalogs\\Messier\\Messier.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  MESSIER=fopen(filename.c_str(),"rb");

  //Note that there are 71 bytes per record. You must provide
  //an n+1 byte buffer to read in an n byte char because of the
  //addition of an end character during the read. Records are
  //listed by catalogue number, so we can seek directly to
  //record location in the file.
  fseek(MESSIER,(CatNum[4]-1)*71,SEEK_SET);

  dsNum->Caption="Messier Catalog Number: "+IntToStr(CatNum[4]);

  //NGC X-reference
  fread(sptr,2,1,MESSIER);
  CatNum[2]=*sptr;
  if (CatNum[2]!=0){
    ShowAsRNGC->Enabled=true;
  }

  //Object type
  fread(sptr,2,1,MESSIER);
  type=*sptr;
  dsType->Caption="Object type: "+TypeS[type];
  ObjType=type;

  //Get ra and dec
  fread(fptr,4,1,MESSIER);
  ra_=*fptr;
  fread(fptr,4,1,MESSIER);
  dec_=*fptr;

  //Proper motions
  ra_pm=0;
  dec_pm=0;

  //Magnitude
  fread(fptr,4,1,MESSIER);
  vismag_=*fptr;

  //Show object astrometric information
  ShowInfo(Sender,PS.time);

  //Description
  fgets(obuf,56,MESSIER);
  OldDesc=obuf;
  fclose(MESSIER);
  Desc->Clear();
  ObjectInfo->Repaint();

  //Format the descriptor
  if (OldDesc.Trim()!=""){
    Desc->SelAttributes->Style=Desc->DefAttributes->Style<<fsBold;
    Desc->SelAttributes->Color=clBlue;
    Desc->Lines->Add("Description: ");
    Desc->SelAttributes->Color=Desc->DefAttributes->Color;
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;

    Desc->Lines->Add(OldDesc);
    Desc->SelAttributes->Style=Desc->DefAttributes->Style>>fsBold;
  }
}
//---------------------------------------------------------------------------
//Reads a record from the SAO catalogue and displays the results.
void __fastcall TObjectInfo::ShowStarInfo(TObject *Sender,int CatNum)
{
  String sbuf,filename,fstring,Flamsteed,ComName;
  int *iptr,idum=0,SAO,SAO_,i;
  float *fptr,fdum=0.0;
  char buf[3],bufFlam[10],bufName[24];
  bool found;
  iptr=&idum;
  fptr=&fdum;

  FILE *mycat,*namefile;
  filename=PS.ProgPath+"Catalogs\\SAO\\SAO_FULL.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  mycat=fopen(filename.c_str(),"rb");

  //Skip to the desired record number.  Seeks relative to the
  //file beginning.
  //Note that there are 31 bytes per record (you must provide 4 byte
  //buffer to read in the 3 byte char spectral type value because of the
  //appended end character)
  fseek(mycat,(CatNum-1)*31,SEEK_SET);

  //Cat. number
  fread(iptr,4,1,mycat);
  SAO=*iptr;
  sbuf=IntToStr(SAO);
  stNum->Caption="SAO catalogue number: "+sbuf;

  //Get the Flamsteed label and common name if applicable
  filename=PS.ProgPath+"Catalogs\\SAO\\NAME_CROSSREF.DAT";
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }  
  namefile=fopen(filename.c_str(),"rb");
  i=0;
  found=false;
  CommonName="";
  while ((i<3126)&&(!found)){
    fseek(namefile,38*i,SEEK_SET);
    //Get the SAO number
    fread(iptr,4,1,namefile);
    SAO_=*iptr;
    if (SAO_==SAO) found=true;
    i++;
  }
  if (found){
    //Get the Flamsteed label
    fgets(bufFlam,11,namefile);
    Flamsteed=bufFlam;
    //Get the common name
    fgets(bufName,25,namefile);
    ComName=bufName;
    ComName=ComName.Trim();
    //Call the formating and display routine
    FormatName(Sender,Flamsteed,ComName,true);
  } else {
    Pre->Caption="";
    Greek->Caption="";
    Suf->Caption="";
  }

  fclose(namefile);

  //RA
  fread(fptr,4,1,mycat);
  ra_=*fptr;

  //RA Proper motion
  fread(fptr,4,1,mycat);
  stRApm->Caption="Right ascension proper motion: "+FormatFloat("0.0000",*fptr)+"\"/y";
  ra_pm=*fptr;

  //Declination
  fread(fptr,4,1,mycat);
  dec_=*fptr;

  //Declination proper motion
  fread(fptr,4,1,mycat);
  stDECpm->Caption="Declination proper motion:        "+FormatFloat("#0.0000##",*fptr)+"\""+"/y";
  dec_pm=*fptr;

  //Visual magnitude
  fread(iptr,4,1,mycat);
  vismag_=*iptr/100.0;

  //Photographic magnitude
  fread(iptr,4,1,mycat);

  //Spectral type
  fgets(buf,4,mycat);
  sbuf=buf;
  stType->Caption="Spectral type: "+sbuf;
  fclose(mycat);
  ObjType=sbuf+" star";

  //Show object astrometric information
  ShowInfo(Sender,PS.time);
}
//---------------------------------------------------------------------------
//Format and display a string containing Greek letters and
//constellation abbreviations (possessive latin form, i.e.
//as used in star names)
AnsiString __fastcall TObjectInfo::FormatName(TObject *Sender,\
  String Flamsteed,String ComName,bool Display)
{
  String GLetter[24]={"Alp","Bet","Gam","Del","Eps","Zet",\
    "Eta","The","Iot","Kap","Lam","Mu ","Nu ","Xi ","Omi",\
    "Pi ","Rho","Sig","Tau","Ups","Phi","Chi","Psi","Ome"};
  String GFLetter[24]={"Alpha","Beta","Gamma","Delta",\
    "Epsilon","Zeta","Eta","Theta","Iota","Kappa","Lambda",\
    "Mu","Nu","Xi","Omicron","Pi","Rho","Sigma","Tau",\
    "Upsilon","Phi","Chi","Psi","Omega"};
  String Equival[24]={"a","b","g","d","e","z","h","q","i",\
    "k","l","m","n","x","o","p","r","s","t","u","f","c",\
    "y","w"};
  String ConstAbr[88]={"And","Ant","Aps","Aqr","Aql","Ara",\
    "Ari","Aur","Boo","Cae","Cam","Cnc","CVn","CMa","CMi",\
    "Cap","Car","Cas","Cen","Cep","Cet","Cha","Cir","Col",\
    "Com","CrA","CrB","Crv","Crt","Cru","Cyg","Dlp","Dor",\
    "Dra","Equ","Eri","For","Gem","Gru","Her","Hor","Hya",\
    "Hyi","Ind","Lac","Leo","LMi","Lep","Lib","Lup","Lyn",\
    "Lyr","Men","Mic","Mon","Mus","Nor","Oct","Oph","Ori",\
    "Pav","Peg","Per","Phe","Pic","Psc","PsA","Pup","Pyx",\
    "Ret","Sge","Sgr","Sco","Scl","Sct","Ser","Sex","Trs",\
    "Tel","Tri","TrA","Tuc","UMa","UMi","Vel","Vir","Vol","Vul"};

  String ConstNam[88]={"Andromedae","Ant","Apodis","Aquarii","Aquilae",\
    "Arae","Arietis","Aurigae","Botis","Caeli","Camelopardi","Cancri",\
    "Canum Venaticorum","Canis Majoris","Canis Minoris","Capricorni",\
    "Carinae","Cassiopeiae","Centauri","Cephei","Ceti","Chamaeleontis",\
    "Circini","Colombae","Comae Berenices","Coronae Australis",\
    "Coronae Borealis","Corvi","Crateris","Crucis","Cygni","Delphini",\
    "Doradus","Draconis","Equulei","Eridani","Fornacis","Geminorum",\
    "Gruis","Herculis","Horologium","Hydrae","Hydri","Indi","Lacertae",\
    "Leonis","Leonis Minoris","Leporis","Librae","Lupi","Lyncis","Lyrae",\
    "Mensae","Microscopii","Monocerotis","Muscae","Normae","Octantis",\
    "Ophiuchii","Orionis","Pavonis","Pegasi","Persei","Phoenicis",\
    "Pictoris","Piscium","Piscis Austrini","Puppis","Pyxidis","Reticuli",\
    "Sagittae","Sagittarii","Scorpii","Sculptoris","Scuti","Serpentis",\
    "Sextantis","Tauri","Telescopii","Trianguli","Trianguli Australis",\
    "Tucanae","Ursae Majoris","Ursae Minoris","Velorum","Virginis",\
    "Volantis","Vulpeculae"};

  int j;
  String SrcStr,prefix,suffix,grk,multinum,abrev;
  bool found;

  //Get the greek letter (if any)
  grk=Flamsteed.SubString(4,3);
  found=false;
  j=-1;
  do {
    j++;
    if (GLetter[j]==grk) found=true;
  } while ((j<24)&&(!found));

  //Assign the appropriate character code
  if (found) {grk=Equival[j]; CommonName=GFLetter[j];}
  else {grk=""; CommonName="";}
  Greek->Caption=grk;

  //The name prefix (the Flamsteed number)
  prefix=(Flamsteed.SubString(1,3)).Trim();
  CommonName=prefix+" "+CommonName;

  //The multiple star label
  multinum=Flamsteed.SubString(7,1);
  if (multinum!=" ") multinum=" "+multinum;
  else multinum="";
  CommonName+=multinum;

  //Get the full constellation name from the constellation
  //abreviation
  abrev=Flamsteed.SubString(8,3);
  j=-1;
  found=false;
  do {
    j++;
    if (ConstAbr[j]==abrev) found=true;
  } while ((j<88)&&(!found));

  //Assign the constellation name
  if (found) suffix=multinum+" "+ConstNam[j]+", "+ComName;
  else suffix=multinum+", "+ComName;
  CommonName+=suffix;

  if (Display){
    Pre->Caption="Names: "+prefix;
    Greek->Left=Pre->Left+Pre->Width+2;
    Suf->Left=Greek->Left+Greek->Width;
    Suf->Caption=suffix;
  } else {
    return prefix+suffix;
  }
}
//---------------------------------------------------------------------------
//Display object information that may require updating with time
void __fastcall TObjectInfo::ShowInfo(TObject *Sender, struct TIME_ curtime)
{
  String sbuf;
  float lambda,q;
  lambda=ra_;
  q=dec_;

  //Deep sky objects
  if (SelectedCatalogue>=0){
    //RA
    lambda=ra_+ra_pm*(curtime.jul-PS.StartEpoch[SelectedCatalogue+3])/365.25;
    sbuf=GetRAString(lambda);
    dsRA->Caption="Right ascension: "+sbuf;

    //Declination
    q=dec_+dec_pm*(curtime.jul-PS.StartEpoch[SelectedCatalogue+3])/365.25;
    sbuf=GetDecString(q,false);
    dsDec->Caption="Declination: "+sbuf;

    //Galactic longitude and latitude
    EqToGal(&lambda,&q);
    sbuf=GetDecString(lambda,false);
    dsGalRA->Caption="Galactic long.: "+sbuf;
    sbuf=GetDecString(q,false);
    dsGalDec->Caption="Galactic lat.: "+sbuf;

    //Magnitude
    if (vismag_!=-100) sbuf=FormatFloat("#0.0",vismag_);
    else sbuf="n.a.";
    dsVMag->Caption="Visual magnitude: "+sbuf;

    //Elevation
    EquToHor(&ra_,&dec_,&lambda,&q,AL.l_ori);
    sbuf=GetDecString(q,false);
    dsElevation->Caption="Elevation: "+sbuf;

    //Azimuth
    sbuf=GetDecString(lambda,false);
    dsAzimuth->Caption="Azimuth: "+sbuf;

    //Rise and set times
    dsRiseTime->Caption="Rising time: "+GetRiseTime(ra_,dec_,AL.l_ori,curtime);
    dsSetTime->Caption="Setting time: "+GetSetTime(ra_,dec_,AL.l_ori,curtime);
  }

  //Planets
  double T=curtime.jul;
  ViewingTime=T;
  if (SelectedCatalogue==-1){
    //Best to simply recompute the ephemeris because some objects in this
    //class (e.g. planetary satelites) may be moving quite
    //rapidly
    GetOE(&T);
    GetObliquity(&T);
    EclSatCoord(0,&T);          //Get the sun position
    EclSatCoord(CatNum[1],&T);  //Get the heliocentric coordinates of the planet
    GeoSatCoord(CatNum[1]);     //Convert to geocentric equatorial coordinates

    //RA
    ra_=(float)OElem[CatNum[1]][9];

    //Declination
    dec_=(float)OElem[CatNum[1]][10];

    //RA
    sbuf=GetRAString(ra_);
    plRA->Caption="Right ascenssion: "+sbuf;

    //Declination
    sbuf=GetDecString(dec_,false);
    plDec->Caption="Declination: "+sbuf;

    //Elevation
    EquToHor(&ra_,&dec_,&lambda,&q,AL.l_ori);
    sbuf=GetDecString(q,false);
    plElevation->Caption="Elevation: "+sbuf;

    //Azimuth
    sbuf=GetDecString(lambda,false);
    plAzimuth->Caption="Azimuth: "+sbuf;

    //Rise and set times
    plRiseTime->Caption="Rising time: "+GetRiseTime(ra_,dec_,AL.l_ori,curtime);
    plSetTime->Caption="Setting time: "+GetSetTime(ra_,dec_,AL.l_ori,curtime);

    //Illumination fraction
    if (CatNum[1]!=0){
      double k=0, mag=0;
      Illumination(CatNum[1],&k,&mag);
      if (mag!=-20) plMag->Caption="Visual magnitude: "+FormatFloat("###0.00",mag);
      else plMag->Caption="Visual magnitude: ?";
      plIllumination->Caption="Illuminated fraction: "+FormatFloat("##0.0",k*100)+"\%";
    } else {
      plMag->Caption="Visual magnitude: -26.75";
      plIllumination->Caption="Illuminated fraction: 100\%";
    }

    //Draw the planet image
    DrawPlanet(T);
  }

  //Stars
  if (SelectedCatalogue==-2){
    //RA and declination
    q=dec_;
    lambda=ra_;
    Precess_sp(&lambda,&q,&(PS.time.jul),&(PS.StartEpoch[SelectedCatalogue+3]));

    lambda += ra_pm*(curtime.jul-PS.StartEpoch[SelectedCatalogue+3])*1.327347e-8;
    sbuf=GetRAString(lambda);
    stRA->Caption="Right ascension: "+sbuf;

    q=dec_+dec_pm*(curtime.jul-PS.StartEpoch[SelectedCatalogue+3])*1.327347e-8;
    sbuf=GetDecString(q,false);
    stDec->Caption="Declination:         "+sbuf;

    //Visual magnitude
    sbuf=FormatFloat("#0.00",vismag_);
    stVmag->Caption="Visual magnitude:             "+sbuf;

    //Elevation
    EquToHor(&ra_,&dec_,&lambda,&q,AL.l_ori);
    sbuf=GetDecString(q,false);
    stElevation->Caption="Elevation: "+sbuf;

    //Azimuth
    sbuf=GetDecString(lambda,false);
    stAzimuth->Caption="Azimuth:   "+sbuf;

    //Galactic longitude and latitude
    EqToGal(&lambda,&q);
    sbuf=GetDecString(lambda,false);
    stGalRA->Caption="Galactic long.: "+sbuf;
    sbuf=GetDecString(q,false);
    stGalDec->Caption="Galactic lat.:    "+sbuf;

    //Rise and set times
    stRiseTime->Caption="Rising time: "+GetRiseTime(ra_,dec_,AL.l_ori,curtime);
    stSetTime->Caption="Setting time: "+GetSetTime(ra_,dec_,AL.l_ori,curtime);
  }

  ObjectInfo->Repaint();
}
//---------------------------------------------------------------------------
//Shows the planet information
void __fastcall TObjectInfo::PlanetInfo(TObject *Sender)
{
  String ostring;

  //Name
  plName->Caption=SSData[CatNum[1]].Name;
  CommonName=SSData[CatNum[1]].Name;

  //Show object astrometric information, and draw the
  //planet
  ShowInfo(Sender,PS.time);
}
//---------------------------------------------------------------------------
//Draw the planet image
void __fastcall TObjectInfo::DrawPlanet(double T)
{
  float lambda,q,dl,dq,lambda_p,q_p,ScreenScale,dp,pl_radius;
  int x1=0,y1=0;
  int h2,w2,r;
  h2=PImage->Height/2;
  w2=PImage->Width/2;

  //Clear the canvas
  PlBitmap->Canvas->Pen->Color=clBlack;
  PlBitmap->Canvas->Brush->Color=clBlack;
  PlBitmap->Canvas->Rectangle(0,0,PImage->Width,PImage->Height);

  GetOE(&T);
  GetObliquity(&T);
  EclSatCoord(0,&T);          //Get the sun position
  EclSatCoord(CatNum[1],&T);  //Get the heliocentric coordinates of the planet
  GeoSatCoord(CatNum[1]);     //Convert to geocentric equatorial coordinates

  lambda_p=OElem[CatNum[1]][9];
  q_p=OElem[CatNum[1]][10];

  if (CatNum[1]==100) {
    //Draw the moon
    DrawMoonDisk();
  } else {
    //Draw the planet disk
    DrawPlanetDisk(T);
  }

  //Draw the satelites (other than the moon)

  //Get the screen scale
  ScreenScale=PScale*w2*360.0/(BasePScale*pi);

  pl_radius=asin(SSData[CatNum[1]].OE[18]/OElem[CatNum[1]][11]);
  for (int i=BaseIndex;i<BaseIndex+NSat;i++){
    //Get the planetocentric (or other) coordinates of the satelite (usually
    //refered to the earth's equator and 0 meridian)
    EclSatCoord(i,&T);
    dp=OElem[CatNum[1]][6]*OElem[i][6]+OElem[CatNum[1]][7]*OElem[i][7]+OElem[CatNum[1]][8]*OElem[i][8];

    //Convert to geocentric equatorial coordinates
    GeoSatCoord(i);
    lambda=OElem[i][9];
    q=OElem[i][10];

    //Check to see that the satelite is not eclipsed by the planet
    dl=lambda-OElem[CatNum[1]][9];
    dq=q-OElem[CatNum[1]][10];

    if ((dp>0)||(sqrt(dl*dl+dq*dq)>pl_radius)){
      //Get the screen coordinates (a flat projection is sufficient here).
      if (!XMirrorOn) x1=w2+(lambda-lambda_p)*ScreenScale;
      else x1=w2-(lambda-lambda_p)*ScreenScale;
      if (!YMirrorOn) y1=h2-(q_p-q)*ScreenScale;
      else y1=h2+(q_p-q)*ScreenScale;
      r=ScreenScale*SSData[i].OE[18]*4.2635e-5/OElem[CatNum[1]][11];
      if (r<1) r=1;
      if (sqrt(dl*dl+dq*dq)>pl_radius){
        //Satelite is away from the planet
        PlBitmap->Canvas->Pen->Color=clWhite;
        PlBitmap->Canvas->Brush->Color=clWhite;
      } else {
        //Satelite is in transit across the face of the planet
        PlBitmap->Canvas->Pen->Color=clBlack;
        PlBitmap->Canvas->Brush->Color=clBlack;
      }
      PlBitmap->Canvas->Ellipse(x1-r,y1-r,x1+r,y1+r);
    }
  }

  //Draw satelite path
  double Period, dT;
  if (PS.DrawPPath[CatNum[1]]){
    for (int i=BaseIndex;i<BaseIndex+NSat;i++){
      Period=twopi/SSData[i].OE[11];
      dT=Period/400.0;
      for (double T_=T;T_<T+Period;T_+=dT){
        GetOE(&T_);
        GetObliquity(&T_);
        EclSatCoord(0,&T_);          //Get the sun position
        EclSatCoord(CatNum[1],&T_);  //Get the heliocentric coordinates of the planet
        GeoSatCoord(CatNum[1]);     //Convert to geocentric equatorial coordinates
        lambda_p=OElem[CatNum[1]][9];
        q_p=OElem[CatNum[1]][10];

        //Get the planetocentric (or other) coordinates of the satelite (usually
        //refered to the earth's equator and 0 meridian)
        EclSatCoord(i,&T_);
        dp=OElem[CatNum[1]][6]*OElem[i][6]+OElem[CatNum[1]][7]*OElem[i][7]+OElem[CatNum[1]][8]*OElem[i][8];

        //Convert to geocentric equatorial coordinates
        GeoSatCoord(i);
        lambda=OElem[i][9];
        q=OElem[i][10];

        //Check to see that the satelite is not eclipsed by the planet
        dl=lambda-OElem[CatNum[1]][9];
        dq=q-OElem[CatNum[1]][10];
        pl_radius=asin(SSData[CatNum[1]].OE[18]/OElem[CatNum[1]][11]);

        if ((sqrt(dl*dl+dq*dq)>pl_radius)||(dp>0)){
          //Get the screen coordinates (a flat projection is sufficient here).
          if (!XMirrorOn) x1=w2+(lambda-lambda_p)*ScreenScale;
          else x1=w2-(lambda-lambda_p)*ScreenScale;
          if (!YMirrorOn) y1=h2-(q_p-q)*ScreenScale;
          else y1=h2+(q_p-q)*ScreenScale;
          PlBitmap->Canvas->Pixels[x1][y1]=clGray;
        }
      }
    }
    //Return coordinates to current time
    GetOE(&T);
    GetObliquity(&T);
    EclSatCoord(0,&T);          //Get the sun position
    EclSatCoord(CatNum[1],&T);  //Get the heliocentric coordinates of the planet
    GeoSatCoord(CatNum[1]);     //Convert to geocentric equatorial coordinates
    for (int i=BaseIndex;i<BaseIndex+NSat;i++){
      EclSatCoord(i,&T);
      GeoSatCoord(i);
    }
  }

  //Text strings
  PlBitmap->Canvas->Font->Color=clWhite;
  PlBitmap->Canvas->Brush->Color=clBlack;
  if (!YMirrorOn){
    PlBitmap->Canvas->TextOut(w2-5,2,"N");
    PlBitmap->Canvas->TextOut(w2-5,2*h2-15,"S");
  } else {
    PlBitmap->Canvas->TextOut(w2-5,2,"S");
    PlBitmap->Canvas->TextOut(w2-5,2*h2-15,"N");
  }
  if (!XMirrorOn){
    PlBitmap->Canvas->TextOut(2,h2,"E");
    PlBitmap->Canvas->TextOut(w2*2-12,h2,"W");
  } else {
    PlBitmap->Canvas->TextOut(2,h2,"W");
    PlBitmap->Canvas->TextOut(w2*2-12,h2,"E");
  }
  String label;
  float FOV=BasePScale/PScale;
  if (FOV>=10) label="Field of view: "+FormatFloat("00.0",FOV)+"";
  if (FOV<10 && FOV>=1) label="Field of view: "+FormatFloat("0.00",FOV)+"";
  if (FOV<1 && FOV>=.1) label="Field of view: "+FormatFloat("0.00",FOV)+"";
  if (FOV<.1 && FOV>=.01) label="Field of view: "+FormatFloat("0.000",FOV)+"";
  if (FOV<.01) label="Field of view: "+FormatFloat("0.0000",FOV)+"";
  PlBitmap->Canvas->TextOut(2,2,label);
  PlBitmap->Canvas->TextOut(2,h2*2-15,"JD: "+FormatFloat("#######.##",ViewingTime));

  PImage->Picture->Bitmap=PlBitmap;
  PImage->Repaint();
}
//---------------------------------------------------------------------------
//Draw the grid that defines the planet's orientation and the planet texture
void __fastcall TObjectInfo::DrawPlanetGrid(int n,float res,double T_)
{
  float ScreenScale;
  double lambda=0,q=0,t2,l2,t,l,incr,smallincr,r;
  int x_,w2,h2;
  bool Proceed;
  TColor color;
  TPoint points[5000];

  w2=PImage->Width/2;
  h2=PImage->Height/2;

  //Render surface using polygons
  if (PScale<=4) incr=pi/(18.0*res);
  if (PScale==8) incr=pi/(36.0*res);
  if (PScale>8) incr=pi/(72.0*res);

  //Get the screen scale
  ScreenScale=PScale*2*w2/(BasePScale*pi/180);

  r=atan(SSData[n].OE[18]/OElem[n][11]); //Apparent angular planet radius

  //Initialize variables in coordinate transformation routine
  l=0.001; t=0.001;
  PlanetToGeoCoord(&n,&r,&T_,&l,&t,&lambda,&q,true);

  smallincr=incr/4;
  PlBitmap->Canvas->Pen->Width=1;
  for (l=0;l<=twopi;l+=incr){
    x_=l*57.29577951;
    for (t=-pihalved;t<=pihalved;t+=incr){
      //Get the polygon color
      color=TextureMap->Canvas->Pixels[x_][(pihalved-t)*57.29577951];
      PlBitmap->Canvas->Brush->Color=color;
      PlBitmap->Canvas->Pen->Color=color;

      //FIRST CORNER OF THE POLYGON
      //Convert planet surface coordinates to geocentric equatorial coordinates
      l2=l; t2=t;
      Proceed=PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false);

      if (Proceed){
        if (XMirrorOn) points[0].x=w2-lambda*ScreenScale;
        else points[0].x=w2+lambda*ScreenScale;
        if (YMirrorOn) points[0].y=h2+q*ScreenScale;
        else points[0].y=h2-q*ScreenScale;

        //SECOND CORNER OF THE POLYGON
        //Convert planet surface coordinates to geocentric equatorial coordinates
        l2=l+incr;
        Proceed=PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false);

        //Find a good point if the last one was hidden
        if (!Proceed){
          do {l2-=smallincr;} while (!PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false));
        }

        if (XMirrorOn) points[1].x=w2-lambda*ScreenScale;
        else points[1].x=w2+lambda*ScreenScale;
        if (YMirrorOn) points[1].y=h2+q*ScreenScale;
        else points[1].y=h2-q*ScreenScale;

        //THIRD CORNER OF THE POLYGON
        //Convert planet surface coordinates to geocentric equatorial coordinates
        l2=l+incr; t2=t+incr;
        Proceed=PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false);

        //Find a good point if the last one was hidden
        if (!Proceed){
          do {t2-=smallincr;} while (!PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false) && t2>t);
          if (t2<t){
            t2=t;
            do {l2-=smallincr;} while (!PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false) && l2>l);
          }
        }

        if (XMirrorOn) points[2].x=w2-lambda*ScreenScale;
        else points[2].x=w2+lambda*ScreenScale;
        if (YMirrorOn) points[2].y=h2+q*ScreenScale;
        else points[2].y=h2-q*ScreenScale;

        //FOURTH CORNER OF THE POLYGON
        //Convert planet surface coordinates to geocentric equatorial coordinates
        l2=l; t2=t+incr;
        Proceed=PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false);

        //Find a good point if the last one was hidden
        if (!Proceed){
          do {t2-=incr/10;} while (!PlanetToGeoCoord(&n,&r,&T_,&l2,&t2,&lambda,&q,false) && t2>t);
        }

        if (XMirrorOn) points[3].x=w2-lambda*ScreenScale;
        else points[3].x=w2+lambda*ScreenScale;
        if (YMirrorOn) points[3].y=h2+q*ScreenScale;
        else points[3].y=h2-q*ScreenScale;

        PlBitmap->Canvas->Polygon(points,3);
      }
    }
  }
}
//---------------------------------------------------------------------------
//Draw Saturn's rings
void __fastcall TObjectInfo::SaturnRings(float res, double T_)
{
  float ScreenScale, PlanetSize2;
  double lambda=0,q=0,t,l,incr,r;
  int n,w2,h2,i;
  bool Proceed,Drawing=false;
  TPoint points[1000];

  w2=PImage->Width/2;
  h2=PImage->Height/2;

  //Get the screen scale
  ScreenScale=PScale*2*w2/(BasePScale*pi/180);

  //Initialize variables in coordinate transformation routine
  l=0.001; t=0.001;  r=1; n=5;
  PlanetToGeoCoord(&n,&r,&T_,&l,&t,&lambda,&q,true);

  t=0;
  PlanetSize2=atan(SSData[5].OE[18]/OElem[5][11]);
  PlanetSize2=PlanetSize2*PlanetSize2; //Apparent angular radius of the planet squared
  if (PScale>3) PlBitmap->Canvas->Pen->Width=2;
  else PlBitmap->Canvas->Pen->Width=1;
  int n_ring=25*PScale;
  incr=twopi/120.0;
  for (int j=0;j<n_ring;j++){
    r=(j*4.186836e-4/n_ring)+5.106375e-4;

    if (r<=5.106375e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x709090;
    if (r>5.106375e-4 && r<=6.162866e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x7090a0;
    if (r>6.162866e-4 && r<=6.710677e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0xc0f0ff;
    if (r>6.710677e-4 && r<=6.876976e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x407080;
    if (r>6.876976e-4 && r<=7.043275e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x90c0d0;
    if (r>7.043275e-4 && r<=7.581303e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x80b0c0;
    if (r>7.581303e-4 && r<=7.825862e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0xa0e0f0;
    if (r>7.825862e-4 && r<=8.021508e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0xc0e0f0;
    if (r>8.021508e-4 && r<=8.275849e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x000000;
    if (r>8.275849e-4 && r<=8.980177e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x80c0d0;
    if (r>8.980177e-4 && r<=9.077999e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0x407080;
    if (r>9.077999e-4 && r<=9.293211e-4) PlBitmap->Canvas->Pen->Color=(Graphics::TColor)0xa0e0f0;
    r=r/OElem[5][11];

    i=0;
    Drawing=false;
    for (l=0;l<=twopi+.0001;l+=incr){
      //Convert planet surface coordinates to geocentric
      //equatorial coordinates
      Proceed=PlanetToGeoCoord(&n,&r,&T_,&l,&t,&lambda,&q,false);

      //Check that the coordinate is not behind the planet
      if (!Proceed && !(lambda*lambda+q*q<=PlanetSize2)) Proceed=true;

      if (Proceed){
        if (XMirrorOn) lambda=-lambda;
        if (YMirrorOn) q=-q;
        points[i].x=w2+lambda*ScreenScale;
        points[i].y=h2-q*ScreenScale;
        i++;
        Drawing=true;
      } else {
        if (Drawing) {
          PlBitmap->Canvas->Polyline(points,i-1);
          i=0;
        }
        Drawing=false;
      }
    }
    if (Drawing) PlBitmap->Canvas->Polyline(points,i-1);
  }
}
//---------------------------------------------------------------------------
//Apply illumination to planet disk
void __fastcall TObjectInfo::IlluminateDisk(int N)
{
  double x,y,z,xh,yh,zh,l,q,l_,dp,k,ScreenScale;
  double *X, dbuf;
  int r,g,b,col,w2,h2,i_,j_;
  float luminance;
  double pr_e; //Planet's radius in earth radii
  double pr;   //Planet's apparent geocentric angular radius (in radians)
  double pr_h; //Distance to planet from the sun (in a.u.)
  double pr2,r2;

  w2=PImage->Width/2;
  h2=PImage->Height/2;

  X=d_vector(0,3);

  //Get the screen scale
  ScreenScale=PScale*2*w2/(BasePScale*pi/180);

  pr_e=SSData[N].OE[18];
  pr=1.01*atan(pr_e/OElem[N][11])+2/ScreenScale;
  pr2=pr*pr;

  //Compute the vector from the sun to the planet
  xh=OElem[N][6]-OElem[0][6];
  yh=OElem[N][7]-OElem[0][7];
  zh=OElem[N][8]-OElem[0][8];
  pr_h=sqrt(xh*xh+yh*yh+zh*zh);

  //Render the image
  for (int i=-pr*ScreenScale;i<pr*ScreenScale;i++){
    l=i/ScreenScale;
    dbuf=OElem[N][9]+l;
    FirstCircle2(&dbuf);
    l_=dbuf;
    for (int j=-pr*ScreenScale;j<pr*ScreenScale;j++){
      q=j/ScreenScale;
      r2=l*l+q*q;
      if (r2<=pr2){
        //Convert geocentric spherical coordinates into
        //rectangular coordinates of the planet disk point
        X[0]=OElem[N][11];
        X[2]=l_;
        dbuf=OElem[N][10]+q;
        FirstCircle2(&dbuf);
        X[1]=dbuf;
        SphToCosines2(X);

        //Compute the normal vector (x,y,z) to planet's surface
        k=pr_e*sqrt(1-(r2/pr2))/OElem[N][11];
        x=X[0]-OElem[N][6]+k*OElem[N][6];
        y=X[1]-OElem[N][7]+k*OElem[N][7];
        z=X[2]-OElem[N][8]+k*OElem[N][8];

        //Dot the vector from the sun to the planet with the normal
        //vector to determine the illumination obliquity.
        dp=(x*xh+y*yh+z*zh)/(pr_e*pr_h);
        if (dp<0) dp=0;

        luminance=pow(fabs(dp),.5);
        if (luminance>1) luminance=1;
        if (luminance<=0) luminance=.05;

        if (XMirrorOn) i_=-i;
        else i_=i;
        if (YMirrorOn) j_=-j;
        else j_=j;
        col=(int)PlBitmap->Canvas->Pixels[i_+w2][h2-j_];
        r=(int)(luminance*(col&0x0000ff));
        g=(int)(luminance*((col&0x00ff00)>>8))<<8;
        b=(int)(luminance*((col&0xff0000)>>16))<<16;
        PlBitmap->Canvas->Pixels[i_+w2][h2-j_]=(Graphics::TColor)(b|g|r);
      }
    }
  }
  free_d_vector(X,0,3);
}
//---------------------------------------------------------------------------
//Draws the moon's disk
void __fastcall TObjectInfo::DrawMoonDisk()
{
  double x,y,z,l_,step,dp,k,ScreenScale,theta,sint,cost;
  double *X, dbuf, T;
  int i,j,bmh2,greylevel,w2,h2;
  float luminance;
  double mr_e; //Moon's radius in earth radii
  double mr; //Moon's apparent geocentric radius in radians
  double mr2,r2;
  double TempMOElem[12];
  Graphics::TColor color;
  Graphics::TBitmap *bufbmp = new Graphics::TBitmap();

  TRect MoonRect;

  w2=PImage->Width/2;
  h2=PImage->Height/2;

  X=d_vector(0,3);

  while (DrawingMap){};
  DrawingMap=true;
  mr_e=SSData[100].OE[18];
  mr=atan(mr_e/OElem[100][11]);
  mr2=mr*mr;

  //Get the screen scale
  ScreenScale=PScale*w2*360.0/(BasePScale*pi);
  bufbmp->Assign(PlBitmap);
  bmh2=ScreenScale*mr;
  bufbmp->Height=bmh2*2;
  bufbmp->Width=bmh2*2;
  bufbmp->Canvas->Brush->Color=clBlack;
  bufbmp->Canvas->Rectangle(0,0,bmh2*2,bmh2*2);
  PlBitmap->Canvas->Brush->Color=clBlack;
  PlBitmap->Canvas->Rectangle(0,0,w2*2,h2*2);

  //Draw the moon bitmap, first you must compute the angle by which to rotate
  //the moon image.  This is determined by the angle of the moon's orbital plane
  //projected on the sky (i.e. the moon's pole should always be perpendicular
  //to its orbital plane).  This is most easily computed by simply looking at
  //the difference in the sky coordinates a small time distance appart
  for (int i=0;i<12;i++){TempMOElem[i]=OElem[100][i];} //Backup the orbit parameters
  T=PS.time.jul + 0.1;
  GetOE_(100,&T);
  EclSatCoord(100,&T);
  GeoSatCoord(100);
  theta=atan((OElem[100][10]-TempMOElem[10])/(OElem[100][9]-TempMOElem[9]));
  for (int i=0;i<12;i++){OElem[100][i]=TempMOElem[i];} //Restore the orbit parameters
  sint=sin(theta);
  cost=cos(theta);

  MoonRect=Rect(bmh2-mr*ScreenScale,bmh2-mr*ScreenScale,bmh2+mr*ScreenScale,bmh2+mr*ScreenScale);
  bufbmp->Canvas->StretchDraw(MoonRect,MoonBitmap);

  //Render the image; compute the grayscale level to use
  //at each point on the moon's disk
  step=0.99/ScreenScale;

  for (double l=-mr;l<mr;l+=step){
    dbuf=OElem[100][9]+l;
    FirstCircle2(&dbuf);
    l_=dbuf;
    for (double q=-mr;q<mr;q+=step){
      r2=l*l+q*q;
      if (r2<=mr2){
        //Convert geocentric spherical coordinates into
        //rectangular geocentric coordinates of the planet
        //disk point
        X[0]=OElem[100][11];
        X[2]=l_;
        dbuf=OElem[100][10]+q;
        FirstCircle2(&dbuf);
        X[1]=dbuf;
        SphToCosines2(X);

        //Compute the normal vector (x,y,z) to moon's surface
        k=mr_e*sqrt(1-(r2/mr2))/OElem[100][11];
        x=X[0]-OElem[100][6]+k*OElem[100][6];
        y=X[1]-OElem[100][7]+k*OElem[100][7];
        z=X[2]-OElem[100][8]+k*OElem[100][8];

        //Assume the vector from the sun to earth is close
        //to the vector from the sun to the moon.  Dot the
        //solar vector with the normal vector to determine
        //the illumination obliquity.
        dp=-(x*OElem[0][6]+y*OElem[0][7]+z*OElem[0][8])/(mr_e*OElem[0][11]);
        if (dp<0) dp=0;

        luminance=pow(fabs(dp),.5);
        if (luminance>1) luminance=1;
        if (luminance<0||luminance==0) luminance=.15;

        i=ScreenScale*l;
        j=ScreenScale*q;
        if (XMirrorOn) i=-i;
        if (YMirrorOn) j=-j;
        greylevel=(((int)bufbmp->Canvas->Pixels[bmh2+i*cost+j*sint][bmh2+i*sint-j*cost])&0x0000ff)*luminance;
        color=(Graphics::TColor)((greylevel<<16)|(greylevel<<8)|(greylevel));
        PlBitmap->Canvas->Pixels[i+w2][h2-j]=color;
      }
    }
  }
  PImage->Picture->Bitmap=PlBitmap;
  DrawingMap=false;

  free_d_vector(X,0,3);
}
//---------------------------------------------------------------------------
//Draw the planet disk
void __fastcall TObjectInfo::DrawPlanetDisk(double T)
{
  String filename = PS.ProgPath;
  switch (CatNum[1]) {
    case 0: filename += "\\Planets\\Bitmaps\\sun_txt.bmp"; break;
    case 1: filename += "\\Planets\\Bitmaps\\mer_txt.bmp"; break;
    case 2: filename += "\\Planets\\Bitmaps\\ven_txt.bmp"; break;
    case 3: filename += "\\Planets\\Bitmaps\\mars_txt.bmp"; break;
    case 4: filename += "\\Planets\\Bitmaps\\jup_txt.bmp"; break;
    case 5: filename += "\\Planets\\Bitmaps\\sat_txt.bmp"; break;
    case 6: filename += "\\Planets\\Bitmaps\\ura_txt.bmp"; break;
    case 7: filename += "\\Planets\\Bitmaps\\nep_txt.bmp"; break;
    case 8: filename += "\\Planets\\Bitmaps\\plu_txt.bmp"; break;
  }
  if (!FileExists(filename)) {
    String osbuf = "Could not open file " + filename;
    Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    return;
  }
  TextureMap->LoadFromFile(filename);

  DrawPlanetGrid(CatNum[1],PS.PlanetResolution,T);
  if (CatNum[1]>0) IlluminateDisk(CatNum[1]); //Apply illumination to disk
  if (CatNum[1]==5) SaturnRings(PS.PlanetResolution,T);
}
//---------------------------------------------------------------------------
//Highlight the the currently displayed object
void __fastcall TObjectInfo::HighlightButClick(TObject *Sender)
{
  if (!highlighted){

    //Add object to the highlight list
    HO[HighlightNum].ra=ra_;
    HO[HighlightNum].dec=dec_;
    HO[HighlightNum].CatIndex=SelectedCatalogue;
    HO[HighlightNum].CatNum=CatNum[SelectedCatalogue+2];
    HO[HighlightNum].CatExt=ExtNum;

    MainF->Highlight(&HighlightNum,true);
    MainF->Image->Repaint();
    HighlightBut->Caption="Remove Highlight";
    highlighted=true;
    HighlightNum++;

    //If the object is in the selected object list
    //then change the highlight setting
    int i=0;
    while ((CatNum[SelectedCatalogue+2]!=SO[i].CatNum||ExtNum!=SO[i].CatExt\
      ||SelectedCatalogue!=SO[i].CatIndex)&&(i<ObjList->Items->Count)){
      i++;
    }
    if (i<ObjList->Items->Count){
      //Object found, make the changes
      SO[i].Highlight=true;
      if (ObjList->ItemIndex==i) HighlightBut->Caption="Remove Highlight";
    } else {
      //Not found, add object to selected objects list
      AddToListButClick(Sender);
    }
  } else {
    //Find the object in the highlight list
    int i=0;
    while ((CatNum[SelectedCatalogue+2]!=HO[i].CatNum||ExtNum!=HO[i].CatExt||\
      SelectedCatalogue!=HO[i].CatIndex)&&(i<HighlightNum)){
      i++;
    }
    MainF->Highlight(&i,false);
    MainF->Image->Repaint();
    highlighted=false;
    HighlightNum--;

    //Remove object from the list and shift the contents of
    //the list up to fill the gap in the array
    if (HighlightNum>1){
      for (int j=i;j<HighlightNum-1;j++){
        HO[j].ra=HO[j+1].ra;
        HO[j].dec=HO[j+1].dec;
        HO[j].CatIndex=HO[j+1].CatIndex;
        HO[j].CatNum=HO[j+1].CatNum;
        HO[j].CatExt=HO[j+1].CatExt;
      }
    }

    //If the object is in the selected object list
    //then change the highlight setting
    i=0;
    while ((CatNum[SelectedCatalogue+2]!=SO[i].CatNum||ExtNum!=SO[i].CatExt||\
      SelectedCatalogue!=SO[i].CatIndex)&&(i<ObjList->Items->Count)){
      i++;
    }
    if (i<ObjList->Items->Count){
      //Object found, make the changes
      SO[i].Highlight=false;
      if (ObjList->ItemIndex==i) HighlightBut->Caption="Highlight";
    }

    HighlightBut->Caption="Highlight";
  }
}
//---------------------------------------------------------------------------
//Add an object to the selected object list
void __fastcall TObjectInfo::AddToListButClick(TObject *Sender)
{
  int index=ObjList->Items->Count;
  String sbuf;

  String Ext[10]={"","A","B","C","D","E","F","G","H","I"};
  String CatName[12]={"SAO ","","NGC ","IC ","M ","Lynd ","Barnard ","Arp ","","","",""};

  if (SelectedCatalogue==-1){
    sbuf=CommonName;
  } else {
    if (CommonName!="") sbuf=CatName[SelectedCatalogue+2]+IntToStr(CatNum[SelectedCatalogue+2])+", "+CommonName;
    else sbuf=CatName[SelectedCatalogue+2]+IntToStr(CatNum[SelectedCatalogue+2]);

    //Special cases (over-rides above default)
    //NGC
    if (SelectedCatalogue==0){
      if (CommonName!="") sbuf="NGC "+IntToStr(CatNum[2])+Ext[ExtNum]+", "+CommonName;
      else sbuf="NGC "+IntToStr(CatNum[2])+Ext[ExtNum];
      SO[index].CatExt=ExtNum;
    }

    //Barnard
    if (SelectedCatalogue==4){
      if (CommonName!="") sbuf="Barnard "+IntToStr(CatNum[6])+Ext[ExtNum]+", "+CommonName;
      else sbuf="Barnard "+IntToStr(CatNum[6])+Ext[ExtNum];
      SO[index].CatExt=ExtNum;
    }
  }
  SO[index].CatNum=CatNum[SelectedCatalogue+2];
  SO[index].Name=sbuf;
  SO[index].ra=ra_;
  SO[index].dec=dec_;
  SO[index].CatIndex=SelectedCatalogue;
  SO[index].Highlight=highlighted;

  ObjList->Items->Add(sbuf);
  ObjList->ItemIndex=ObjList->Items->Count-1;
  AddToListBut->Enabled=false;
}
//---------------------------------------------------------------------------
//Remove an item from the list
void __fastcall TObjectInfo::RemoveButClick(TObject *Sender)
{
  //Shift the items in the structure as required
  for (int i=ObjList->ItemIndex;i<ObjList->Items->Count-1;i++){
    SO[i].ra=SO[i+1].ra;
    SO[i].dec=SO[i+1].dec;
    SO[i].Name=SO[i+1].Name;
    SO[i].CatIndex=SO[i+1].CatIndex;
    SO[i].CatNum=SO[i+1].CatNum;
    SO[i].CatExt=SO[i+1].CatExt;
    SO[i].Highlight=SO[i+1].Highlight;
  }
  if (CurrentObject==ObjList->ItemIndex+1) CurrentObject--;

  ObjList->Items->Delete(ObjList->ItemIndex);
}
//---------------------------------------------------------------------------
//Selects the item as the currently displayed item
void __fastcall TObjectInfo::SelectButClick(TObject *Sender)
{
  int index=ObjList->ItemIndex;
  CurrentObject=index+1;

  ra_=SO[index].ra;
  dec_=SO[index].dec;
  SelectedCatalogue=SO[index].CatIndex;
  highlighted=SO[index].Highlight;
  CatNum[SelectedCatalogue+2]=SO[index].CatNum;
  ExtNum=SO[index].CatExt;
  FormActivate(Sender);
  AddToListBut->Enabled=false;
}
//---------------------------------------------------------------------------
//Center the map on the selected object
void __fastcall TObjectInfo::CenterButClick(TObject *Sender)
{
  float azim,elev;
  EquToHor(&ra_,&dec_,&azim,&elev,AL.l_ori);
  AL.incl=elev;
  AL.azim=azim;
  AL.cosi=cos(AL.incl);
  AL.sini=sin(AL.incl);
  AL.cos_azim=cos(AL.azim);
  AL.sin_azim=sin(AL.azim);

  //Redraw the map and coordinates
  int X=PS.w2, Y=PS.h2;
  MainF->DrawCoord(&X,&Y,true,true);
  if (!DrawingMap){
    DrawingMap=true;
    MainF->DrawMap(PS.MaxStarMag,false);
    DrawingMap=false;
  }
}
//---------------------------------------------------------------------------
//Disable lower buttons
void __fastcall TObjectInfo::SelectionsShow(TObject *Sender)
{
  AddToButEn=AddToListBut->Enabled;
  CenterViewOnButEn=CenterBut->Enabled;

  AddToListBut->Enabled=false;
  CenterBut->Enabled=false;
}
//---------------------------------------------------------------------------
//Return lower buttons to original state
void __fastcall TObjectInfo::SelectionsHide(TObject *Sender)
{
  AddToListBut->Enabled=AddToButEn;
  CenterBut->Enabled=CenterViewOnButEn;
}
//---------------------------------------------------------------------------
//Update the list display
void __fastcall TObjectInfo::ObjListKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
  ObjListClick(Sender);
}
//---------------------------------------------------------------------------
//Update the list display
void __fastcall TObjectInfo::ObjListClick(TObject *Sender)
{
  if (SO[ObjList->ItemIndex].Highlight){
    HighlightBut->Caption="Remove Highlight";
  } else {
    HighlightBut->Caption="Highlight";
  }
  if (ObjList->ItemIndex==CurrentObject-1){
    SelectBut->Enabled=false;
  } else {
    SelectBut->Enabled=true;
  }
}
//---------------------------------------------------------------------------
//Zoom into planet image
void __fastcall TObjectInfo::ZoomInClick(TObject *Sender)
{
  PScale=PScale*2;
  if (.1/PScale<.001) PScale=PScale/2;

  //Redraw the planet image
  DrawPlanet(ViewingTime);
}
//---------------------------------------------------------------------------
//Zoom out of planet image
void __fastcall TObjectInfo::ZoomOutButClick(TObject *Sender)
{
  PScale=PScale/2;
  if (.1/PScale>100) PScale=PScale*2;

  //Redraw the planet image
  DrawPlanet(ViewingTime);
}
//---------------------------------------------------------------------------
//Zoom to normal scale
void __fastcall TObjectInfo::Zoom1ButClick(TObject *Sender)
{
  PScale=1;

  //Redraw the planet image
  DrawPlanet(ViewingTime);
}
//---------------------------------------------------------------------------
//Flip the image right to left
void __fastcall TObjectInfo::XFlipButClick(TObject *Sender)
{
  XMirrorOn=!XMirrorOn;
  DrawPlanet(ViewingTime);
}
//---------------------------------------------------------------------------
//Flip the image top to bottom
void __fastcall TObjectInfo::YFlipButClick(TObject *Sender)
{
  YMirrorOn=!YMirrorOn;
  DrawPlanet(ViewingTime);
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::PlanetResClick(TObject *Sender,
      TUDBtnType Button)
{
  PS.PlanetResolution=PlanetRes->Position;
  DrawPlanet(ViewingTime);
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::TimeDateButClick(TObject *Sender)
{
  DrawMoonDisk();
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::ShowAsICClick(TObject *Sender)
{
  SelectedCatalogue=1;
  ShowICInfo(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::ShowAsRNGCClick(TObject *Sender)
{
  SelectedCatalogue=0;
  ShowRNGCInfo(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::ShowAsMesClick(TObject *Sender)
{
  SelectedCatalogue=2;
  CatNum[0]=CatNum[0];
  ShowMesInfo(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::XCatBtnClick(TObject *Sender)
{
  SelectedCatalogue=SourceCatalog;
  if (SourceCatalog==3) ShowLyndInfo(Sender);
  if (SourceCatalog==4) ShowBarnInfo(Sender);
  if (SourceCatalog==5) ShowArpInfo(Sender);
  if (SourceCatalog==6) ShowRC3Info(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::ShowSatPathClick(TObject *Sender)
{
  PS.DrawPPath[CatNum[1]]=!ShowSatPath->Down;
  DrawPlanet(ViewingTime);
}
//---------------------------------------------------------------------------
//Display angular position on diagram and determine the name of the nearest
//object
void __fastcall TObjectInfo::PImageMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
  double ScreenScale, r, lambda, q, dlambda, dq, lambdap, qp, dr, bestdr;
  int w2,h2,besti;

  w2=PImage->Width/2;
  h2=PImage->Height/2;

  //Get the screen scale
  ScreenScale=PScale*2*w2/(BasePScale*pi/180);

  //Convert (X,Y) coordinates to angular distances in declination and
  //longitude from the planet
  if (YMirrorOn) Y=h2*2-Y;
  if (XMirrorOn) X=w2*2-X;
  lambda=(X-w2)/ScreenScale;
  q=(h2-Y)/ScreenScale;
  r=sqrt(lambda*lambda+q*q);
  AngDist->Caption="R: "+GetDecString(r,true);
  PosAngle->Caption="W: "+GetDecString(lambda,true)+"    N: "+GetDecString(q,true);

  //Find the closest object to the cursor and display its name
  lambdap=OElem[CatNum[1]][9]+lambda;
  qp=OElem[CatNum[1]][10]-q;
  bestdr=10e+20;
  for (int i=BaseIndex;i<BaseIndex+NSat;i++){
    dlambda=OElem[i][9]-lambdap;
    dq=qp-OElem[i][10];
    dr=dlambda*dlambda+dq*dq;
    if (dr<bestdr){
      bestdr=dr;
      besti=i;
    }
  }
  dr=lambda*lambda+q*q;
  if ((sqrt(bestdr)<10/ScreenScale)||(sqrt(dr)<10/ScreenScale)){
    if (dr<bestdr){
      //Planet center is closest
      SatName->Caption="Object name: "+SSData[CatNum[1]].Name;
    } else {
      //Satelite center is closest
      SatName->Caption="Object name: "+SSData[besti].Name;
    }
  } else {
    SatName->Caption="Object name: ?";
  }
}
//---------------------------------------------------------------------------
//Find an object by its catalog number and display the information
void __fastcall TObjectInfo::FindByCatBtnClick(TObject *Sender)
{
  SelectedCatalogue=SearchCat_->ItemIndex-2;
  SourceCatalog=SelectedCatalogue;
  ExtNum=FindExt_->ItemIndex;
  CatNum[SelectedCatalogue+2]=FindNum_->Value;

  //Make sure the catalogue number isn't too large
  int maxval[20]={258997,0,8163,5386,110,1791,348,338,23011,0,0,0,0,0,0,0,0,0,0,0};
  if (CatNum[SelectedCatalogue+2]>maxval[SelectedCatalogue+2]) CatNum[SelectedCatalogue+2]=maxval[SelectedCatalogue+2];
  FindNum_->Value=CatNum[SelectedCatalogue+2];

  ShowObjectInfo(Sender);

  //Center the field of view on the object if requested
  if (CenterOnFound->Checked) CenterButClick(Sender);
}
//---------------------------------------------------------------------------
//Find an object by a string
void __fastcall TObjectInfo::FindByStrBtnClick(TObject *Sender)
{
  String SearchStr, filename, DescStr, NameStr, Descriptor;
  int MaxCatN[20]={3126,213,193,193,193,1791,348,338,23011,0,0,0,0,0,0,0,0,0,0,0};
  String ExtS[8]={"A","B","C","D","E","F","G","H"};
  int CatN, BestMatchNChar, MatchNChar, N, N_, N_Ext, NMatch, MaxNMatch;
  int *iptr,idum=0,j,i,SearchStrLen;
  short *sptr, sdum=0;
  float *fptr,fdum=0.0;
  iptr=&idum;
  sptr=&sdum;
  fptr=&fdum;

  CatN=SearchCat_->ItemIndex;
  SearchStr=SearchString_->Text.Trim();
  SearchStrLen=SearchStr.Length();
  if (SearchStr.Length()<1) return;
  MaxNMatch=NMatch_->Value;

  BestMatchNChar=0;
  NMatch=0;

  SearchIndicator->Caption="Searching...";
  SearchIndicator->Repaint();

  for (i=0;i<50;i++){
    BestMatches[i][0]=0;
    BestMatches[i][1]=0;
    BestMatches[i][2]=0;
    BestMatches[i][3]=0;
    BestMatchDesc[i]="";
  }

  //Case for stars
  if (CatN==0){
    char bufName[24],bufFlam[11];
    String Flamsteed;

    FILE *namefile;
    filename=PS.ProgPath+"Catalogs\\SAO\\NAME_CROSSREF.DAT";
    if (!FileExists(filename)) {
      String osbuf = "Could not open file " + filename;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
      return;
    }
    namefile=fopen(filename.c_str(),"rb");

    //Note that there are 31 bytes per record.
    i=0;
    while (i<MaxCatN[CatN]){
      //Get the SAO number
      fread(iptr,4,1,namefile);
      N=*iptr;

      //Get the flamsteed number
      fgets(bufFlam,11,namefile);
      Flamsteed=bufFlam;

      //Get the common name
      fgets(bufName,25,namefile);
      NameStr=bufName;
      NameStr=NameStr.Trim();

      //Formating the flamsteed name
      NameStr=FormatName(Sender,Flamsteed,NameStr,false);
      if (NameStr.SubString(NameStr.Length(),1)==",") NameStr=NameStr.SubString(1,NameStr.Length()-1);

      //Search the name for the string
      MatchNChar=CompareString(SearchStr,NameStr);

      if ((MatchNChar>=BestMatchNChar)&&(MatchNChar>0)) {
        Descriptor="SAO "+IntToStr(N)+", "+NameStr;
        AddToFoundList(CatN,N,0,MatchNChar,Descriptor,MaxNMatch);
        BestMatchNChar=MatchNChar;
        NMatch++;
      }
      i++;
    }
    fclose(namefile);
  }

  //Case for solar system objects
  if (CatN==1){
    for (N=0;N<MaxCatN[CatN];N++){
      NameStr=SSData[N].Name;
      //Search the name for the string
      if (NameStr!=""){
        MatchNChar=CompareString(SearchStr,NameStr);
        if ((MatchNChar>=BestMatchNChar)&&(MatchNChar>0)) {
          AddToFoundList(CatN,N,0,MatchNChar,NameStr,MaxNMatch);
          BestMatchNChar=MatchNChar;
          NMatch++;
        }
      }
    }
  }

  //Case for RNGC, IC and Messier objects
  if ((CatN<5)&&(CatN>1)){
    char pnamebuf[121];

    FILE *NAME;
    filename=PS.ProgPath+"Catalogs\\XRef\\Names.dat";
    if (!FileExists(filename)) {
      String osbuf = "Could not open file " + filename;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
      return;
    }
    NAME=fopen(filename.c_str(),"rb");

    if (CatN==3) fseek(NAME,6,SEEK_SET);
    if (CatN==4) fseek(NAME,4,SEEK_SET);

    //Search the name list
    N_Ext=0;
    for (i=0;i<MaxCatN[2];i++) {
      fread(sptr,2,1,NAME);
      N=*sptr;
      if (N>0){
        if (CatN==2) {
          fread(sptr,2,1,NAME);
          N_Ext=*sptr;
          fseek(NAME,4,SEEK_CUR);
        }
        if (CatN==4) fseek(NAME,2,SEEK_CUR);
        fgets(pnamebuf,121,NAME);
        NameStr=pnamebuf;
        NameStr=NameStr.Trim();

         //Search the name string
        if (NameStr!=""){
          MatchNChar=CompareString(SearchStr,NameStr);
          if ((MatchNChar>=BestMatchNChar)&&(MatchNChar>0)) {
            if (CatN==2) Descriptor="NGC "+IntToStr(N)+ExtS[N_Ext]+", "+NameStr;
            if (CatN==3) Descriptor="IC "+IntToStr(N)+", "+NameStr;
            if (CatN==4) Descriptor="M "+IntToStr(N)+", "+NameStr;
            AddToFoundList(CatN,N,N_Ext,MatchNChar,Descriptor,MaxNMatch);
            BestMatchNChar=MatchNChar;
            NMatch++;
          }
        }
        if (CatN==3) fseek(NAME,6,SEEK_CUR);
        if (CatN==4) fseek(NAME,4,SEEK_CUR);
      } else {
        fseek(NAME,126,SEEK_CUR);
      }
    }
    fclose(NAME);
  }

  //Case for Lynd and Barnard dark nebulae
  if ((CatN>4)&&(CatN<7)){
    char namebuf[21];
    int xref;
    FILE *infile;
    filename=PS.ProgPath+"Catalogs\\Barnard\\Barnard.dat";
    if (!FileExists(filename)) {
      String osbuf = "Could not open file " + filename;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
      return;
    }
    infile=fopen(filename.c_str(),"rb");

    //Note that there are 34 bytes per record.
    fseek(infile,0,SEEK_SET);

    for (i=0;i<MaxCatN[6];i++){
      fread(sptr,2,1,infile);
      N=*sptr;
      fread(sptr,2,1,infile); //Barnard catalogue only has a few "A" extensions
      N_Ext=*sptr;
      fseek(infile,8,SEEK_CUR);
      fgets(namebuf,21,infile);
      NameStr=namebuf;
      NameStr=NameStr.Trim();
      //Find a cross-reference to the Lynd catalog
      if (CatN==5){
        //X-Reference
        fread(sptr,2,1,infile);
        xref=*sptr;
      } else {
        fseek(infile,2,SEEK_CUR);
      }
      if (NameStr!=""){
        MatchNChar=CompareString(SearchStr,NameStr);
        if ((MatchNChar>=BestMatchNChar)&&(MatchNChar>0)) {
          if (CatN==6) Descriptor="Barnard "+IntToStr(N)+ExtS[N_Ext]+", "+NameStr;
          else Descriptor="Lynd "+IntToStr(xref)+", "+NameStr;
          AddToFoundList(CatN,N,N_Ext,MatchNChar,Descriptor,MaxNMatch);
          BestMatchNChar=MatchNChar;
          NMatch++;
        }
      }
    }
    fclose(infile);
  }

  //Case for Arp galaxies
  if (CatN==7){
    char namebuf_[56];

    FILE *NAME;
    filename=PS.ProgPath+"Catalogs\\Arp\\arp_xref.dat";
    if (!FileExists(filename)) {
      String osbuf = "Could not open file " + filename;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
      return;
    }
    NAME=fopen(filename.c_str(),"rb");
    fseek(NAME,0,SEEK_CUR);

    for (i=0;i<216;i++){
      fread(sptr,2,1,NAME);
      N=*sptr;
      fseek(NAME,8,SEEK_CUR);
      fgets(namebuf_,56,NAME);
      NameStr=namebuf_;
      NameStr=NameStr.Trim();
      if (NameStr!=""){
        MatchNChar=CompareString(SearchStr,NameStr);
        if ((MatchNChar>=BestMatchNChar)&&(MatchNChar>0)) {
          Descriptor="Arp "+IntToStr(N)+", "+NameStr;
          AddToFoundList(CatN,N,0,MatchNChar,Descriptor,MaxNMatch);
          BestMatchNChar=MatchNChar;
          NMatch++;
        }
      }
    }
    fclose(NAME);
  }

  //Case for RC3 galaxies (cross references to NGC names)
  if (CatN==8){
    int N_, N_Ext_;
    char pnamebuf[120];

    FILE *xref_file, *NAME;
    filename=PS.ProgPath+"Catalogs\\RC3\\RC3_XREF.dat";
    if (!FileExists(filename)) {
      String osbuf = "Could not open file " + filename;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
      return;
    }
    xref_file=fopen(filename.c_str(),"rb");
    fseek(xref_file,0,SEEK_SET);
    filename=PS.ProgPath+"Catalogs\\XRef\\Names.dat";
    if (!FileExists(filename)) {
      String osbuf = "Could not open file " + filename;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
      return;
    }
    NAME=fopen(filename.c_str(),"rb");

    for (i=0;i<MaxCatN[CatN];i++){
      //Determine any cross-references to NGC objects
      fseek(xref_file,8*i,SEEK_SET);
      fread(sptr,2,1,xref_file);
      N=*sptr;
      fread(sptr,2,1,xref_file);
      N_Ext=*sptr;
      fseek(xref_file,2,SEEK_CUR);
      if (N!=0){
        //See if NGC object has a name
        fseek(NAME,0,SEEK_SET);
        N_Ext_=0; N_=0; j=0;
        while (((N!=N_)||(N_Ext!=N_Ext_))&&(j<MaxCatN[2])) {
          fread(sptr,2,1,NAME);
          N_=*sptr;
          if (N_==N){
            fread(sptr,2,1,NAME);
            N_Ext_=*sptr;
            if (N_Ext_==N_Ext){
              fseek(NAME,4,SEEK_CUR);
              fgets(pnamebuf,121,NAME);
              NameStr=pnamebuf;
              NameStr=NameStr.Trim();
              //Search the name string
              if (NameStr!=""){
                MatchNChar=CompareString(SearchStr,NameStr);
                if ((MatchNChar>=BestMatchNChar)&&(MatchNChar>0)) {
                  Descriptor="RC3 "+IntToStr(i)+", "+NameStr;
                  AddToFoundList(CatN,N,0,MatchNChar,Descriptor,MaxNMatch);
                  BestMatchNChar=MatchNChar;
                  NMatch++;
                }
              }
            } else {
              fseek(NAME,4,SEEK_CUR);
            }
          } else {
            fseek(NAME,126,SEEK_CUR);
          }
          j++;
        }
      } else {
        fseek(xref_file,4,SEEK_CUR);
      }
    }
    fclose(NAME);
    fclose(xref_file);
  }

  //Search the custom database for matches

  //Display search results
  for (i=0;i<50;i++){
    if (i<MaxNMatch) FoundList->Items->Strings[i]=BestMatchDesc[i];
    else FoundList->Items->Strings[i]=" ";
  }
  FoundList->TopIndex=0;
  SearchIndicator->Caption="";
}
//---------------------------------------------------------------------------
//Returns the maximum number of consecutive matching characters in Str1
//and Str2: NOT CASE SENSITIVE.  Str1 should be the short string.
int CompareString(String Str1, String Str2)
{
  int k, len1, len2, maxk=0;
  bool EndedMatchStr;
  bool BegunMatchStr;

  Str1=Str1.UpperCase();
  len1=Str1.Length();
  Str2=AnsiString::StringOfChar(' ',len1-1)+Str2.UpperCase()+AnsiString::StringOfChar(' ',len1-1);
  len2=Str2.Length();

  EndedMatchStr=true;
  BegunMatchStr=false;
  for (int j=0;j<=len2-len1;j++){
    k=0;
    for (int i=1;i<=len1;i++){
      if (Str1[i]==Str2[i+j]&&EndedMatchStr){
        BegunMatchStr=true;
        k++;
      } else {
        if (BegunMatchStr){
          BegunMatchStr=false;
          EndedMatchStr=true;
          if (k>maxk) maxk=k;
          k=0;
        }
      }
    }
    if (BegunMatchStr) {
      BegunMatchStr=false;
      EndedMatchStr=true;
      if (k>maxk) maxk=k;
    }
    if (k==len1) return maxk;
  }
  return maxk;
}
//---------------------------------------------------------------------------
//Add object to the match list, sorted in order of decreasing match "goodness"
void AddToFoundList(int CatN,int N,int N_Ext,int MatchNChar,String Desc,int MaxNMatch)
{
  //BestMatch[i][0]=Object catalogue
  //BestMatch[i][1]=Object number
  //BestMatch[i][2]=Object extension number
  //BestMatch[i][3]=Number of matching characters
  //BestMatchDesc[i]=Object descriptor string;

  //Find where to insert the data
  int i=0;
  while (MatchNChar<=BestMatches[i][3] && i<MaxNMatch){i++;}

  //Shift the list items to make room for the new item
  if (i<MaxNMatch){
    for (int j=MaxNMatch-1;j>i;j--){
      for (int k=0;k<4;k++){
        BestMatches[j][k]=BestMatches[j-1][k];
        BestMatchDesc[j]=BestMatchDesc[j-1];
      }
    }
    //Insert the item
    BestMatches[i][0]=CatN;
    BestMatches[i][1]=N;
    BestMatches[i][2]=N_Ext;
    BestMatches[i][3]=MatchNChar;
    BestMatchDesc[i]=Desc;
  }
}
//---------------------------------------------------------------------------
//Display the information for the object selected from the search list
void __fastcall TObjectInfo::FoundListClick(TObject *Sender)
{
  int i=FoundList->ItemIndex;
  SelectedCatalogue=BestMatches[i][0]-2;
  CatNum[SelectedCatalogue+2]=BestMatches[i][1];
  ExtNum=BestMatches[i][2];

  ShowObjectInfo(Sender);

  //Center the field of view on the object if requested
  if (CenterOnFound->Checked) CenterButClick(Sender);
}
//---------------------------------------------------------------------------
//Associate an image file with the object or change the image file
void __fastcall TObjectInfo::OImageClick(TObject *Sender)
{
  String filename;
  if (!ImageExists){
    OpenD->InitialDir=PS.ProgPath+"Images";
    OpenD->FileName="";
    OpenD->Execute();
    filename=OpenD->FileName;
    //Add an image to the list
    if (FileExists(filename)){
      short *sptr, sdum=0;
      sptr=&sdum;
      int NRec;

      FILE *iofile;
      String filename2 = PS.ProgPath+"Images\\ImageFiles.dat";
      if (!FileExists(filename2)) {
        String osbuf = "Could not open file " + filename2;
        Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
        return;
      }
      iofile=fopen(filename2.c_str(),"r+b");

      //The first 2 bytes of the output file contain the number of records
      fseek(iofile,0,SEEK_SET);
      fread(sptr,1,2,iofile);
      NRec=*sptr;

      fseek(iofile,0,SEEK_SET);
      sdum=NRec+1;
      fwrite(&sdum,1,2,iofile);
      fseek(iofile,NRec*142,SEEK_CUR);  //skip to the last record

      //NGC #
      if (ShowAsRNGC->Enabled) sdum=CatNum[2];
      else sdum=0;
      fwrite(&sdum,1,2,iofile);
      sdum=ExtNum;
      fwrite(&sdum,1,2,iofile);

      //IC #
      if (ShowAsIC->Enabled) sdum=CatNum[3];
      else sdum=0;
      fwrite(&sdum,1,2,iofile);

      //Messier #
      if (ShowAsMes->Enabled) sdum=CatNum[4];
      else sdum=0;
      fwrite(&sdum,1,2,iofile);

      //Other catalogue information
      if (SelectedCatalogue>2){
        sdum=SelectedCatalogue;
        fwrite(&sdum,1,2,iofile);
        sdum=CatNum[SelectedCatalogue+2];
        fwrite(&sdum,1,2,iofile);
        sdum=ExtNum;
        fwrite(&sdum,1,2,iofile);
      } else {
        sdum=0;
        fwrite(&sdum,1,2,iofile);
        fwrite(&sdum,1,2,iofile);
        fwrite(&sdum,1,2,iofile);
      }

      //The file name (you must remove the path - it must be stored in
      //the "Images" directory)
      ImageFile=filename;
      int i=filename.LastDelimiter("\\");
      filename=filename.SubString(i+1,filename.Length()-i);
      while (filename.Length()<128) filename+=" ";
      fprintf(iofile,"%s",filename.c_str()); //Name as 128 byte char
      fclose(iofile);

      DisplayImage(true,ImageFile);
      ImageExists=true;
    }
  } else {
    //Change the image associated with the object.  To keep things simple,
    //records can never be removed, only edited.  If the file is removed from
    //the system the code will not fail.
    OpenD->InitialDir=PS.ProgPath+"Images";
    OpenD->FileName=ImageFile;
    OpenD->Execute();
    filename=OpenD->FileName;

    if (FileExists(filename)){
      short *sptr, sdum=0;
      sptr=&sdum;
      int NRec;

      FILE *iofile;
      String filename2 = PS.ProgPath+"Images\\ImageFiles.dat";
      if (!FileExists(filename2)) {
        String osbuf = "Could not open file " + filename2;
        Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
        return;
      }
      iofile=fopen(filename2.c_str(),"r+b");

      fseek(iofile,(ImageIndex-1)*142+2,SEEK_CUR);  //skip to record to be edited

      //NGC #
      if (ShowAsRNGC->Enabled) sdum=CatNum[2];
      else sdum=0;
      fwrite(&sdum,1,2,iofile);
      sdum=ExtNum;
      fwrite(&sdum,1,2,iofile);

      //IC #
      if (ShowAsIC->Enabled) sdum=CatNum[3];
      else sdum=0;
      fwrite(&sdum,1,2,iofile);

      //Messier #
      if (ShowAsMes->Enabled) sdum=CatNum[4];
      else sdum=0;
      fwrite(&sdum,1,2,iofile);

      //Other catalogue information
      if (SelectedCatalogue>2){
        sdum=SelectedCatalogue;
        fwrite(&sdum,1,2,iofile);
        sdum=CatNum[SelectedCatalogue+2];
        fwrite(&sdum,1,2,iofile);
        sdum=ExtNum;
        fwrite(&sdum,1,2,iofile);
      } else {
        sdum=0;
        fwrite(&sdum,1,2,iofile);
        fwrite(&sdum,1,2,iofile);
        fwrite(&sdum,1,2,iofile);
      }

      //The file name (you must remove the path - it must be stored in
      //the "Images" directory)
      ImageFile=filename;
      int i=filename.LastDelimiter("\\");
      filename=filename.SubString(i+1,filename.Length()-i);
      while (filename.Length()<128) filename+=" ";
      fprintf(iofile,"%s",filename.c_str()); //Name as 128 byte char
      fclose(iofile);

      DisplayImage(true,ImageFile);
    }
  }
}
//---------------------------------------------------------------------------
//Display the deep-sky object image
void __fastcall TObjectInfo::DisplayImage(bool Found,String filename)
{
  if (Found) FullImgBtn->Enabled=true;
  else FullImgBtn->Enabled=false;

  if (Found && FileExists(filename)){
    float w_to_h;
    Graphics::TBitmap *bmp = new Graphics::TBitmap();
    bmp->LoadFromFile(filename);
    w_to_h = (float)bmp->Width/(float)bmp->Height;
    if (w_to_h<1){
      OImage->Height=OPanel->Height-6;
      OImage->Width=OImage->Height*w_to_h;
      OImage->Top=2;
      OImage->Left=(OPanel->Width-OImage->Width)/2+3;
    } else {
      OImage->Width=OPanel->Width-6;
      OImage->Height=OImage->Width/w_to_h;
      OImage->Left=3;
      OImage->Top=(OPanel->Height-OImage->Height)/2+2;
    }
    OImage->Picture->Bitmap=bmp;
    ImageFile=filename;
    ImageExists=true;
  } else {
    OImage->Top=2;
    OImage->Left=3;
    OImage->Height=OPanel->Height-6;
    OImage->Width=OPanel->Width-6;
    String bmpfile = PS.ProgPath+"Images\\NoImage.bmp";
    if (!FileExists(bmpfile)) {
      String osbuf = "Could not open file " + bmpfile;
      Application->MessageBox(osbuf.c_str(), "SkyAtlas", MB_OK);
    } else {
      OImage->Picture->LoadFromFile(bmpfile);
    }  
  }
}
//---------------------------------------------------------------------------
//Display the full resolution image in a separate window
void __fastcall TObjectInfo::FullImgBtnClick(TObject *Sender)
{
  ImgForm->ShowModal();
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::OkButClick(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------------
void __fastcall TObjectInfo::FormClose(TObject *Sender,
      TCloseAction &Action)
{
  PS.PlanetResolution = PlanetRes->Position;
}
//---------------------------------------------------------------------------

