//---------------------------------------------------------------------------
#include <vcl.h>
#include <stdio.h>
#pragma hdrstop

#include "PPaths.h"
#include "Main.h"
#include "ServiceRoutines.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TPPathForm *PPathForm;

extern Settings PS;
extern OrbitalElements SSData[213]; //Information for all solar system objects
extern bool DrawingMap;

extern double StripBlank(String s);
extern void LoadPPSettings(void);
extern void SavePPSettings(void);

//---------------------------------------------------------------------------
__fastcall TPPathForm::TPPathForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
//Form opens
void __fastcall TPPathForm::FormShow(TObject *Sender)
{
  LoadPPSettings();
  LoadForm(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TPPathForm::OKBtnClick(TObject *Sender)
{
  SavePPSettings();
  Close();
}
//---------------------------------------------------------------------------
//Load values onto the form
void __fastcall TPPathForm::LoadForm(TObject *Sender)
{
  int N;

  if (PlanetName->ItemIndex<0) PlanetName->ItemIndex=0;
  N=PlanetName->ItemIndex;

  //All non-standard satelites (of earth and sun) are appended to the end
  //of the list
  if (N>8 && N<10) N=100;
  if (N>10 && N<12) N=N-10+200;
  if (N>11 && N<16) N=N-12+202;
  if (N>15 && N<21) N=N-16+206;

  Increment->Text=FormatFloat("#######0.0",SSData[N].Increment);
  EndTime->Text=FormatFloat("#######0.0",SSData[N].PathED);
  StartTime->Text=FormatFloat("#######0.0",SSData[N].PathSD);
  DrawPath->Checked=SSData[N].ShowPath;
  DrawPoints->Checked=SSData[N].ShowPoints;
  DrawLines->Checked=SSData[N].ShowLines;
  ShowDates->Checked=SSData[N].ShowLabels;

  PointColImg->Canvas->Pen->Color=SSData[N].PointCol;
  PointColImg->Canvas->Brush->Color=SSData[N].PointCol;
  PointColImg->Canvas->Rectangle(0,0,PointColImg->Width,PointColImg->Height);

  LineColImg->Canvas->Pen->Color=SSData[N].LineCol;
  LineColImg->Canvas->Brush->Color=SSData[N].LineCol;
  LineColImg->Canvas->Rectangle(0,0,LineColImg->Width,LineColImg->Height);

  LabelColImg->Canvas->Pen->Color=SSData[N].LabelCol;
  LabelColImg->Canvas->Brush->Color=SSData[N].LabelCol;
  LabelColImg->Canvas->Rectangle(0,0,LabelColImg->Width,LabelColImg->Height);

  Repaint();
}
//---------------------------------------------------------------------------
void __fastcall TPPathForm::PlanetNameSelect(TObject *Sender)
{
  PlanetName->Text=PlanetName->Items->Strings[PlanetName->ItemIndex];
  LoadForm(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TPPathForm::ApplyBtnClick(TObject *Sender)
{
  int N;

  if (PlanetName->ItemIndex<0) PlanetName->ItemIndex=0;
  N=PlanetName->ItemIndex;

  //All non-standard satelites (of earth and sun) are appended to the end
  //of the list
  if (N>8 && N<10) N=100;
  if (N>10 && N<12) N=N-10+200;
  if (N>11 && N<16) N=N-12+202;
  if (N>15 && N<21) N=N-16+206;

  SSData[N].Increment=StripBlank(Increment->Text);
  SSData[N].PathED=StripBlank(EndTime->Text);
  SSData[N].PathSD=StripBlank(StartTime->Text);
  SSData[N].ShowPath=DrawPath->Checked;
  SSData[N].ShowPoints=DrawPoints->Checked;
  SSData[N].ShowLines=DrawLines->Checked;
  SSData[N].ShowLabels=ShowDates->Checked;

  SSData[N].PointCol=PointColImg->Canvas->Pixels[0][0];
  SSData[N].LineCol=LineColImg->Canvas->Pixels[0][0];
  SSData[N].LabelCol=LabelColImg->Canvas->Pixels[0][0];

  while (DrawingMap);
  DrawingMap=true;
  MainF->DrawMap(PS.MaxStarMag,false);
  DrawingMap=false;
}
//---------------------------------------------------------------------------
void __fastcall TPPathForm::PointColImgClick(TObject *Sender)
{
  ColorDialog->Color=PointColImg->Canvas->Pixels[0][0];
  ColorDialog->Execute();
  PointColImg->Canvas->Pen->Color=ColorDialog->Color;
  PointColImg->Canvas->Brush->Color=ColorDialog->Color;
  PointColImg->Canvas->Rectangle(0,0,PointColImg->Width,PointColImg->Height);
}
//---------------------------------------------------------------------------
void __fastcall TPPathForm::LineColImgClick(TObject *Sender)
{
  ColorDialog->Color=LineColImg->Canvas->Pixels[0][0];
  ColorDialog->Execute();
  LineColImg->Canvas->Pen->Color=ColorDialog->Color;
  LineColImg->Canvas->Brush->Color=ColorDialog->Color;
  LineColImg->Canvas->Rectangle(0,0,LineColImg->Width,LineColImg->Height);
}
//---------------------------------------------------------------------------
void __fastcall TPPathForm::LabelColImgClick(TObject *Sender)
{
  ColorDialog->Tag=2;
  ColorDialog->Color=LabelColImg->Canvas->Pixels[0][0];
  ColorDialog->Execute();
  LabelColImg->Canvas->Pen->Color=ColorDialog->Color;
  LabelColImg->Canvas->Brush->Color=ColorDialog->Color;
  LabelColImg->Canvas->Rectangle(0,0,LabelColImg->Width,LabelColImg->Height);
}
//---------------------------------------------------------------------------


